/*  $Id: ActionStack.cc,v 1.10 1998/08/22 03:35:07 shaggy Exp $
 *      C/C++ Parser (yacc grammar)
 *      
 *      relative path: .//lib/cparse
 *  
 *      Shag/OS: an Object-Oriented MicroKernel Operating System
 *       Libraries
 *        C/C++ Parsing classes
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the Shag/OS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare Non-profit Public License (SNPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare Non-profit Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

/*  
 *  $Log: ActionStack.cc,v $
 *  Revision 1.10  1998/08/22 03:35:07  shaggy
 *  header update
 *
 *  Revision 1.9  1998/08/12 15:18:30  shaggy
 *  better typing; fixed some type coercions;
 *  cleanup for #include's, and fixed some constructor call bugs
 *
 *  Revision 1.8  1998/07/28 16:43:14  shaggy
 *  moved lots of Deco into generic CParse lib
 *  CAction split into vCAction and CAction
 *  printaction -> part of CAction
 *  ccint -> CRep (C Representation)
 *  ccout -> COut (C output)
 *
 *  Revision 1.7  1998/07/19 06:46:43  shaggy
 *  1st stage of Item cleanup
 *
 *  Revision 1.6  1998/07/04 01:23:28  shaggy
 *  anonymous unions work now
 *
 *  Revision 1.5  1998/05/14 05:08:38  shaggy
 *  inheritence works and lots of bugs fixed
 *
 *  Revision 1.4  1998/03/23 14:17:01  shaggy
 *  lots more changes
 *
 *  Revision 1.3  1998/02/10 12:43:35  shaggy
 *  testing
 *
 *  Revision 1.2  1998/01/12 04:12:48  shaggy
 *  got dependencies working when re-generating code
 *
 *  Revision 1.1  1998/01/11 05:06:33  shaggy
 *  Initial revision
 *
 */


#include <defines>
#include <types.h>
#include <inttypes.h>
#include <posix/stdarg.h>
#include <posix/string.h>
#include <vCAction>
#include "ActionStack.h"


#define INTERNAL_ERROR	_caction->internal_error(__FILE__, __LINE__)

#define T(x)	vCAction::x


ActionStack::ActionStack(vCAction *caction)
{
	/* XXX - needs to be able to dynamically enlarge this later,
	   and hopefully dynamically shrink it too, if memory becomes tight,
	   and it's not currently being used anymore.
	   Actually, we could just dynamically add pieces to the stack,
	   and check the index to see if it's within the range of the
	   current piece, and find the correct piece otherwise 
	   - need to be careful about it "thrashing" on the boundary though */

	_caction = caction;
	_stacksize = 1024;
	_stack = new Node[_stacksize];
	_sp = 0;
}

ActionStack::~ActionStack()
{
	delete _stack;
}

void
ActionStack::push(Item *range, Item_ linkrange, int actionindex,
		Item_ p1, Item_ p2, Item_ p3)
{
	if(_sp >= _stacksize) 
		_caction->abort("action stack overflow");
	Node *csp = &_stack[_sp];
	csp->actionindex = actionindex;
	csp->param1 = p1;  csp->param2 = p2;  csp->param3 = p3;
	range->token = T(ACTION);
	if(linkrange.token == T(ACTION)) {
		// make sure we're linking to the next item on the stack...
		if(linkrange.Action.top+1 != _sp) INTERNAL_ERROR;
		range->Action.bottom = linkrange.Action.bottom;
	} else
		range->Action.bottom = _sp;
	range->Action.top = _sp++;
}


void
ActionStack::pop_lifo(Item *result, Item_ range, 
			const Action *action_table, Item_ param)
{
	if(range.token == T(NONE)) {
		*result = param;
		return;
	}
	if(range.token != T(ACTION)) INTERNAL_ERROR;
	int top = range.Action.top;
	int bottom = range.Action.bottom;

	*result = param;
	while(top >= bottom) {
		// _caction->eprintf("lifo: bottom=%d top=%d\n", bottom, top);
		Node *csp = &_stack[top--];
		if(csp->param3.token == T(TBC)) csp->param3 = *result;
		if(csp->param2.token == T(TBC)) csp->param2 = *result;
		if(csp->param1.token == T(TBC)) csp->param1 = *result;

		if(csp->param3.token == T(ACTION)) pop_lifo(&csp->param3,
				csp->param3, action_table, *result);
		if(csp->param2.token == T(ACTION)) pop_lifo(&csp->param2,
				csp->param2, action_table, *result);
		if(csp->param1.token == T(ACTION)) pop_lifo(&csp->param1,
				csp->param1, action_table, *result);

		Action action = action_table[csp->actionindex];
		if(action) 
			(_caction->*action)(result, csp->param1,
				csp->param2, csp->param3);
		else
			*result = csp->param2;
	}
	if(bottom < _sp)
		_sp = bottom;
}

#if 0	/* probably not needed anymore */
void
ActionStack::pop_fifo(Item *result, Item_ range,
			const Action *action_table, Item_ param)
{
	if(range.token == T(NONE)) {
		*result = param;
		return;
	}
	if(range.token != T(ACTION)) INTERNAL_ERROR;
	int top = range.Action.top;
	int bottom = range.Action.bottom;

	*result = param;
	while(bottom <= top) {
		_caction->eprintf("fifo: bottom=%d top=%d\n", bottom, top);
		Node *csp = &_stack[bottom++];
		if(csp->param1.token == T(TBC)) csp->param1 = *result;
		if(csp->param2.token == T(TBC)) csp->param2 = *result;
		if(csp->param3.token == T(TBC)) csp->param3 = *result;

		if(csp->param1.token == T(ACTION)) pop_lifo(&csp->param1,
				csp->param1, action_table, *result);
		if(csp->param2.token == T(ACTION)) pop_lifo(&csp->param2,
				csp->param2, action_table, *result);
		if(csp->param3.token == T(ACTION)) pop_lifo(&csp->param3,
				csp->param3, action_table, *result);

		Action action = action_table[csp->actionindex];
		if(action) 
			(_caction->*action)(result, csp->param1,
				csp->param2, csp->param3);
		else
			*result = csp->param2;
	}
	if(bottom < _sp)
		_sp = bottom;
}
#endif

