/*  $Id: CAction.cc,v 1.57 1998/08/22 03:35:08 shaggy Exp $
 *      standard C/C++ parsing actions
 *      
 *      relative path: .//lib/cparse
 *  
 *      Shag/OS: an Object-Oriented MicroKernel Operating System
 *       Libraries
 *        C/C++ Parsing classes
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the Shag/OS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare Non-profit Public License (SNPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare Non-profit Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

//  $Log: CAction.cc,v $
//  Revision 1.57  1998/08/22 03:35:08  shaggy
//  header update
//
//  Revision 1.56  1998/08/17 17:37:55  shaggy
//  fixed to work with gcc 2.8.1
//
//  Revision 1.55  1998/08/17 06:38:14  shaggy
//  more fixes for forward refs; __attribute__ support in aggrs
//
//  Revision 1.54  1998/08/13 04:26:33  shaggy
//  bug fixes and dynamic sized arrays
//
//  Revision 1.53  1998/08/12 15:18:30  shaggy
//  better typing; fixed some type coercions;
//  cleanup for #include's, and fixed some constructor call bugs
//
//  Revision 1.52  1998/08/11 12:50:41  shaggy
//  getting conditional assignment to work
//
//  Revision 1.51  1998/08/11 02:05:20  shaggy
//  deco can finally parse itself
//
//  Revision 1.50  1998/08/10 21:11:00  shaggy
//  fixed bug in matching constructor calls
//
//  Revision 1.49  1998/08/10 13:09:42  shaggy
//  some code cleanup (casts, etc)
//
//  Revision 1.48  1998/08/10 03:11:57  shaggy
//  deco 0.10
//
//  Revision 1.47  1998/08/09 03:57:32  shaggy
//  lots of work on output dependency checking
//
//  Revision 1.46  1998/08/04 00:45:33  shaggy
//  lots of fixes so it builds under Solaris again
//
//  Revision 1.45  1998/08/03 15:00:04  shaggy
//  conversion operators, constructor conversions, default params
//
//  Revision 1.44  1998/08/03 00:12:54  shaggy
//  tons of work-- friends, class scope fixes, tons of bugfixes,
//  lots more cleanup, more DC++ work, etc
//
//  Revision 1.43  1998/08/02 04:29:38  shaggy
//  more fixes
//
//  Revision 1.42  1998/08/01 21:02:04  shaggy
//  much more cleanup; many bugfixes; DC++ progress
//
//  Revision 1.41  1998/08/01 14:56:44  shaggy
//  about to restructure deco
//
//  Revision 1.40  1998/07/31 23:44:23  shaggy
//  running out of time
//
//  Revision 1.39  1998/07/31 04:15:24  shaggy
//  need sleep
//
//  Revision 1.38  1998/07/30 22:27:17  shaggy
//  broken
//
//  Revision 1.37  1998/07/30 16:16:28  shaggy
//  switched to new vsn of shacc and removed useless underscore prefixes
//  as well as lot of extra configuration code that is no longer needed
//
//  Revision 1.36  1998/07/30 04:40:29  shaggy
//  much more dtype and dclass work
//
//  Revision 1.35  1998/07/29 03:35:09  shaggy
//  dtype support nearly complete, initial dclass support
//
//  Revision 1.34  1998/07/28 19:55:24  shaggy
//  moved most of generic parsing from Deco into CAction
//
//  Revision 1.33  1998/07/28 16:43:15  shaggy
//  moved lots of Deco into generic CParse lib
//  CAction split into vCAction and CAction
//  printaction -> part of CAction
//  ccint -> CRep (C Representation)
//  ccout -> COut (C output)
//
//  Revision 1.32  1998/07/19 06:46:44  shaggy
//  1st stage of Item cleanup
//
//  Revision 1.31  1998/07/05 20:32:53  shaggy
//  tons of bugfixes, enhancements and parser improvements
//
//  Revision 1.30  1998/07/04 20:38:55  shaggy
//  lots of bug fixes; fn, aggr, etc improvements, new/delete
//
//  Revision 1.29  1998/06/29 14:14:15  shaggy
//  more DECO conversions, support for 'this', etc
//
//  Revision 1.28  1998/06/29 01:36:50  shaggy
//  fixed inline attr, output of pure virtuals, etc...
//
//  Revision 1.27  1998/06/02 14:50:38  shaggy
//  in the midst of more full parsing updates
//
//  Revision 1.26  1998/06/01 13:28:57  shaggy
//  switchable code save/parse
//
//  Revision 1.25  1998/05/26 05:02:51  shaggy
//  parameter scoping works correctly
//  most basic C/C++ statements now work
//
//  Revision 1.24  1998/05/26 01:39:45  shaggy
//  params almost working right
//
//  Revision 1.23  1998/05/25 23:28:41  shaggy
//  major rework of func/param info
//
//  Revision 1.22  1998/05/15 07:03:35  shaggy
//  dependency ordering fixed; constructor initializers; lots of bug fixes, etc
//
//  Revision 1.21  1998/05/15 00:17:23  shaggy
//  ctor initializers work, and enums are fixed so declarations
//  in an enum can be immediately re-used
//
//  Revision 1.20  1998/05/14 18:02:02  shaggy
//  fixed constant handling, enums, function cv tags, etc
//
//  Revision 1.19  1998/05/13 00:17:40  shaggy
//  scoping works and enums fixed (mostly)
//
//  Revision 1.18  1998/05/12 05:00:11  shaggy
//  more scope work
//
//  Revision 1.17  1998/05/10 18:53:39  shaggy
//  just about ready to start fixing scoping
//
//  Revision 1.16  1998/05/10 00:03:29  shaggy
//  string concatenation works, and comments are being
//  added to get ready to fix scoping
//
//  Revision 1.15  1998/05/04 03:09:17  shaggy
//  support for ctors, dtors, conversion operators, operator overloading, etc
//
//  Revision 1.14  1998/05/03 02:47:06  shaggy
//  added inheritence to parser
//
//  Revision 1.13  1998/04/16 13:35:43  shaggy
//  access control
//
//  Revision 1.12  1998/04/15 13:39:42  shaggy
//  added types to constants
//
//  Revision 1.11  1998/04/11 21:49:26  shaggy
//  added token saving (for regurgitation)
//
//  Revision 1.10  1998/04/06 14:28:06  shaggy
//  got enums almost working
//
//  Revision 1.9  1998/03/22 03:14:55  shaggy
//  almost there...
//
//  Revision 1.8  1998/03/21 00:23:48  shaggy
//  All kinds of new changes...
//  and it works!!!
//
//  Revision 1.7  1998/02/11 04:33:44  shaggy
//  print scopes
//
//  Revision 1.6  1998/02/08 19:07:05  shaggy
//  about to start splitting up everything
//
//  Revision 1.5  1998/01/12 04:12:48  shaggy
//  got dependencies working when re-generating code
//
//  Revision 1.4  1998/01/11 05:06:33  shaggy
//  about to switch to standard ref/ptr semantics
//
//  Revision 1.3  1996/11/23  17:21:35  shaggy
//  last checkin before switching to virtual functions
//
//  Revision 1.2  1996/10/19  21:47:37  shaggy
//  modified to handle new tokens
//
//  Revision 1.1  1996/04/10  12:13:48  shaggy
//  Initial revision
//

#define CACTION_DEBUG

#include <defines>
#include <types.h>
#include <inttypes.h>
#include <posix/stdarg.h>
#include <posix/string.h>
#include <posix/stdio.h>
#include <posix/assert.h>
#include <vCAction>
#include <CObj>
#include <CLex>
#include <StrTab>
#include <SymTab>
#include <ParsePath>
#include <CParse>
#include <CRep>
#include <COut>

#include "CAction.hh"


// some lexical shortcuts:
typedef vStrTab::Id Id;
typedef vCAction::Item Item;
typedef const Item &Item_;
typedef vCAction::Token Token;
typedef CLex::Token LToken;

const Item vCAction::none(T(NONE));


CAction::CAction(CLex &clp, vStrTab &strtab, vSymTab &globlsym)
                  : _clex(clp), _strtab(strtab), _global_scope(globlsym),
	   	    ccout(*this, &globlsym)
{
	_complete = true;
	_num_errors = 0;
	_num_warnings = 0; 
	_anon = 0;
	_notedecl = false;
	_notedef = false;
	_warnerr = false;
	_realtypes = false;
	_allscopes += global();
	_otype = Decl::OT_CODE;
	_inclevel = 0;

	// predefines for C++:
	_id_this = strid("this");
}


void
CAction::accept() {}

void
CAction::eol() {}


csz *
CAction::cstr(Id id) const
{
	switch((ulong)id) {
	case 0:	return "<nil>";
	case (ulong)ID_ANON: return "<anon>";
	case (ulong)ID_INHERIT: return "<inherit>";
	case (ulong)ID_USING: return "<using>";
	case (ulong)ID_FRIEND: return "<friend>";
	default: return _strtab.string(id);
	}
}
 
csz *
CAction::cstr(Token token) const
{
	switch(token) {
	case TEMP:	return "<TEMP>";
	case RESULT:	return "<RESULT>";
	case INSN:	return "<INSN>";
	case LIST:	return "<LIST>";
	case TREE:	return "<TREE>";
	case ARRAY:	return "<ARRAY>";
	case OBJ:	return "<OBJ>";
	case TYPE:	return "<TYPE>";
	case TYPELIST:	return "<TYPELIST>";
	case PARMLIST:	return "<PARMLIST>";
	case EXPRLIST:	return "<EXPRLIST>";
	case STMTLIST:	return "<STMTLIST>";
	case ERR:	return "<ERROR>";
	case ERRLIST:	return "<ERRLIST>";
	case STRUCT:	return "<STRUCT>";
	case CLASS:	return "<CLASS>";
	case VAR:	return "<VAR>";
	case FUNC:	return "<FUNC>";
	default:
		return _clex.token_string((LToken)token);
	}                                                   
}


sz *
CAction::cstr(Item_ item, sz *buf, size_t buflen,
				bool show_type) const
{
	sz internal_buf[4096];
	csz *s = internal_buf;
	sz *bp = buf;
	Token t = item.token;

	if(show_type) {
		size_t len = sprintf(buf, "%s:", cstr(t));
		if(len >= buflen)
			return 0;
		bp += len;	
		buflen -= len;
	}

	/* XXX - these are here during initial development, but should
	   be removed later, since they rely on other XXX_ID types having
	   the same layout as an ID, which is not guaranteed
 	*/
	switch(t) {
	case AGGR_ID: case ENUM_ID: case NSPACE_ID: case ASPACE_ID:
	case TYPE_ID: case TPLATE_ID:
	case VAL_ID: case DTOR_ID: case CONV_ID: case OPER_ID:
		t = ID;
	default:
		break;
	}

	switch(t) {
        case NUM:
		s = _clex.token_string((LToken)t, item.val,
				item.c.is_signed, item.c.is_long);
                break;
        case CHAR:
		s = _clex.token_string((LToken)t, item.val);
                break;
	case BOOL:
		s = _clex.token_string((LToken)t, (bool)item.val);
                break;
        // case LINE:
        // case CSTR:
	//	s = _clex.token_string((LToken)t, item.str);
  	//      break;
        case ID:
        case STR:
        case BACKSTR:
        case ANGSTR:

		s = _clex.token_string((LToken)t, cstr(item.id));
		if(t == ID) {
			sprintf(internal_buf, "%p:%s", item.scope, s);
			s = internal_buf;
		}
                break;
        default:
                sprintf(internal_buf, "%p:0x%llx", item.scope, item.val);       
	}

	size_t len = strlen(s);
	if(len < buflen)
		strcpy(bp, s);
	else {
		memcpy(bp, s, buflen/2);
		memcpy(bp+buflen/2, "...", 3);
		memcpy(bp+buflen/2+3, s+len-(buflen+1)/2+4, (buflen+1)/2-3);
	}
	
	return buf; 
}


void
CAction::eprintf(csz *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	veprintf(fmt, ap);
	va_end(ap);
}


void
CAction::vmprintf(csz *prefix, csz *fmt, va_list args)
{
	eprintf("%s:%d: %s", _clex.source(), _clex.line_num(), prefix);
	veprintf(fmt, args);
	switch(_clex.last_token()) {
	case NONE:
		eprintf("\n");
		break;
	case EOF_COMPLETE:
	case EOF_INCOMPLETE:
		eprintf("; at end of file\n"); 
		break;
	case INVALID:
		eprintf("; at `%s'\n", _clex.token_string());
		break;
	default:
		eprintf("; before `%s'\n", _clex.token_string());
	}
	_clex.set_token((LToken)NONE);
}


void
CAction::mprintf(csz *prefix, csz *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	vmprintf(prefix, fmt, ap);
	va_end(ap);
}


void
CAction::error(csz *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	_num_errors++;
	vmprintf("", fmt, ap);
	va_end(ap);
}


void
CAction::warning(csz *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	_num_warnings++;
	vmprintf("warning: ", fmt, ap);
	va_end(ap);
}


void
CAction::abort(csz *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	vmprintf("abort: ", fmt, ap);
	va_end(ap);
	abort_exit();
}


void
CAction::internal_error(csz *file, int line, csz *module)
{
	if(module) 
		mprintf("Internal Error ", "in %s (%s:%d)",
			module, file, line);
	else
		mprintf("Internal Error ", "(%s:%d)", file, line);
	internal_error_exit();
}


void
CAction::_assertfail(csz *cond, csz *file, int line)
{
	mprintf("", "Assertion \"%s\" failed", cond);
	internal_error(file, line, 0);
}


void
CAction::mprint_action(csz *prefix, csz *name,
			Item_ p1, Item_ p2, Item_ p3)
{
	sz buf1[50], buf2[50], buf3[50];
	mprintf(prefix, "%s(%s, %s, %s)", name,
		cstr(p1, buf1, sizeof(buf1), true),
		cstr(p2, buf2, sizeof(buf2), true),
		cstr(p3, buf3, sizeof(buf3), true));
}


void
CAction::mprint_result(csz *prefix, Item_ r)
{
	sz buf[50];
	mprintf(prefix, "\tresult: %s", cstr(r, buf, sizeof(buf), true));
}



void
CAction::unsupported(csz *name,
			Item_ p1, Item_ p2, Item_ p3)
{
	mprint_action("unsupported: ", name, p1, p2, p3);
	internal_error(__FILE__, __LINE__);
}

void
CAction::Test(Item *r, Item_ p1, Item_ p2, Item_ p3)
{
	mprint_action("", "Test", p1, p2, p3);
	*r = p2;
}

void
CAction::Index(Item *r, Item_ array, Item_ index, Item_ )
{
	BinaryOp(r, OP_INDEX, array, index);
}

/*
void
CAction::DoWhile(Item *r, Item_ scond, Item_ code, Item_ econd)
{
	While(r, scond, code, econd);
}
*/

void
CAction::Break(Item *r, Item_ where, Item_, Item_)
{
	Goto(r, where, none, none);
}

void
CAction::Continue(Item *r, Item_ where, Item_, Item_)
{
	Goto(r, where, none, none);
}

void
CAction::NewAggr(Item *r, Item_ type, Item_ sym, Item_ scope)
{
	NewScope(r, type, sym, scope);
}

void
CAction::NewNspace(Item *r, Item_ type, Item_ sym, Item_ scope)
{
	NewScope(r, type, sym, scope);
}

void
CAction::NewLocal(Item *r, Item_ type, Item_ sym, Item_ scope)
{
	NewScope(r, type, sym, scope);
}

void
CAction::FreeScope(Item *, Item_, Item_, Item_) {}


TDecl::DataType
datatypeof(Token tok)
{
	switch(tok) {
	case CAction::KW_INT: return TDecl::dt_int;
	case CAction::KW_CHAR: return TDecl::dt_char;
	case CAction::KW_WCHAR: return TDecl::dt_wchar;
	case CAction::KW_BOOL: return TDecl::dt_bool;
	case CAction::KW_VOID: return TDecl::dt_void;
	case CAction::KW_FLOAT:	return TDecl::dt_float;
	case CAction::KW_DOUBLE: return TDecl::dt_double;
	case CAction::KW_ENUM: return TDecl::dt_enum;
	case CAction::ENUM_ID: return TDecl::dt_enumval;
	case CAction::KW_STRUCT: return TDecl::dt_struct;
	case CAction::KW_CLASS:	return TDecl::dt_class;
	case CAction::KW_UNION:	return TDecl::dt_union;
	case CAction::KW_NAMESPACE: return TDecl::dt_namespace;
	case CAction::KW_DTYPE:	return TDecl::dt_dtype;
	case CAction::KW_DCLASS: return TDecl::dt_dclass;
	default: return TDecl::dt_default;
	}
}

/*
static Token
tokenof(TDecl::DataType dt)
{
	switch(dt) {
	case TDecl::dt_int: return CAction::KW_INT;
	case TDecl::dt_char: return CAction::KW_CHAR;
	case TDecl::dt_wchar: return CAction::KW_WCHAR;
	case TDecl::dt_bool: return CAction::KW_BOOL;
	case TDecl::dt_void: return CAction::KW_VOID;
	case TDecl::dt_float: return CAction::KW_FLOAT;
	case TDecl::dt_double: return CAction::KW_DOUBLE;
	case TDecl::dt_enum: return CAction::KW_ENUM;
	case TDecl::dt_struct: return CAction::KW_STRUCT;
	case TDecl::dt_class: return CAction::KW_CLASS;
	case TDecl::dt_union: return CAction::KW_UNION;
	case TDecl::dt_namespace: return CAction::KW_NAMESPACE;
	case TDecl::dt_dtype: return CAction::KW_DTYPE;
	case TDecl::dt_dclass: return CAction::KW_DCLASS;
	default: return CAction::NONE;
	}
}
*/

type_t
CAction::rtypeof(Item_ val)
{
	switch(val.token) {
	case LITERAL:
		return (type_t)val.cptr;
	case VAL_ID:
		return uValId(val).decl->rtype;
	case EXPR:
		return uExpr(val).rtype();
	default:
		assertfail(val.token == LITERAL|VAL_ID|EXPR);
		for(;;);
	}
}


static bool
can_deref(type_t type)
{
	if(type->is_type(Type_PtrTo) || type->is_type(Type_Array))
		return true;
	return false;
}

type_t
CAction::deref(type_t type)
{
	if(PtrTo *ptr = dcast_PtrTo(type)) {
		return ptr->rtype();
	} else if(Array *arr = dcast_Array(type)) {
		return arr->rtype();
	} else {
		sz buf[MAXTYPELEN];
		buf[0] = '\0';
		error("cannot dereference `%s'", type_string(type, buf));
		return 0;
	}
}

Expr
CAction::exprof(Item_ val)
{
	Item v;
	switch(val.token) {
	case LITERAL:
	case VAL_ID:
		if(makeconst(&v, val)) {
			type_t t = (type_t)v.cptr;
			return Expr(Lit(t, v.val)).set_rtype(realtype(t));
		} else 
			return Expr(uValId(val).decl);
	case EXPRLIST:
	case EXPR:
	case STMTLIST:
		return uExpr(val);
	case TOKENLIST:
		return Expr(val.TokenList);
	case NONE:
		return null_expr;
	case ID:
		return Expr(val.id);
	default:
		assertfail(val.token == EXPRLIST|LITERAL|NONE);
		for(;;);
	}
}

Stmt
CAction::stmtof(Item_ val)
{
	switch(val.token) {
	case STMTLIST:
	case STMT:
	case EXPRLIST:
	case EXPR:
		return uStmt(val);
	case DECL: 
		return Stmt(ExprDecl(uDecl(val).decl));
	case NONE:
		return null_stmt;
	default:
		INTERNAL_ERROR;
	}
}

StmtList
CAction::stmtlistof(Item_ val)
{
	switch(val.token) {
	case STMTLIST:
		return uStmtList(val);
	default:
		INTERNAL_ERROR;
	}
}



Id
CAction::new_anon(Scope, csz *tname)
{
	sz buf[80];
	sprintf(buf, "__anon_%s_%d__", tname, _anon++);
	return strid(buf);
}


csz*
CAction::printable_scope_string(Scope sc)
{
	static sz buf[MAXIDLEN];
	sz *s;
	if(sc.parent()) {
		s = scoped_string(sc, buf);
		strpcat(buf, "scope '");
		strcat(buf, "'");
	} else {
		s = "global scope";
	}
	return s;
}


void
CAction::error_id(csz *what, Id id, Scope sc, bool showtype)
{
	sz buf[MAXIDLEN];
	Decl *decl = sc.get(id);
	if(decl && decl->tagged() && !showtype)
		sprintf(buf, "%s %s", 
			Decl::dtname[decl->dtype], cstr(id));
	else
		strcpy(buf, cstr(id));
	if(showtype && decl && decl->ptype) {
		csz *s;
		sz buf2[MAXTYPELEN];
		buf2[0] = '\0';
		if(decl->tagged())
			s = Decl::dtname[decl->dtype];
		else
			s = type_string(decl->ptype, buf2);
		error("'%s' %s as `%s' in %s",
			 buf, what, s, printable_scope_string(sc));
	} else  
		error("'%s' %s in %s",
			 buf, what, printable_scope_string(sc));
}


Scope
CAction::newscope(Scope parent, Decl *decl)
{
	Scope nsc = new SymTab(parent, decl);
	_allscopes += nsc;
	return nsc;
}

Scope
CAction::newaggrscope(Decl *decl)
{
	Aggr *aggr = dcast_Aggr(decl->ptype);
	assert(aggr);
	if(aggr->scope()) {
		error_dupdef(decl->id, decl->scope);
		return aggr->scope();
	} else {
		Scope nsc = newscope(decl->scope, decl);
		aggr->set_scope(nsc);
		return nsc;
	}
}

Scope
CAction::getaggrscope(Decl *decl)
{
	Aggr *aggr = dcast_Aggr(decl->ptype);
	assert(aggr);
	return aggr->scope();
}

vSymTab*
CAction::get_symtab(Decl *dsc) 
{
	vSymTab *symtab = 0;
	if(!dsc || !dsc->rtype)
		return 0;
	if(dsc->dtaggr()) {
		if(!(symtab = getaggrscope(dsc))) {
			error("aggregate `%s' is undefined", cstr(dsc->id));
			symtab = newscope(dsc->scope, dsc);
		}
	}
	return symtab;
}

vSymTab*
CAction::get_symtab(Item_ scope) 
{
	vSymTab *symtab;
	switch(scope.token) {
	case NONE:
		symtab = &_global_scope;
		break;
	case AGGR_ID:
		symtab = get_symtab(uAggrId(scope).decl); 
		break; 
	case NSPACE_ID: case ASPACE_ID:
		symtab = 0;
		break;
	case SCOPE:
		symtab = scope.scope;
		break;
	default:
		assert(scope.token == NONE|SCOPE|AGGR_ID);
		return 0;
	}
	assert(symtab);
	return symtab;
}

vSymTab*
CAction::get_symtab(Item_ sym, Item_ scope) 
{
	vSymTab *symtab = sym.scope;
	if(!symtab)
		symtab = get_symtab(scope);
	return symtab;
}


struct TypeList {
	type_t type;
	TypeList *next;
};

static type_t
addtype(TypeList **tl, type_t type)
{
	TypeList *tn = new TypeList;
	tn->type = type;
	tn->next = *tl;
	*tl = tn;
	return type;
}

type_t
type_ptrto(type_t t, bool is_const = false, bool is_volatile = false)
{
	static TypeList *tl = 0;
	for(TypeList *tp = tl; tp; tp = tp->next) {
		PtrTo *ptr = (PtrTo*)tp->type->ptr();
		if(ptr->rtype() == t && ptr->is_const() == is_const
				&& ptr->is_volatile() == is_volatile) 
			return tp->type;
	}
	return addtype(&tl, new obj_PtrTo(t, is_const, is_volatile));
}

type_t
type_refto(type_t t, bool is_const = false, bool is_volatile = false)
{ 
	static TypeList *tl = 0;
	for(TypeList *tp = tl; tp; tp = tp->next) {
		RefTo *ref = (RefTo*)tp->type->ptr();
		if(ref->rtype() == t && ref->is_const() == is_const
				&& ref->is_volatile() == is_volatile) 
			return tp->type;
	}
	return addtype(&tl, new obj_RefTo(t, is_const, is_volatile));
}

type_t
type_qualifier(type_t t, bool is_const, bool is_volatile)
{ 
	static TypeList *tl = 0;
	for(TypeList *tp = tl; tp; tp = tp->next) {
		Qualifier *qual = (Qualifier*)tp->type->ptr();
		if(qual->rtype() == t && qual->is_const() == is_const
				&& qual->is_volatile() == is_volatile) 
			return tp->type;
	}
	return addtype(&tl, new obj_Qualifier(t, is_const, is_volatile));
}


type_t
type_const(type_t t)
{
	return type_qualifier(t, true, false);
}

type_t
type_declref(Decl *d)
{
	static TypeList *tl = 0;
	for(TypeList *tp = tl; tp; tp = tp->next)
		if(((DeclRef*)tp->type->ptr())->decl() == d)
			return tp->type;
	return addtype(&tl, new obj_DeclRef(d));
}

type_t
type_array(type_t t, uint size, void *dsize = 0)
{ 
	static TypeList *tl = 0;
	for(TypeList *tp = tl; tp; tp = tp->next) {
		Array *array = (Array*)tp->type->ptr();
		if(array->rtype() == t && array->size() == size
				&& array->dsize() == dsize)
			return tp->type;
	}
	return addtype(&tl, new obj_Array(t, size, dsize));
}

type_t
type_aggr(Id id, Scope sc)
{ return new obj_Aggr(id, sc); }

type_t
type_enum(Decl *d)
{ return new obj_Enum(d); }

type_t
type_psig(uint np, const type_t *p, bool var = false)
{ 
	static TypeList *tl = 0;
	for(TypeList *tp = tl; tp; tp = tp->next) {
		Params *par = (Params*)tp->type->ptr();
		if(par->nparam() == np
		    && par->is_variable() == var) {
			for(uint i=0; i<np; i++) {
				if(par->ptype(i) != p[i])
					goto typeloop;
			}
			return tp->type;
		}
	    typeloop: ;
	}
	p = (type_t*)memcpy(new type_t[np], p, sizeof(type_t)*np);
	return addtype(&tl, new obj_Params(np, p, var));
}

type_t
type_func(type_t rt, type_t p, vSymTab *s)
{ 
	static TypeList *tl = 0;
	for(TypeList *tp = tl; tp; tp = tp->next) {
		Func *func = (Func*)tp->type->ptr();
		if(func->rtype() == rt && func->psig() == p
		    && func->scope() == s)
			return tp->type;
	}
	return addtype(&tl, new obj_Func(rt, p, s));
}

type_t
type_method(type_t o, type_t f)
{
	static TypeList *tl = 0;
	for(TypeList *tp = tl; tp; tp = tp->next) {
		Method *meth = (Method*)tp->type->ptr();
		if(meth->otype() == o && meth->ftype() == f)
			return tp->type;
	}
	return addtype(&tl, new obj_Method(o, f));
}

type_t
realtype(type_t ptype)
{
	bool is_const = false;
	bool is_volatile = false;
	type_t rt;
	for(;;) {
		if(Qualifier *qual = dcast_Qualifier(ptype)) {
			if(qual->is_const())
				is_const = true;
			if(qual->is_volatile())
				is_volatile = true;
			ptype = qual->rtype();
			continue;
		} else if(DeclRef *dref = dcast_DeclRef(ptype)) {
			ptype = dref->rtype();
			continue;
		}
		break;
	} 
	if(PtrTo *ptr = dcast_PtrTo(ptype)) {
		rt = type_ptrto(realtype(ptr->rtype()),
				ptr->is_const() || is_const,
				ptr->is_volatile() || is_volatile);
		is_const = is_volatile = false;
	} else if(RefTo *ref = dcast_RefTo(ptype)) {
		/* XXX - what about const/volatile? */
		rt = realtype(ref->rtype());
	} else if(Array *arr = dcast_Array(ptype)) {
		rt = type_array(realtype(arr->rtype()), arr->size(), arr->dsize());
	} else if(ptype == &Type_Str) {
		rt = type_ptrto(&Type_char);
	} else if(Method *meth = dcast_Method(ptype)) {
		type_t otype = meth->otype();
		while(Qualifier *q = dcast_Qualifier(otype)) {
			if(q->is_const())
				is_const = true;
			if(q->is_volatile())
				is_volatile = true;
			otype = q->rtype();
		}
		otype = realtype(otype);
		if(is_const || is_volatile)
			otype = type_qualifier(otype, is_const, is_volatile);
		rt = type_method(otype, realtype(meth->ftype()));
		is_const = is_volatile = false;
	} else if(Func *func = dcast_Func(ptype)) {
		rt = type_func(realtype(func->rtype()),
				realtype(func->psig()), func->scope());
	} else if(Params *par = dcast_Params(ptype)) {
		int np = par->nparam();
		type_t pt[np];
		for(int i=0; i<np; i++)
			pt[i] = realtype(par->ptype(i));
		rt = type_psig(np, pt, par->is_variable());
	} else
		rt = ptype;
	if(is_const || is_volatile)  {
		rt = type_qualifier(rt, is_const, is_volatile);
	}
	return rt;
}


int
CAction::type_sizeof(type_t type)
{
	if(type->is_type(Type_Func))
		return 0;
	else if(DeclRef *dref = dcast_DeclRef(type)) {
		Decl *decl = dref->decl();
		return type_sizeof(decl->rtype);
	} else if(type->is_type(Type_Aggr)) {
		// Aggr *aggr = (Aggr*)type->ptr();
		// XXX - need to loop
		return 0;
	} else if(Array *array = dcast_Array(type)) {
		return array->size()*type_sizeof(array->rtype());
	} else if(type->is_type(Type_PtrTo)) {
		return sizeof(void*);
	} else if(RefTo *ref = dcast_RefTo(type)) {
		return type_sizeof(ref->rtype());
	} else if(Integral *integral = dcast_Integral(type)) {
		return integral->size();
	} else if(type == &Type_bool) {
		return 1;
	} else if(type == &Type_void) {
		return 0;
	} else
		return 0; 
}

int
CAction::type_alignment(type_t type)
{
	if(type->is_type(Type_Func))
		return 0;
	else if(DeclRef *dref = dcast_DeclRef(type)) {
		Decl *decl = dref->decl();
		return type_alignment(decl->rtype);
	} else if(type->is_type(Type_Aggr)) {
		// Aggr *aggr = (Aggr*)type->ptr();
		// XXX - need to loop
		return 0;
	} else if(Array *array = dcast_Array(type)) {
		return type_alignment(array->rtype());
	} else if(type->is_type(Type_PtrTo)) {
		return sizeof(void*);
	} else if(RefTo *ref = dcast_RefTo(type)) {
		return type_alignment(ref->rtype());
	} else if(Integral *integral = dcast_Integral(type)) {
		return 1+(((1<<integral->alignment())-1)>>3);
	} else if(type == &Type_bool) {
		return 1;
	} else if(type == &Type_void) {
		return 0;
	} else
		return 0; 
}

Expr
CAction::dotexpr(type_t t, Item_ m, type_t *rt)
{
	bool is_const = false;
	bool is_volatile = false;
	if(Qualifier *qual = dcast_Qualifier(t)) {
		if(qual->is_const())
			is_const = true;
		if(qual->is_volatile())
			is_volatile = true;
		t = qual->rtype();
	}
	if(Aggr *aggr = dcast_Aggr(t)) {
		Scope scope = m.scope ? m.scope : aggr->scope();
		Decl *d = scope.ufind(m.id);
		if(!d) {
			sz buf[MAXIDLEN];
			error("`%s' is not a member of aggregate `%s'",
				cstr(m.id),
				scoped_string(scope.decl(), buf));
			return null_expr;
		}
		Expr e(d);
		if(is_const || is_volatile)
			e.set_rtype(realtype(type_qualifier(d->rtype,
				is_const,
				is_volatile)));
		if(rt)
			*rt = e.rtype();
		return e;
	} else {
		sz buf[MAXTYPELEN];
		buf[0] = '\0';
		error("`%s' is not an aggregate", type_string(t, buf));
		if(rt)
			*rt = 0;
		return null_expr;
	}
}

Expr
CAction::ptrexpr(type_t t, Item_ m, type_t *rt)
{
	if(!(t = deref(t))) 
		return null_expr;
	return dotexpr(t, m, rt);
}

type_t
CAction::decl_type(Item_ item, bool impwarn)
{
	if(item.token == NONE) {
		if(impwarn) {
			warning("implicit int");
			return &Type_int;
		} else
			return 0;
	}
	if(item.token == ENUM_ID) {
		INTERNAL_ERROR;
	}
	assert(item.token == TYPELIST);

	type_t t;

	TDecl &decl = *(TDecl*)&item.val;
	if(decl.ptype) {
		if(DeclRef *dr = dcast_DeclRef(decl.ptype)) 
			dr->decl()->anon = false;
		t = decl.ptype;
	} else {
	  if(decl.longcount && !decl.dtlongok()) 
		  error("long invalid for %s", TDecl::dtname[decl.dtype]);
	  if(decl.is_short && !decl.dtshortok())
		  error("short invalid for %s", TDecl::dtname[decl.dtype]);
	  if((decl.is_signed || decl.is_unsigned) && !decl.dtsignok())
		  error("signed/unsigned invalid for %s",
				  TDecl::dtname[decl.dtype]);
	  if(decl.longcount && decl.is_short)
		  error("long and short specified together");
	  if(decl.is_signed && decl.is_unsigned)
		  error("signed and unsigned specified together");
	  if(decl.longcount > 2)
		  error("long long long not supported");
	  switch(decl.dtype) {
	  case TDecl::dt_void: t = &Type_void; break;
	  case TDecl::dt_bool: t = &Type_bool; break;
	  case TDecl::dt_char: t = decl.is_signed ? &Type_schar
				  : decl.is_unsigned ? &Type_uchar
				  : &Type_char; break;
	  case TDecl::dt_wchar: t = &Type_wchar; break;
	  case TDecl::dt_float: t = &Type_float; break;
	  case TDecl::dt_double: t = &Type_double; break;
	  case TDecl::dt_default:
	  case TDecl::dt_int:
		  switch(decl.is_short ? -1 : (int)decl.longcount) {
		  case -1: t = decl.is_unsigned ? &Type_ushort : &Type_short;
			  break;
		  case 0: t = decl.is_unsigned ? &Type_uint : 
				  ((impwarn || decl.dtype == TDecl::dt_int)
				  ? &Type_int : (type_t)0);
			  break;
		  case 1:	t = decl.is_unsigned ? &Type_ulong : &Type_long;
			  break;
		  default: t = decl.is_unsigned ? &Type_ulong2 : &Type_long2;
			  break;
		  }
		  break;
	  default:
		  INTERNAL_ERROR;
	  }
	}
	if(decl.is_const || decl.is_volatile) 
		t = type_qualifier(t, decl.is_const, decl.is_volatile);
	if(_realtypes) 
		t = realtype(t);
	return t;
}

type_t
CAction::decl_type(Item *r, Item_ type, bool impwarn)
{
	if(type.token == NONE) 
		BuildTypeList(r, none, none, none);
	else
		*r = type;
	TDecl *td = (TDecl*)&r->val;
	td->ptype = decl_type(type, impwarn);
	td->is_const = false;
	td->is_volatile = false;
	return td->ptype;
}


/* need a "getconst" function that gets a constant value if possible ...
	CParse might need access to this too, so it should probably
	be a GetConst function */


bool
CAction::makeconst(Item *r, Item_ arg)
{
	switch(arg.token) {	
	case LITERAL:
		*r = arg;
		return true;
	case VAL_ID: {
		Decl *decl = uValId(arg).decl;
		if(decl && decl->storage == Decl::sc_constant) {
			r->token = LITERAL;
			r->cptr = decl->ptype;
			r->val = decl->c.val;
			return true;
		}
/*
		if(decl && decl->initialized && decl->ptype->is_type(Type_Qualifier)) {
			r->token = LITERAL;
			r->cptr = decl->ptype;
			r->val = decl->val;
			return true;
		}
*/
		} break; 
	default:
		r->token = LITERAL;
		r->cptr = &Type_int;
		r->val = 0;
		break;
	}
	return false;
}

void
CAction::Primary(Item *r, Item_ val, Item_ scope, Item_ )
{
	PRINT_ACTION(r, val, scope, none);
	Scope sc = get_symtab(scope);
	Decl *d;
	switch(val.token) {
	case NONE:
		*r = val;
		return;
	case VAL_ID:
		d = uValId(val).decl;
		break;
	default:
		INTERNAL_ERROR;
	}
	if(d->field) {
		Decl *dt = sc.ufind(_id_this);
		if(!dt) {
			sz buf[MAXIDLEN];
			error("cannot access field `%s' without a `this' pointer", 
				scoped_string(d, buf));
			*r = val;
			return;
		}
		Expr em = exprof(val);
		type_t rt = rtypeof(val);
		uExpr(r) = Expr(OP_PTRTO, Expr(dt), em).set_rtype(rt);
		r->token = EXPR;
	} else
		*r = val;
}

void
CAction::Literal(Item *r, Item_ val, Item_, Item_ )
{
	PRINT_ACTION(r, val, none, none);
	switch(val.token) {
	case NUM:
		switch(val.c.is_long) {
		case 0:
			r->cptr = val.c.is_signed ? &Type_int : &Type_uint;
			break;
		case 1:
			r->cptr = val.c.is_signed ? &Type_long : &Type_ulong;
			break;
		case 2:
			r->cptr = val.c.is_signed ? &Type_long2 : &Type_ulong2;
			break;
		default:
			error("%d L's too large", val.c.is_long);
		}
		r->val = val.Num;
		break;
	case CHAR:
		r->cptr = &Type_char;
		r->val = val.Char;
		break;
	case BOOL:
		r->cptr = &Type_bool;
		r->val = val.Bool;
		break;
	case STR:
		r->cptr = &Type_Str;
		r->val = (int)val.id;
		break;
	case FLOAT:		/* currently unsupported */
		*r = val;
		return;
	case VAL_ID:
	case TEMP:
		if(makeconst(r, val)) 
			return;
	case EXPR:
		error("not a constant");
		break;
	case LITERAL:
		*r = val;
		return;
	default:
		assertfail(val.token ==
				NUM|CHAR|STR|BOOL|FLOAT|VAL_ID|LITERAL);
	}
	r->token = LITERAL;
}

void
CAction::EnumNext(Item *r, Item_ prev, Item_, Item_ )
{
	PRINT_ACTION(r, prev, none, none);
	if(prev.token == NONE) {
		r->cptr = &Type_int;
		r->val = 0;
		r->token = LITERAL;
	} else {
		assert(prev.token == LITERAL);
		*r = prev;
		r->val = r->val+1;
	}
}

void
CAction::Undeclared(Item *r, Item_ sym, Item_ scope, Item_)
{
	PRINT_ACTION(r, sym, scope, none);
	vSymTab *symtab = get_symtab(sym, scope);
	error_notdecl(sym.id, Scope(symtab));
	*r = none;
}

csz*
CAction::opname(Token t)
{
	switch(t) {
	case OP_REF:	return "take the address of";
	case OP_PTRTO:
	case OP_PTR:	return "dereference";
	case OP_INC:	return "increment";
	case OP_DEC:	return "decrement";
	case OP_ASN:
	case OP_ASNADD: case OP_ASNSUB:
	case OP_ASNMUL: case OP_ASNDIV: case OP_ASNMOD:
	case OP_ASNAND: case OP_ASNOR: case OP_ASNXOR:
	case OP_ASNLSHIFT: case OP_ASNRSHIFT:
		return "assign";
	case OP_CONDASN:
		return "conditionally assign";
	default:
		INTERNAL_ERROR;
	}
}

void
CAction::UnaryOp(Item *r, Item_ op, Item_ arg, Item_ )
{
	PRINT_ACTION(r, op, arg, none);
	Item a;
	if(arg.token == NONE) {
		*r = none;
		return;
	}
	if(makeconst(&a, arg)) {
		switch(op.token) {
		case OP_UPLUS:	r->val = a.val; break;
		case OP_UMINUS:	r->val = -a.val; break;
		case OP_NOT:	r->val = ~a.val; break;
		case OP_LNOT:	r->val = !a.val; break;
		case OP_REF:
		case OP_INC:
		case OP_DEC:
			error("cannot %s a constant",
				opname(op.token));
			/* XXX - these imply assignment as well... */
		case OP_PTR: {
			type_t t = type_ptrto((type_t)a.cptr);
			uExpr(r) = Expr(op.token, exprof(arg))
						.set_rtype(t);
			r->token = EXPR;
			}
			return;
		default:
			assertfail(op.token == operator);
		}
		r->token = LITERAL;
		r->cptr = a.cptr;
	} else if(op.token == OP_TSIZEOF) {
		type_t type = decl_type(arg);
		if(type)
			r->val = type_sizeof(type);
		r->token = LITERAL;
		r->cptr = &Type_int;
	} else {
		type_t t = rtypeof(arg);
		switch(op.token) {
		case OP_REF: t = type_ptrto(t); break;
		case OP_PTR: t = deref(t); break;
		case OP_SIZEOF: t = &Type_int; break;
		case OP_LNOT: t = &Type_bool; break;
		default: ;
		}
		uExpr(r) = Expr(op.token, exprof(arg)).set_rtype(t);
		r->token = EXPR;
	}
}

void
CAction::BinaryOp(Item *r, Item_ op, Item_ arg1, Item_ arg2)
{
	PRINT_ACTION(r, op, arg1, arg2);
	Item a1, a2;
	if(arg1.token == NONE || arg2.token == NONE) {
		*r = none;
		return;
	}
	if(makeconst(&a1, arg1)) {
		switch(op.token) {
		case OP_ASN:
		case OP_ASNADD: case OP_ASNSUB:
		case OP_ASNMUL: case OP_ASNDIV: case OP_ASNMOD:
		case OP_ASNAND: case OP_ASNOR: case OP_ASNXOR:
		case OP_ASNLSHIFT: case OP_ASNRSHIFT:
			error("cannot %s a constant", opname(op.token));
		default: ;
		}
		if(makeconst(&a2, arg2)) {
			switch(op.token) {
			case OP_ADD:  r->val = a1.val + a2.val; break;
			case OP_SUB:  r->val = a1.val - a2.val; break;
			case OP_MUL:  r->val = a1.val * a2.val; break;
			case OP_DIV:  r->val = a1.val / a2.val; break;
			case OP_MOD:  r->val = a1.val % a2.val; break;
			case OP_AND:  r->val = a1.val & a2.val; break;
			case OP_OR:   r->val = a1.val | a2.val; break;
			case OP_XOR:  r->val = a1.val ^ a2.val; break;
			case OP_LAND: r->val = a1.val && a2.val; break;
			case OP_LOR:  r->val = a1.val || a2.val; break;
			case OP_LXOR: r->val=!(!a1.val && !a2.val); break;
			case OP_LSS:  r->val = a1.val < a2.val; break;
			case OP_GTR:  r->val = a1.val > a2.val; break;
			case OP_LEQ:  r->val = a1.val <= a2.val; break;
			case OP_GEQ:  r->val = a1.val >= a2.val; break;
			case OP_EQU:  r->val = a1.val == a2.val; break;
			case OP_NEQ:  r->val = a1.val != a2.val; break;
			case OP_LSHIFT:  r->val = a1.val << a2.val; break;
			case OP_RSHIFT:  r->val = a1.val >> a2.val; break;
			case OP_COMMA:	r->val = a2.val; break;
			default:
				assertfail(op.token == operator);
			}
			r->cptr = a1.cptr;	// XXX - need table
			r->token = LITERAL;
			return;
		}
	} 
	Expr e1 = exprof(arg1);
	type_t t1 = rtypeof(arg1);
	type_t rt;
	Expr e2 = null_expr;
	switch(op.token) {
	case OP_DOT:
		e2 = dotexpr(t1, arg2, &rt);
		break;
	case OP_PTRTO:
		e2 = ptrexpr(t1, arg2, &rt);
		break;
	case OP_MDOT:
	case OP_MPTRTO:
		{
		type_t pt;
		e2 = exprof(arg2);
		assert((pt = e2.rtype()));
		if(!(pt = deref(pt)) || !pt->is_type(Type_Method)) {
			error("not a member function pointer");
		} else {
			Method *meth = (Method*)pt->ptr();
			Func *func = (Func*)meth->ftype()->ptr();
			rt = func->rtype();
		}
		break; }
	case OP_INDEX:
		e2 = exprof(arg2);
		if(Array *arr = dcast_Array(t1)) {
			rt = arr->rtype();
		} else if(PtrTo *ptr = dcast_PtrTo(t1)) {
			rt = ptr->rtype();
		} else {
			error("not an array/ptr");
			rt = t1;
		}
		break;
	case OP_LSS: case OP_GTR: case OP_LEQ: case OP_GEQ:
	case OP_EQU: case OP_NEQ:
	case OP_LAND: case OP_LOR: case OP_LXOR:
		e2 = exprof(arg2);
		rt = &Type_bool;
		break;
	case OP_COMMA:
		e2 = exprof(arg2);
		rt = e2.rtype();
		break;
	default:
		e2 = exprof(arg2);
		rt = t1;
	}
	uExpr(r) = Expr(op.token, e1, e2).set_rtype(rt);
	r->token = EXPR;
}

void
CAction::OpNew(Item *r, Item_ type, Item_ init, Item_ params)
{
	PRINT_ACTION(r, type, init, params);

	assert(type.token == TYPELIST);
	type_t nt = decl_type(type);
	Expr e(OP_NEW, nt, exprof(init), exprof(params));
	e.set_rtype(type_ptrto(realtype(nt)));
	uExpr(r) = e;
	r->token = EXPR;
}

// this breaks type-safety, but C++
// gives us no choice since 0 == NULL
static bool
couldbenull(Expr e)
{
	if(e.op() == T(LITERAL)) {
		if(e.lit().type == &Type_int && e.lit().val == 0)
			return true;
	} 
	return false;
}

void
CAction::CondExpr(Item *r, Item_ cond, Item_ iftrue, Item_ iffalse)
{
	PRINT_ACTION(r, cond, iftrue, iffalse);
	Item c, t, f;
	if(makeconst(&c, cond) && makeconst(&t, iftrue)
				&& makeconst(&f, iffalse)) {
		r->val = c.val ? t.val : f.val;
		r->cptr = c.val ? t.cptr : f.cptr;
		r->token = LITERAL;
		return;
	}
	Expr eiftrue = exprof(iftrue);
	Expr eiffalse = exprof(iffalse);
	type_t rt = eiftrue.rtype();
	if(rt != eiffalse.rtype() && !typematch(rt, eiffalse.rtype())) {
		if(typematch(eiffalse.rtype(), rt))
			rt = eiffalse.rtype();
		else if(couldbenull(eiftrue) && can_deref(eiffalse.rtype()))
			rt = eiffalse.rtype();
		else if(!can_deref(rt) || !couldbenull(eiffalse))
			error("type mismatch in conditional");
	}
	uExpr(r) = Expr(QMARK, exprof(cond), eiftrue, eiffalse)
					.set_rtype(rt);
	r->token = EXPR;
}



long
CAction::check_repeat(long a, long b, csz *name)
{
	if(a && b) {
		if(a == b)
			error("repeated %s", name);
		else
			error("more than one %s", name);
	}
	return a ? a : b;
}

void
CAction::BuildTypeList(Item *r,
		Item_ declist, Item_ item, Item_ )
{
	PRINT_ACTION(r, declist, item, none);
	TDecl td;
	switch(declist.token) {
	case TYPELIST:
		td = uTypeList(declist);
		break;
	case NONE:
		td.clear();
		break;
	default:
		assertfail(declist.token == TYPELIST|NONE);
		break;
	}

	TDecl n;
	n.clear();

	switch(item.token) {
	case KW_AUTO:		n.storage = TDecl::sc_auto; break;
	case KW_STATIC:		n.storage = TDecl::sc_static; break;
	case KW_EXTERN:		n.storage = TDecl::sc_extern; break;
	case KW_REGISTER:	n.storage = TDecl::sc_register; break;
	case KW_MUTABLE:	n.storage = TDecl::sc_mutable; break;
	case KW_TYPEDEF:	n.storage = TDecl::sc_typedef; break;
	case KW_FRIEND:		n.storage = TDecl::sc_friend; break;
	case KW_INT:		n.dtype = TDecl::dt_int; break;
	case KW_CHAR:		n.dtype = TDecl::dt_char; break;
	case KW_WCHAR:		n.dtype = TDecl::dt_wchar; break;
	case KW_FLOAT: 		n.dtype = TDecl::dt_float; break;
	case KW_DOUBLE:		n.dtype = TDecl::dt_double; break;
	case KW_BOOL:		n.dtype = TDecl::dt_bool; break;
	case KW_VOID:		n.dtype = TDecl::dt_void; break;
	case KW_VOLATILE:	n.is_volatile = true; break;
	case KW_CONST:		n.is_const = true; break;
	case KW_SHORT:		n.is_short = true; break;
	case KW_SIGNED:		n.is_signed = true; break;
	case KW_UNSIGNED:	n.is_unsigned = true; break;
	case KW_LONG:		n.longcount++; break;
	case KW_INLINE:		n.is_inline = true; break;
	case KW_VIRTUAL:	n.is_virtual = true; break;
	case KW_EXPLICIT:	n.is_explicit = true; break;
	case_ANY_TYPE_ID: {
		Decl *d = uValId(item).decl;
		n.ptype = type_declref(d);
		td.useful = td.useful || d->newtype;
		d->newtype = false;
		break; }
	case TYPELIST:
		n = (TDecl&)item.val;
		break;
	case NONE:
		break;
	default:
		error("unsupported declaration");
	}
	td.storage = (TDecl::StorageClass)check_repeat(td.storage,
				n.storage, "storage class");
	td.dtype = (TDecl::DataType)check_repeat(td.dtype, n.dtype, "type");
	td.ptype = (type_t)check_repeat((long)td.ptype,
				(long)n.ptype, "type");
	if(td.dtype && td.ptype)
		error("multiple types (missing semicolon?)");
	td.is_volatile = check_repeat(td.is_volatile,
					n.is_volatile, "volatile");
	td.is_const = check_repeat(td.is_const,
					n.is_const, "const");
	td.is_unsigned = check_repeat(td.is_unsigned,
					n.is_unsigned, "unsigned");
	td.is_signed = check_repeat(td.is_signed,
					n.is_signed, "signed");
	td.is_short = check_repeat(td.is_short,
					n.is_short, "short");
	td.is_inline = check_repeat(td.is_inline,
					n.is_inline, "inline");
	td.is_virtual = check_repeat(td.is_virtual,
					n.is_virtual, "virtual");
	td.is_explicit = check_repeat(td.is_explicit,
					n.is_explicit, "explicit");
	td.longcount += n.longcount;

	uTypeList(r) = td;
	r->token = TYPELIST;
	r->scope = 0;
}


void
CAction::BuildParamList(Item *r,
		Item_ paramlist, Item_ item, Item_ scope)
{
	PRINT_ACTION(r, paramlist, item, scope);
	assert((paramlist.token == PARMLIST)
		| (paramlist.token == NONE));
	r->val = paramlist.val;
	ParamList &plist = (ParamList&)r->val;
	r->token = PARMLIST;
	Param *p;
	switch(item.token) {
	case PARM:
		p = uParm(item);
		if(p->type != &Type_void) {
			p->next = plist.head;
			plist.head = p;
			plist.nparam++;
		}
		break;
	case ELIPSIS:
		plist.variable = true;
		break;
	case NONE:
		break;
	default:
		assertfail(item.token == PARMLIST|ELIPSIS|NONE);
	}
	r->scope = get_symtab(scope);
}


void
CAction::DeclArray(Item *r, Item_ type, Item_ size,
			Item_ /*lower*/)
{
	PRINT_ACTION(r, type, size, none);
	TDecl *td = (TDecl*)&r->val;
	type_t t = decl_type(r, type);
	if(size.token == NONE) 
		td->ptype = type_array(t, NPOS);
	else {
		Item s;
		if(makeconst(&s, size))
			td->ptype = type_array(t, s.val);
		else
			td->ptype = type_array(t, NPOS, exprof(size));
	}
}



void
CAction::DeclPtr(Item *r, Item_ ptrtype, Item_ type, Item_ )
{
	PRINT_ACTION(r, ptrtype, type, none);
	TDecl *td = (TDecl*)&r->val;
	const TDecl &cv = (const TDecl&)ptrtype.val;
	type_t t = decl_type(r, type);
	switch(ptrtype.token) {
	case OP_MPTR: 
		t = type_method(Scope(ptrtype.scope).decl()->ptype, t);
	case OP_PTR:
		td->ptype = type_ptrto(t, cv.is_const, cv.is_volatile);
		break;
	case OP_REF:
		td->ptype = type_refto(t, cv.is_const, cv.is_volatile);
		break;
	default:
		assertfail(ptrtype.token == OP_REF|OP_PTR);
	}
}


void
CAction::FuncScope(Item *r, Item_ sym, Item_ scope, Item_ )
{
	PRINT_ACTION(r, sym, scope, none);
	assert(sym.token == VAL_ID);
	Decl *d = uValId(sym).decl;
	assert(d);
	Func *func;
	if(Method *meth = dcast_Method(d->rtype))
		func = dcast_Func(meth->ftype());
	else
		func = dcast_Func(d->rtype);
	if(!func) {
		error_notfunc(d->id, d->scope); 
		return;
	}
	vSymTab *symtab = func->scope();
	if(d->is_method)
		addthis(d);
	if(!symtab)
		symtab = get_symtab(scope);
	r->scope = symtab;
	r->token = SCOPE;
}


Decl *
CAction::addparam(int i, Scope scope, const Param &p)
{
	Decl *d = new Decl;
	d->clear();
	d->scope = scope;
	d->id = p.name;
	d->ptype = p.type;
	d->rtype = realtype(d->ptype);
	d->where = i;
	d->storage = Decl::sc_param;
	d->defined = true;
	if(p.defval.token != NONE) {
		d->initialized = true;	
		d->v.expr = exprof(p.defval);
	}
	decladd(d);
	return d;
}

Decl*
CAction::addthis(Decl *decl)
{
	Method *meth = dcast_Method(decl->rtype);
	if(!meth)
		abort("cannot add `this' to non-method");
	Func *func = dcast_Func(meth->ftype());
	assert(func);
	if(Decl *d = Scope(func->scope()).uget(_id_this))
		return d;
	Param p;
	p.type = type_ptrto(meth->otype());
	p.name = _id_this;
	p.defval.token = NONE;
	return addparam(-1, func->scope(), p);
}


void
CAction::DeclFunc(Item *r, Item_ type, Item_ paramlist, Item_ cvqual)
{
	PRINT_ACTION(r, type, paramlist, cvqual);
	assert(type.token == TYPELIST || type.token == NONE);
	assert(paramlist.token == PARMLIST);

	ParamList &plist = (ParamList&)paramlist.val;
	const TDecl &cv = (const TDecl&)cvqual.val;

	uint np = plist.nparam;	
	type_t ptype[np];
	Param *param = plist.head;
	Scope scope = newscope(paramlist.scope, 0);
	for(int i=np-1; i>=0; i--, param = param->next) {
		ptype[i] = param->type;
		if(param->name || param->defval.token != NONE)
			addparam(i, scope, *param);
	}
	type_t psig = type_psig(np, ptype, plist.variable);
	TDecl *td = (TDecl*)&r->val;
	type_t rt = decl_type(r, type, false);
	td->ptype = type_func(rt, psig, scope);
	if(cv.is_const || cv.is_volatile)
		td->ptype = type_qualifier(td->ptype, cv.is_const, cv.is_volatile);
}


/* C/C++ extension: (XXX - should be a switchable flag to control this) */
const bool anonaggr_byname = true;

// XXX - not r-safe (and could use some simplification as well,
// although at least by dumping all this here, many of the strange
// quirks of C++, such as the overloaded meaning of "static"
// get kept in this one spot rather than scattered all over)

void
CAction::DeclSym(Item *r, Item_ def, Item_ sym, Item_ scope)
{
	PRINT_ACTION(r, def, sym, scope);
	switch(sym.token) {
	case_ANY_ID:
	case NONE:
		break;
	default:
		error("unknown declaration type");
		return;
	}

	// enumerated values
	if(def.token == ENUM_ID) {
		Scope sc = get_symtab(sym, scope);
		Decl *d = newdecl(sym.id, sc, type_declref(uEnumId(def).decl),
				datatypeof(def.token), true);
		uValId(r).id = sym.id;
		uValId(r).decl = d;
		r->token = VAL_ID;
		return;
	}

	assert(def.token == TYPELIST || def.token == NONE);
	const TDecl &td = uTypeList(def);

	// to handle `friend classname;'
	if(sym.token == AGGR_ID && td.storage == TDecl::sc_friend) {
		Item tl;
		// oops, the parser used the last part of the type as
		// the declarator... we'd better remedy that now:
		BuildTypeList(&tl, def, sym, none);
		DeclSym(r, tl, none, scope);
		return;
	}

	// supposedly nothing to declare... (could be unnamed aggregates)
	if(sym.token == NONE) {
		if(DeclRef *dref = dcast_DeclRef(td.ptype)) {
			Scope sc(get_symtab(scope));
			assert(sc);
			Decl *d = dref->decl();
			bool was_anon = d->anon;
			bool is_anon = was_anon;

			// keep a special list of our friends...
			if(td.storage == TDecl::sc_friend) {
				Decl *d = newdecl(ID_FRIEND, sc,
					decl_type(def), td.dtype, false);
				d->storage = td.storage;
				return;
			}

			// it's a forward ref meant to be in this scope
			if(td.useful && d->rtype->is_type(Type_Aggr)) {
				forwref(d->dtype, d->id, sc);
				return;
			}

			// if an aggregate name is given by itself,
			// treat it like it was anonymous
			if(!is_anon && anonaggr_byname && !td.useful
					&& d->rtype->is_type(Type_Aggr)
					&& sc.decl()
					&& sc.decl()->ptype
					&& sc.decl()->ptype->is_type(Type_Aggr))
				is_anon = true;

			// for anonymous aggregates, create
			// an unnamed variable of that type
			if(is_anon) {
				Decl *decl = newdecl(ID_ANON, sc,
					decl_type(def), td.dtype, false);
				decl->storage = td.storage;
				decl->is_explicit = td.is_explicit;
				decl->is_virtual = td.is_virtual;
				decl->is_inline = td.is_inline;
				decl->field = true;
				d->anon = was_anon; /* restore d->anon */
				return;
			} 
		}

		if(!td.useful) 
			warning("useless declaration");
		return;
	}

	Scope sc = get_symtab(sym, scope);
	type_t ptype = decl_type(def);
	type_t rtype = realtype(ptype);
	bool is_const = false;
	bool is_volatile = false;
	while(Qualifier *qual = dcast_Qualifier(rtype)) {
		is_const = qual->is_const();
		is_volatile = qual->is_volatile();
		rtype = qual->rtype();
	}
	bool ismem = sc.decl() && sc.decl()->dtaggr();
	int np = 0;
	bool copyctor = false;
	if(Func *func = dcast_Func(rtype)) {
		np = func->nparam();
		if(sym.token == DTOR_ID) {
			if(uDtorId(sym).aggrid != sc.pchid())
				error("destructor does not match class name");
			if(func->rtype())
				error("destructor cannot have return type");
		} else if(sym.token == CTOR_ID) {
			if(func->rtype())
				error("constructor cannot have return type");
			if(np == 1) {
				type_t tr = func->ptype(0);
				while(Qualifier *q = dcast_Qualifier(tr))
					tr = q->rtype();
				if(tr == sc.decl()->rtype) 
					copyctor = true;
			}
		} else if(sym.token == CONV_ID) {
			if(func->rtype()) {
				warning("conversion operator has return type");
				if(func->rtype() != uConvId(sym).type)
					error("return type doesn't match operator");
			}
		} else {
			if(!func->rtype()) {
				warning("function implicitly returns int");
				func->change_rtype(&Type_int);
			}
		}
		if(ismem && td.storage != TDecl::sc_static) {
			type_t atype = sc.decl()->rtype;
			if(is_const || is_volatile) {
				atype = type_qualifier(atype, is_const, is_volatile);
				while(Qualifier *qual = dcast_Qualifier(ptype)) 
					ptype = qual->rtype();
			}
			ptype = type_method(atype, ptype);
		} else if(is_const || is_volatile)
			error("cannot apply const/volatile to %s function",
				ismem ? "static member" : "non-member");
	}

	if(td.storage == TDecl::sc_static && sym.scope)
		warning("`static' should not be used on "
					"explicitly scoped declarations");

	Decl *decl = newdecl(sym.id, sc, ptype, td.dtype, false);
	decl->storage = td.storage;
	if(ismem) {
		switch(td.storage) {
		case TDecl::sc_static:
			decl->storage = TDecl::sc_default;
			break;
		case TDecl::sc_default:
			if(!decl->is_func) {
				decl->field = true;
				if(uScope(scope).dtype == TDecl::dt_dimpl)
					error("fields are not allowed in implementation sections");
				else if(uScope(scope).dtype == TDecl::dt_dtype)
					error("fields are not allowed in dynamic types");
			}
			break;
		case TDecl::sc_typedef:
		case TDecl::sc_mutable:
		case TDecl::sc_friend:
			break;
		default:
			error("`%s' invalid in aggregate", TDecl::scname[td.storage]);
		}
	} else
		decl->storage = td.storage;

	decl->is_explicit = td.is_explicit;
	decl->is_virtual = td.is_virtual;
	decl->is_inline = td.is_inline;

	decl->access = uScope(scope).access;
	if(DeclRef *dr = dcast_DeclRef(decl->ptype)) 
		dr->decl()->anon = false;

	assert(decl->scope);

	rtype = decl->rtype;

	if(!decl->is_func) {
		if(decl->storage != TDecl::sc_extern) {
			Decl *dsc = sc.decl();
			if(decl->storage == TDecl::sc_typedef
					|| !dsc || !dsc->dtaggr())
				notesymdef(decl);
		}
		if(sym.token == DTOR_ID)
			error("destructor must be a function");
	}

	uValId(r).id = sym.id;
	uValId(r).decl = decl;
	r->token = decltoken(decl);
	if(!decl->odeclared) {
		if(sym.token == CONV_ID)
			addconv(sc, uConvId(sym).type, decl);
		else if(sym.token == CTOR_ID && np == 1 && !copyctor) 
			addctor(sc, decl);
		return;
	}

	decl->odeclared = false;

	Decl *odecl = decl->link;
	if(get_symtab(scope) != global())
		error_dupdecl(decl->id, decl->scope);
	if(odecl->defined) {
		if(decl->defined) {
			if(decl->storage != TDecl::sc_typedef
			   || decl->storage != odecl->storage
			   || decl->rtype != odecl->rtype)
				error_dupdef(sym.id, decl->scope);
		}
	} else {
		if(odecl->field && !decl->field)
			error("field redeclared as static");
		if(decl->defined) {
			notesymdef(odecl);
			if(odecl->storage == TDecl::sc_extern)
				odecl->storage = TDecl::sc_default;
		}
	}
	uValId(r).decl = odecl;
	delete decl;
}


void
CAction::DeclParam(Item *r, Item_ type, Item_ name, Item_)
{
	PRINT_ACTION(r, type, name, none);

	assert(type.token == TYPELIST || type.token == NONE);
	const TDecl &td = uTypeList(type);

	if(name.token == AGGR_ID && td.dtype == TDecl::dt_default
			&& !td.is_short && !td.longcount
			&& !td.is_signed && !td.is_unsigned) {
		Item tl;
		// oops, the parser used the last part of the type as
		// the declarator... we'd better remedy that now:
		BuildTypeList(&tl, type, name, none);
		DeclParam(r, tl, none, none);
		return;
	}

	Param *p = new Param;
	p->type = decl_type(type);
	p->name = (name.token == NONE) ? 0 : name.id;
	p->defval = NONE;
	p->next = 0;
	uParm(r) = p;
	r->token = PARM;
}	

void
CAction::SetAttr(Item *r, Item_ sym, Item_ attr, Item_ val)
{
	PRINT_ACTION(r, sym, attr, val);
	Decl *decl = uValId(sym).decl;
	switch(attr.token) {
	case COLON:		// bitfield
		if(!decl->field)
			error("cannot use non-aggregate-member as bitfield");
		decl->bitfield = true;
		assert(val.token == LITERAL);
		decl->b.bsize = val.val;
		break;
	case KW_INLINE:
		decl->is_inline = true;
		break;
	default:
		assertfail(attr.token == COLON|KW_INLINE);
	}
	*r = sym;
	
}

void
CAction::Define(Item *r, Item_ sym, Item_ val, Item_)
{
	PRINT_ACTION(r, sym, val, none);
	Decl *decl = uValId(sym).decl;
	if(decl->initialized) {
		if(decl->is_func)
			error_dupdef(decl->id, decl->scope);
		else
			error_dupinit(decl->id, decl->scope);
	}
	notesymdef(decl);
	decl->initialized = true;
	Item v;
	if(makeconst(&v, val)) {
		if(Qualifier *q = dcast_Qualifier(decl->rtype)) 
			if(q->is_const()) 
				decl->storage = Decl::sc_constant;
		if(decl->storage == Decl::sc_constant) {
			decl->field = false;
			decl->c.val = v.val;
			*r = sym;
			return;
		}
	}
	if(decl->field)
		error("cannot assign values to fields");
	if(decl->is_func) {
		switch(val.token) {
		case LITERAL:
			if(!decl->is_virtual)
				error("cannot initialize non-virtual method");
			if(val.val != 0)
				error("virtual method initializer is not 0");
			decl->v.expr = Expr(Lit(&Type_int, 0));
			break;
		case VAL_ID: {
			Func *f1;
			type_t o1 = 0;
			bool c1 = false;
			if(Method *meth = dcast_Method(decl->rtype)) {
				f1 = dcast_Func(meth->ftype());
				o1 = meth->otype();
				if(Qualifier *qual = dcast_Qualifier(o1))
					c1 = qual->is_const();
			} else
				f1 = dcast_Func(decl->rtype);
			Decl *d = uValId(val).decl;
			for(; d; d = d->nextfn()) {
				Func *f2;
				type_t o2 = 0;
				bool c2 = false;
				if(Method *meth = dcast_Method(d->rtype)) {
					f2 = dcast_Func(meth->ftype());
					o2 = meth->otype();
					if(Qualifier *qual = dcast_Qualifier(o2))
						c2 = qual->is_const();
				} else
					f2 = dcast_Func(d->rtype);
				if(f1->psig() == f2->psig() && f1->rtype() == f2->rtype()
							&& c1 == c2 && typematch(o2, o1))
					break;
			}
			if(!d) {
				d = uValId(val).decl;
				sz b1[MAXTYPELEN];
				sz b2[MAXTYPELEN];
				while(d && d->rtype != decl->rtype) {
					scoped_string(decl, b1);
					scoped_string(d, b2);
					error("cannot initialize `%s' to `%s'",
					   type_string(decl->rtype, b1),
					   type_string(d->rtype, b2));
					d = d->nextfn();
				}
				d = uValId(val).decl;
			}
			addthis(decl);
			decl->v.expr = Expr(d);
			} break;
		case TOKENLIST:
		case STMTLIST:
		case EXPRLIST:
		case EXPR:	/* XXX - need to check it */
			decl->v.expr = exprof(val);
			break;
		default:
			error("invalid initializer for function");
		}
	} else 
		decl->v.expr = exprof(val);
	*r = sym;
}

void
CAction::DefParam(Item *r, Item_ par, Item_ val, Item_ )
{
	PRINT_ACTION(r, par, val, none);
	assert(par.token == PARM);
	uParm(par)->defval = val;
	*r = par;
}


void
CAction::notesymdecl(Decl *decl)
{
	if(_clex.include_level() > _inclevel)
		decl->otype = Decl::OT_FROMHDR;
	else
		decl->otype = _otype;
	if(!_notedecl)
		return;
	sz buf[MAXIDLEN];
	printf("%s:%d: decl: %s\n",
		_clex.source(), _clex.line_num(), scoped_string(decl, buf));
}

void
CAction::notesymdef(Decl *decl)
{
	decl->defined = true;
	if(_clex.include_level() > 0)
		decl->otype = Decl::OT_FROMHDR;
	else
		decl->otype = _otype;
	if(!_notedef)
		return;
	sz buf[MAXIDLEN];
	printf("%s:%d: def: %s\n",
		_clex.source(), _clex.line_num(), scoped_string(decl, buf));
}

// create a new Decl, and store it 
// optionally report an error if it already exists
// otherwise return a link to it
Decl *
CAction::newdecl(Id id, Scope parent,
			TDecl::DataType dtype, bool err)
{
	if(id) {
		Scope ps; 
		Decl *d;
		if((ps = parent.parent()) && ((d = ps.uget(id))
				&& d->storage == Decl::sc_param))
			error_shadows(id, parent);
	}
		
	Decl *decl = new Decl;
	decl->clear();
	decl->access = TDecl::ac_public;
	decl->scope = parent;
	decl->id = id;
	decl->storage = Decl::sc_typedef;
	decl->dtype = dtype;
	if(!id) {
		id = new_anon(parent, TDecl::dtname[decl->dtype]);
		decl->anon = true;
	}
	decl->id = id;

	notesymdecl(decl);

	if(parent.add(id, decl)) 
		return decl;

	Decl *d = parent.get(id);
	decl->link = d;
	if(decl->tagged()) {
		while(d->link)
			d = d->link;
		if(!d->tagged() || d->converted) {
			d->link = decl;
			decl->link = 0;
			return decl;
		}
	} else {
		if(d->tagged() || id == ID_ANON) {
			parent.set(id, decl);
			return decl;
		}
	}
	decl->odeclared = true; 
	if(err)
		error_dupdecl(id, parent);
	return decl;
}

void
CAction::mergedefs(Func *to, Func *from)
{
	Scope sto = to->scope();
	Scope sfrom = from->scope();
	if(!sfrom)
		return;
	if(!sto) {
		to->change_scope(sfrom);
		return;
	}
	for(uint i=0; i<to->nparam(); i++) {
		Decl *dfrom = param_decl(from, i);
		Decl *dto = param_decl(to, i);
		if(dfrom) {
			if(!dto) {
				dto = new Decl;
				*dto = *dfrom;
				dto->scope = sto;
				decladd(dto);
			} else if(dfrom->initialized) {
				if(dto->initialized
				  && Expr(dto->v.expr) != Expr(dfrom->v.expr)) 
					error("default for parameter %d changed", i+1);
				/* XXX - should check for value clash */
				dto->initialized = true;
				dto->v.expr = dfrom->v.expr;
			}
		}
	}
}


Decl *
CAction::parentdecl(Decl *d)
{
	Decl *pd;
	if(d->dtype == TDecl::dt_enumval) {
		if(Enum *e = dcast_Enum(d->rtype)) {
			pd = e->decl();
		} else
			INTERNAL_ERROR;
	} else if((pd = d->scope.decl())) {
		if(d->scope != global()
		   && (!pd->ptype || !pd->ptype->is_type(Type_Aggr)))
			pd = 0;
	}
	return pd;
}

Decl *
CAction::parentaggrdecl(Decl *d)
{
	Decl *pd = d->scope.decl();
	if(!pd || !pd->ptype || !pd->ptype->is_type(Type_Aggr))
		pd = 0;
	return pd;
}

void
CAction::chaindecl(Decl *d)
{
	if(Decl *pd = parentdecl(d)) {
		Decl **dpp = &pd->a.first;
		while(*dpp)
			dpp = &(*dpp)->next;
		(*dpp) = d;
		if(!d->is_func) 
			pd->a.tsize++;
	}
}

void
CAction::unchaindecl(Decl *d)
{
	if(Decl *pd = parentdecl(d)) {
		Decl **dpp = &pd->a.first;
		while(*dpp && *dpp != d)
			dpp = &(*dpp)->next;
		assert(*dpp == d);
		*dpp = d->next;
		if(!d->is_func) 
			pd->a.tsize--;
		d->next = 0;
	}
}

Decl *
CAction::newdecl(Id id, Scope parent, type_t type,
			TDecl::DataType dtype, bool err)
{
	Decl *d = newdecl(id, parent, dtype, false);
	d->ptype = type;
	d->rtype = realtype(d->ptype);
	if(dtype == TDecl::dt_enumval)
		d->storage = TDecl::sc_constant;
	else
		d->storage = TDecl::sc_default;	
	Decl *od = d->link;
	Func *f1;
	bool c1 = false;
	type_t rtype = d->rtype;
	if(Method *meth = dcast_Method(rtype)) {
		f1 = dcast_Func(meth->ftype());
		if(Qualifier *q = dcast_Qualifier(meth->otype()))
			c1 = q->is_const();
		d->is_method = true;
	} else 
		f1 = dcast_Func(rtype);
	d->is_func = f1 != 0;
	if(d->odeclared && f1 && od->is_func) {
		for(Decl *dp = od; dp; dp = dp->nextfn()) {
			Func *f2;
			bool c2 = false;
			if(Method *meth = dcast_Method(dp->rtype)) {
			    f2 = dcast_Func(meth->ftype());
			    if(Qualifier *q = dcast_Qualifier(meth->otype()))
				    c2 = q->is_const();
			} else
			    f2 = dcast_Func(dp->rtype);
			if(f1->psig() == f2->psig() && c1 == c2) {
				// if it was originally declared as 'static',
				// then we need to convert it from a method back
				// to an ordinary function...
				if(d->is_method && !dp->is_method) {
					d->is_method = false;
					d->ptype = dcast_Method(d->ptype)->ftype();
					d->rtype = dcast_Method(d->rtype)->ftype();
				} else if(dp->is_method && !d->is_method)
					error("method redeclared as static function");

				if(f1->rtype() != f2->rtype()) {
					sz b1[MAXTYPELEN];
					sz b2[MAXTYPELEN];
					b1[0] = '\0';
					b2[0] = '\0';
					error("return type changed from `%s' to `%s'",
						type_string(f2->rtype(), b1),
						type_string(f1->rtype(), b2));
				}
				if(dp->defined) {
					mergedefs(f2, f1);
					d->ptype = dp->ptype;
					d->rtype = dp->rtype;
				} else {
					mergedefs(f1, f2);
					dp->ptype = d->ptype;
					dp->rtype = d->rtype;
				}
				d->link = dp;
				goto found;
			}
		}
		d->odeclared = false;
		d->scope.set(id, d);
	} else if(d->odeclared) {
		if(!idreal(id)) {
			d->odeclared = false;
			d->scope.set(id, d);
		} else if(d->rtype != od->rtype) {
			sz b1[MAXIDLEN];
			sz b2[MAXTYPELEN];
			sz b3[MAXTYPELEN];
			b2[0] = '\0';
			b3[0] = '\0';
			error("type of `%s' redeclared as %s (was `%s')",
				scoped_string(d, b1),
				type_string(d->rtype, b2),
				type_string(od->rtype, b3));
		}
	}
	found:
	if(!d->odeclared)
		chaindecl(d); 
	else if(err)
		error_dupdecl(id, parent);
	if(f1 && f1->scope()) 
		Scope(f1->scope()).decl(d->odeclared ? d->link : d);
	return d;
}


void
CAction::NewScope(Item *r,
	Item_ type, Item_ sym, Item_ scope)
{
	PRINT_ACTION(r, type, sym, scope);
	Scope sc = get_symtab(sym, scope);
	if(type.token == NONE) {
		uScope(r).clear();
		r->scope = newscope(sc, 0);
	} else {
		Decl *d = forwref(datatypeof(type.token), sym.id, sc);
		assert(d);
		d->access = uScope(scope).access;
		r->scope = newaggrscope(d);
		uScope(r).clear();
		uScope(r).dtype = d->dtype;
		uScope(r).ptype = d->ptype;
		uScope(r).access = d->defacc();
	}
	r->token = SCOPE;
}

void
CAction::EndAggr(Item *r,
	Item_ agsc, Item_ scope, Item_)
{
	PRINT_ACTION(r, agsc, scope, none);
	assert(agsc.token == SCOPE && uScope(agsc).dtaggr());
	Decl *d = Scope(agsc.scope).decl();
	assert(d);
	d->newtype = true;
	notesymdef(d);
	uAggrId(r).id = d->id;
	uAggrId(r).decl = d;
	r->token = AGGR_ID;
}

Decl *
CAction::forwref(TDecl::DataType dtype, Id id, Scope sc, bool search)
{
	if(!sc)
		sc = global();
	Decl *decl = id ? sc.tfind(id, true, search) : 0;
	while(decl && decl->converted)
		decl = decl->link;
	if(decl) {
		if(dtype != decl->dtype) {
			error("`%s %s' redeclared as %s",
				TDecl::dtname[decl->dtype],
				cstr(id),
				TDecl::dtname[dtype]);
		}
	} else if(dtype == TDecl::dt_enum) {
		sz buf[MAXIDLEN];
		decl = newdecl(id, sc, 0, dtype, true);
		decl->storage = Decl::sc_typedef;
		error("`%s' cannot be forward declared",
			scoped_string(decl, buf));
	} else {
		type_t newtype = type_aggr(id, 0);
		decl = newdecl(id, sc, newtype, dtype, true);
		decl->storage = Decl::sc_typedef;
	}
	return decl;
}

void
CAction::AggrRef(Item *r,
	Item_ type, Item_ sym, Item_ scope)
{
	PRINT_ACTION(r, type, sym, scope);
	if(sym.token == AGGR_ID) {
		uAggrId(sym).decl->newtype = true;
		*r = sym;
		return;
	}
	Decl *d = forwref(datatypeof(type.token), sym.id, global(), true);
	d->newtype = true;
	uAggrId(r).id = d->id;
	uAggrId(r).decl = d;
	r->scope = 0;
	r->token = decltoken(d);
}

void
CAction::AggrDecl(Item *r,
	Item_ type, Item_ sym, Item_ scope)
{
	PRINT_ACTION(r, type, sym, scope);
	Scope sc = get_symtab(sym, scope);
	Decl *d = forwref(datatypeof(type.token), sym.id, sc);
	d->access = uScope(scope).access;
	d->newtype = true;
	uAggrId(r).id = d->id;
	uAggrId(r).decl = d;
	r->scope = sym.scope;
	r->token = decltoken(d);
}

void
CAction::SetAccess(Item *r, Item_ scope, Item_ acc, Item_ )
{
	PRINT_ACTION(r, scope, acc, none);
	TDecl::Access a;
	assert(scope.token == SCOPE);
	Scope sc = Scope(scope.scope);
	if(uScope(scope).dtype == Decl::dt_dimpl)
		sc = sc.parent();
	Decl *dsc = sc.decl();
	assert(dsc && dsc->dtaggr());
	switch(acc.token) {
	case KW_PUBLIC:		a = TDecl::ac_public; break;
	case KW_PROTECTED:	a = TDecl::ac_protected; break;
	case KW_PRIVATE:	a = TDecl::ac_private; break;
	case AGGR_ID: {
		a = TDecl::ac_public;
		Decl *sdecl = sc.decl();
		Decl *decl = uAggrId(acc).decl;
		if(decl->dtype != Decl::dt_dtype)
			error("access/type specifier is not a dynamic type");
		else if(sdecl->dtype == Decl::dt_dtype
		 || sdecl->dtype == Decl::dt_dclass) {
			Scope nsc;
			Decl *di = forwref(Decl::dt_dimpl, decl->id, sc);
			Aggr *aggr = dcast_Aggr(di->rtype);
			assert(aggr);
			nsc = aggr->scope();
			if(!nsc) {
				nsc = newscope(sc, di);
				aggr->set_scope(nsc);
				di->i.decl = decl;
				inherit(nsc, dsc);
			}
			uScope(r).clear();
			uScope(r).dtype = di->dtype;
			uScope(r).ptype = di->ptype;
			r->scope = nsc;
			r->token = SCOPE;
			return;
		} else {
			sz buf[MAXIDLEN];
			error("dynamic access specifier `%s:'"
				" cannot be used in non-dynamic aggregate"
				" `%s'",
				cstr(decl->id), scoped_string(sdecl, buf)
				);
		}
		break; }
	default:
		assertfail(acc.token == PUBLIC|PROTECTED|PRIVATE|AGGR_ID);
		return;
	}
	uScope(r).clear();
	uScope(r).dtype = dsc->dtype;
	uScope(r).ptype = dsc->ptype;
	uScope(r).access = a; 
	r->scope = sc;
	r->token = SCOPE;
}


Decl*
CAction::addconv(Scope sc, type_t t, Decl *decl)
{
	// XXX - should create decl if it doesn't exist,
	// including creation of the string form of the operator name
	Decl *d = new Decl;
	d->clear();
	d->scope = sc;
	d->id = ID_CONV;
	d->ptype = t;
	d->rtype = realtype(t);
	d->i.decl = decl;
	if(!sc.add(ID_CONV, d)) {
		Decl *od = d->link = sc.get(ID_CONV);
		for(; od; od = od->link)
			if(od->rtype == d->rtype && od->id != d->id) {
				char b1[MAXTYPELEN+MAXIDLEN];
				char b2[MAXTYPELEN+MAXIDLEN];
				error("`%s()' converts the same type as `%s()'",
					scoped_string(decl, b1),
					scoped_string(od->i.decl, b2));
				break;	
			}
		if(!od)
			sc.set(ID_CONV, d);
	}
	return decl;
}

Decl*
CAction::addctor(Scope sc, Decl *decl)
{
	Decl *d = new Decl;
	d->clear();
	d->scope = sc;
	d->id = ID_CTOR;
	d->ptype = decl->ptype;
	d->rtype = decl->rtype;
	d->i.decl = decl;
	if(!sc.add(ID_CTOR, d)) {
		d->link = sc.get(ID_CTOR);
		sc.set(ID_CTOR, d);
	}
	return decl;
}

void
CAction::inherit(Scope sc, Decl *base, Decl::Access acc)
{
	Decl *d = new Decl;
	d->clear();
	d->access = acc;
	d->scope = get_symtab(base);
	d->id = ID_INHERIT;
	if(!sc.add(ID_INHERIT, d)) {
		d->link = sc.get(ID_INHERIT);
		sc.set(ID_INHERIT, d);
	}
	d->ptype = base->ptype;
	d->rtype = base->rtype;
	d->i.decl = base;
}

void
CAction::inherit(Decl *decl, Decl *base, Decl::Access acc)
{
	inherit(get_symtab(decl), base, acc);
}

void
CAction::Inherit(Item *r,
	Item_ scope, Item_ itype, Item_ base)
{
	PRINT_ACTION(r, scope, itype, base);
	assert(scope.token == SCOPE && uScope(scope).dtaggr());
	assert(base.token == AGGR_ID);
	Decl::Access acc;
	switch(itype.token) {
	case KW_PUBLIC:		acc = TDecl::ac_public; break;
	case KW_PROTECTED:	acc = TDecl::ac_protected; break;
	case KW_PRIVATE:	acc = TDecl::ac_private; break;
	default:
		acc = uScope(scope).defacc();
	}
	inherit(scope.scope, uAggrId(base).decl, acc);
	*r = scope;
}
 

void
CAction::EnumRef(Item *r,
	Item_ type, Item_ sym, Item_ scope)
{
	PRINT_ACTION(r, type, sym, scope);
	Scope sc = get_symtab(sym, scope);
	Decl *d = forwref(datatypeof(type.token), sym.id, sc, true);
	d->access = uScope(scope).access;
	d->newtype = true;
	uEnumId(r).id = d->id;
	uEnumId(r).decl = d;
	r->scope = sym.scope;
	r->token = decltoken(d);
}

void
CAction::NewEnum(Item *r,
	Item_ type, Item_ sym, Item_ scope)
{
	PRINT_ACTION(r, type, sym, scope);
	Scope parent = get_symtab(sym, scope);
	Decl *decl = newdecl(sym.id, parent, 0, datatypeof(type.token), true);
	decl->storage = Decl::sc_typedef;
	decl->access = uScope(scope).access;
	decl->rtype = decl->ptype = type_enum(decl);
	decl->newtype = true;
	decl->defined = true;
	uEnumId(r).id = decl->id;
	uEnumId(r).decl = decl;
	r->scope = sym.scope;
	r->token = ENUM_ID;
}


Token
CAction::decltoken(Decl *decl)
{
	if(decl->storage == TDecl::sc_typedef) {
		if(decl->dtaggr())
			return AGGR_ID;
		else if(decl->dtype == TDecl::dt_enum)
			return ENUM_ID; 
		else if(decl->dtype == TDecl::dt_namespace)
			return NSPACE_ID; 
		else
			return TYPE_ID;
	} else
		return VAL_ID;
}

void
CAction::FindSym(Item *r,
	Item_ type, Item_ sym, Item_ scope)
{
	PRINT_ACTION(r, type, sym, scope);
	Scope sc = get_symtab(sym, scope);
	r->scope = sym.scope;
	if(sym.token == NONE && type.token == SCOPE) {
		r->token = SCOPE;
		r->scope = global();
		return;
	}
	Decl *d = sc.tfind(sym.id, type.token != NONE, !sym.scope);
	if(d) {
		uValId(r).id = sym.id;
		uValId(r).decl = d;
		r->token = decltoken(d);
		if(type.token == SCOPE) 
			r->scope = get_symtab(*r);
	} else 
		*r = NONE;
	if((type.token == VAL_ID || type.token == AGGR_ID)
			&& r->token != type.token) {
		if(r->token == NONE)
			error_notdecl(sym.id, sc);
		else
			error("`%s' not valid", cstr(sym.id));
	}
	if(r->token == AGGR_ID) {
		if(scope.token == SCOPE && uScope(scope).dtaggr()) {
			if(sym.id == Scope(scope.scope).decl()->id)
				r->token = CTOR_ID;
		} else if(sym.scope) {
			if(sym.scope == getaggrscope(d))
				r->token = CTOR_ID;
		}
	} else if(r->token == VAL_ID && d->id == d->scope.pchid()) {
		uAggrId(r).decl = d->scope.parent().tget(sym.id);
		r->token = AGGR_ID;	/* convert ctors fn's back to aggr's */
	}
}

void
CAction::SpecialStr(Item *r,
	Item_ type, Item_ scope, Item_ )
{
	PRINT_ACTION(r, type, scope, none);
	sz buf[MAXTYPELEN];
	buf[0] = '\0';
	Decl *d = 0;
	Scope s = get_symtab(scope);
	while(s && !(d = s.decl()))
		s = s.parent();
	assert(d);
	switch(type.token) {
	case KW__FUNC:
		scoped_string(d, buf);
		break;
	case KW__PFUNC:
		scoped_string(d, buf);
		type_string(d->ptype, buf);
		break;
	default:
		assertfail(type.token != CPP_FUNC|CPP__PFUNC);
	}
	r->token = STR;
	r->id = strid(buf);
}

void
CAction::SpecialId(Item *r,
	Item_ type, Item_ val, Item_ scope)
{
	PRINT_ACTION(r, type, val, scope);
	sz buf[MAXTYPELEN];
	csz *n;
	switch(type.token) {
	case CONV_ID: {
		assert(val.token == TYPELIST);
		type_t t = decl_type(val);
		buf[0] = '\0';
		type_string(t, buf);
		strpcat(buf, "operator ");
		uConvId(r).type = t;
		uConvId(r).id = strid(buf);
		r->scope = val.scope;
		r->token = CONV_ID;
		} break;
	case OPER_ID:
		switch(val.token) {
		case OP_FUNC: n = "()"; break;
		case OP_INDEX: n = "[]"; break;
		case OP_NEWA: n = "new[]"; break;
		case OP_DELETEA: n = "delete[]"; break;
		default:
			n = cstr(val.token);
		}
		sprintf(buf, "operator%s%s", n[0] >= 'a' && n[0] <= 'z'
				? " " : "", n);
		uOperId(r).id = strid(buf);
		r->scope = val.scope;
		r->token = OPER_ID;
		break;
	case CTOR_ID: case DTOR_ID:
		sprintf(buf, "%c%s", type.token == DTOR_ID ? '~' : ' ', cstr(val.id));
		uCtorId(r).id = strid(buf);
		uCtorId(r).aggrid = val.id;
		r->scope = val.scope;
		r->token = type.token;
		break;
	case KW_THIS:
		uValId(r).id = _id_this;
		if(!(uValId(r).decl = Scope(get_symtab(scope)).ufind(_id_this))) {
			error("`this' used outside of"
				" non-static member function");
			r->token = NONE;
		} else
			r->token = VAL_ID;
		r->scope = val.scope;
		break;
	default:
		assertfail(type.token != CONV_ID|OPER_ID|CTOR_ID|DTOR_ID|KW_THIS);
	}
}

void
CAction::BuildArray(Item *r, Item_ array, Item_ item, Item_)
{
	PRINT_ACTION(r, array, item, none);
	switch(array.token) {
	case EXPRLIST:
		*r = array;
		break;
	case NONE:
		uExprList(r) = ExprList(ARRAY);
		r->token = EXPRLIST;
		break;
	default:
		assertfail(array.token == EXPRLIST|NONE);
		break;
	}
	if(item.token != NONE)
		uExprList(r) += exprof(item);
}

void
CAction::BuildExprList(Item *r, Item_ explist, Item_ item, Item_)
{
	PRINT_ACTION(r, explist, item, none);
	switch(explist.token) {
	case EXPRLIST:
		*r = explist;
		break;
	case NONE:
		uExprList(r) = ExprList(EXPRLIST);
		r->token = EXPRLIST;
		break;
	default:
		assertfail(explist.token == EXPRLIST|NONE);
		break;
	}
	if(item.token != NONE)
		uExprList(r) += exprof(item);
}

void
CAction::BuildStmtList(Item *r, Item_ stmtlist, Item_ item, Item_ scope)
{
	PRINT_ACTION(r, stmtlist, item, none);
	Scope sc = get_symtab(scope);
	switch(stmtlist.token) {
	case STMTLIST:
		*r = stmtlist;
		break;
	case NONE:
		uStmtList(r) = StmtList(sc);
		r->token = STMTLIST;
		break;
	default:
		assertfail(stmtlist.token == STMTLIST|NONE);
		break;
	}
	if(item.token != NONE) {
		Stmt s = stmtof(item);
		uStmtList(r) += stmtof(item);
	}
}

void
CAction::IfElse(Item *r, Item_ cond, Item_ st, Item_ sf)
{
	PRINT_ACTION(r, cond, st, sf);
	uStmt(r) = StmtIfElse(exprof(cond), stmtof(st), stmtof(sf));
	r->token = STMT;
}

void
CAction::While(Item *r, Item_ sc, Item_ s, Item_ )
{
	PRINT_ACTION(r, sc, s, none);
	uStmt(r) = StmtWhile(exprof(sc), stmtof(s));
	r->token = STMT;
}

void
CAction::DoWhile(Item *r, Item_ , Item_ s, Item_ ec)
{
	PRINT_ACTION(r, none, s, ec);
	uStmt(r) = StmtDoWhile(stmtof(s), exprof(ec));
	r->token = STMT;
}

void
CAction::ForExpr(Item *r, Item_ i, Item_ c, Item_ n)
{
	PRINT_ACTION(r, i, c, n);
	uExpr(r) = ExprFor(stmtof(i), exprof(c), exprof(n));
	r->token = EXPR;
}

void
CAction::For(Item *r, Item_ fe, Item_ s, Item_ )
{
	PRINT_ACTION(r, fe, s, none);
	uStmt(r) = StmtFor(exprof(fe), stmtof(s));
	r->token = STMT;
}

void
CAction::Switch(Item *r, Item_ v, Item_ s, Item_ )
{
	PRINT_ACTION(r, v, s, none);
	uStmt(r) = StmtSwitch(exprof(v), stmtof(s));
	r->token = STMT;
}

void
CAction::Return(Item *r, Item_ v, Item_ , Item_ )
{
	PRINT_ACTION(r, v, none, none);
	uStmt(r) = StmtReturn(exprof(v));
	r->token = STMT;
}

void
CAction::Label(Item *r, Item_ type, Item_ name, Item_ stmt)
{
	PRINT_ACTION(r, type, name, stmt);
	switch(type.token) {
	case NONE: uStmt(r) = StmtLabel(exprof(name), stmtof(stmt)); break;
	case KW_CASE: uStmt(r) = StmtCase(exprof(name), stmtof(stmt)); break;
	case KW_DEFAULT: uStmt(r) = StmtDefault(stmtof(stmt)); break;
	default:
		INTERNAL_ERROR;
	}
	r->token = STMT;
}

void
CAction::CtorInit(Item *r, Item_ ct, Item_ st, Item_)
{
	PRINT_ACTION(r, ct, st, none);
	uExpr(r) = ExprCtor(stmtlistof(ct), stmtlistof(st));
	r->token = EXPR;
}


void
CAction::Cast(Item *r, Item_ arg, Item_ type, Item_ ctype)
{
	PRINT_ACTION(r, arg, type, none);
	Item a;
	if(makeconst(&a, arg)) {
		r->cptr = realtype(decl_type(type));
		r->token = LITERAL;
		r->val = a.val;
	} else {
		assert(type.token == TYPELIST);
		Expr ev = exprof(arg);
		Token t = ctype.token == NONE ? OP_CAST : ctype.token;
		type_t nt = decl_type(type);
		Expr e(t, nt, ev);
		e.set_rtype(realtype(nt));
		uExpr(r) = e;
		r->token = EXPR;
	}
}


int
CAction::_typematch(type_t to, type_t from)
{
	if(from == to)
		return 1;
	if(Qualifier *qfrom = dcast_Qualifier(from)) {
		if(Qualifier *qto = dcast_Qualifier(to)) {
			return _typematch(qto->rtype(), qfrom->rtype());
		}
	} else if(Qualifier *qto = dcast_Qualifier(to)) {
		return 1+_typematch(qto->rtype(), from);
	} else if(PtrTo *pfrom = dcast_PtrTo(from)) {
		if(PtrTo *pto = dcast_PtrTo(to)) 
			return _typematch(pto->rtype(), pfrom->rtype());
	} else if(Array *afrom = dcast_Array(from)) {
		if(Array *ato = dcast_Array(to)) {
			if(afrom->size() != ato->size())
				return 0;
			return _typematch(ato->rtype(), afrom->rtype());
		}
		if(PtrTo *pto = dcast_PtrTo(to)) 
			return 1+_typematch(pto->rtype(), afrom->rtype());
	} else if(Func *ffrom = dcast_Func(from)) {
/*	
		if(PtrTo *pto = dcast_PtrTo(to)) 
			return _typematch(pto->rtype(), from);
*/
		if(Func *fto = dcast_Func(to)) {
			int cr = _typematch(fto->rtype(), ffrom->rtype());
			int cp = _typematch(fto->psig(), ffrom->psig());
			if(cr && cp)
				return cr+cp;
		}
/*
	} else if(Params *pfrom = dcast_Params(from)) {
		if(Params *pto = dcast_Params(to)) {
			uint np = pfrom->nparam();
			if(np != pto->nparam() 
			  || pto->is_variable() != pfrom->is_variable())
				return 0;
			for(uint i=0; i<np; i++) {
				if(pto->ptype(i) != pfrom->ptype(i))
					return 0;
			}
			return 1;
		} 
*/
	} else if(Method *mfrom = dcast_Method(from)) {
		if(Method *mto = dcast_Method(to)) {
			if(mto->otype() != mfrom->otype())
				return 0;
			return _typematch(mto->ftype(), mfrom->ftype());
		}
	} else if(Aggr *afrom = dcast_Aggr(from)) {
		if(Aggr *ato = dcast_Aggr(to)) {
			Scope sc = afrom->scope();
			if(!sc)
				return 0;
			for(Decl *d = sc.get(ID_INHERIT); d; d = d->link)
				if(int r = _typematch(to, d->rtype))
					return 1+r;
			
		}
	}
	return 0;
}


int
CAction::typematchctors(type_t to, type_t from)
{
	if(Aggr *afrom = dcast_Aggr(from)) {
		Scope sc = afrom->scope();
		if(sc) {
			for(Decl *d = sc.get(ID_CONV); d; d = d->link) 
				if(int r = _typematch(to, d->rtype))
					return 2+r;
			for(Decl *d = sc.get(ID_INHERIT); d; d = d->link)
				if(int r = typematchctors(to, d->i.decl->rtype))
					return 3+r;
		}
	}	
	return 0;
}

int
CAction::typematch(type_t to, type_t from)
{
	if(from == to)
		return 1;
	while(Qualifier *qfrom = dcast_Qualifier(from)) 
		from = qfrom->rtype();
	while(Qualifier *qto = dcast_Qualifier(to)) 
		to = qto->rtype();
	if(Integral *ifrom = dcast_Integral(from)) {
		if(to == &Type_bool)
			return 3;
		if(Integral *ito = dcast_Integral(to)) 
			return 2;
	}
	if(Integral *ito = dcast_Integral(to)) {
		if(from == &Type_bool)
			return 3;
	}
	if(Func *ffrom = dcast_Func(from)) {
		if(PtrTo *pto = dcast_PtrTo(to))
			to = pto->rtype();
	}
	if(int r = _typematch(to, from))
		return r;
	if(PtrTo *pto = dcast_PtrTo(to)) {
		type_t ptto = pto->rtype();
		if(from == &Type_null)
			return 1;
		bool tis_const = false;
		bool tis_volatile = false;
		while(Qualifier *q = dcast_Qualifier(ptto)) {
			if(q->is_const())
				tis_const = true;
			if(q->is_volatile())
				tis_volatile = true;
			ptto = q->rtype();
		}
		if(ptto == &Type_void && from->is_type(Type_PtrTo)) {
			type_t ptfrom = dcast_PtrTo(from)->rtype();
			bool fis_const = false;
			bool fis_volatile = false;
			while(Qualifier *q = dcast_Qualifier(ptfrom)) {
				if(q->is_const())
					fis_const = true;
				if(q->is_volatile())
					fis_volatile = true;
				ptfrom = q->rtype();
			}
			if((!fis_const || tis_const)
			 && (!fis_volatile || tis_volatile))
				return 1;
		}
	}
	if(int r = typematchctors(to, from))
		return r;
	if(Aggr *ato = dcast_Aggr(to)) {
		Scope sc = ato->scope();
		if(sc) for(Decl *d = sc.get(ID_CTOR); d; d = d->link) {
			ExprList ea;
			ea += Expr(from);
			if(int r = fnmatch(d->i.decl->rtype, ea, false))
				return 3+r;
		}
	}
	if(Enum *efrom = dcast_Enum(from)) {
		if(Integral *ito = dcast_Integral(to)) 
			return (1<<14);
	}
	return 0;
}

int
CAction::fnmatch(type_t ft, ExprList arg, bool thisconst)
{
	int add = 0;
	bool isconst = false;
	Func *f;
	if(Method *meth = dcast_Method(ft)) {
		f = dcast_Func(meth->ftype());
		if(Qualifier *q = dcast_Qualifier(meth->otype())) {
			if(q->is_const())
				isconst = true;
		}
		if(thisconst && !isconst)
			return 0;
		if(isconst)
			add = 1;
	} else
		f = dcast_Func(ft);
	assert(f);
	Params *p = dcast_Params(f->psig());
	uint np = arg.len();
	uint c = 1;
	if(np < p->nparam()) {
		Decl *dp = param_decl(f, np);
		if(!dp || !dp->initialized)
			return 0;
		else
			c += (1<<15);	// pick this as a last resort
	} else if(np > p->nparam()) {
		if(p->is_variable()) {
			np = p->nparam();
			c += (1<<15);	// big enough to override all
					// non-elipsis matches
		} else
			return 0;
	} 
	Expr ep = arg.first();
	for(uint i=0; i<np; i++, ep = ep.next()) {
		uint cc = typematch(p->ptype(i), ep.rtype());
		// printf("p[%d]=%d\n", i, cc);
		if(!cc && couldbenull(ep)) {
			cc = typematch(p->ptype(i), &Type_null);
			c += 3;
		}
		if(!cc)
			return 0;
		c += cc;
	}
	return c*2+add;
}
 
Decl *
CAction::findfn(Decl *df, ExprList arg, bool thisconst)
{
	Decl *dm = 0;
	bool ambig = false;
	uint c = (uint)-1;
	Decl *d;
	for(d = df; d; d = d->nextfn()) {
		uint cc = fnmatch(d->rtype, arg, thisconst);
		if(cc && cc <= c) {
			if(cc == c)
				ambig = true;
			else {
				c = cc;
				dm = d;
				ambig = false;
			}
		}	
	}
	if(!dm) {
		sz buf[MAXTYPELEN];
		sz *bp = buf;
		scoped_string(df, buf);
		bp += strlen(bp);
		bp += sprintf(bp, "(");
		Expr ep = arg.first();
		while(ep) {
			type_string(ep.rtype(), bp);
			bp += strlen(bp);
			if((ep = ep.next()))
				bp += sprintf(bp, ", ");
		}
		bp += sprintf(bp, ")");
		if(thisconst)
			bp += sprintf(bp, " const");
		error("no match for call `%s'", buf);
		mprintf("", "must correspond to one of:");
		c = 0;
	}
	if(ambig) {
		error("ambiguous function call");
		mprintf("", "could be any of:");
		dm = 0;
	}
	if(!dm) {
		for(d = df; d; d = d->nextfn()) {
			sz buf[MAXTYPELEN];
			uint cc = fnmatch(d->rtype, arg, thisconst);
			if(cc == c || !c) {
				scoped_string(d, buf);
				type_string(realtype(d->ptype), buf);
				mprintf("", "  %s", buf);
			}
		}
		dm = df;
	}
	return dm;
}

void
CAction::Call(Item *r, Item_ f, Item_ a, Item_ scope)
{
	PRINT_ACTION(r, f, a, none);
	if(f.token == NONE)
		return;
	Expr ef;
	Decl *d;
	if(f.token == VAL_ID && ((d = uValId(f).decl)
			&& d->is_method && !d->is_ctor())) {
		Scope sc = get_symtab(scope);
		Decl *dt = sc.ufind(_id_this);
		if(!dt) {
			sz buf[MAXIDLEN];
			error("cannot call method `%s' without a `this' pointer", 
				scoped_string(d, buf));
			*r = f;
			return;
		}
		Expr em = exprof(f);
		type_t rt = rtypeof(f);
		ef = Expr(OP_PTRTO, Expr(dt), em).set_rtype(rt);
	} else
		ef = exprof(f);
	ExprList ela = ExprList(exprof(a));
	if(ef.op() == VAL_ID && ef.decl()->is_func) 
		ef = Expr(findfn(ef.decl(), ela));
	else if(ef.op() == OP_PTRTO || ef.op() == OP_DOT) {
		Expr emf = ef.right();
		assert(emf.op() == VAL_ID);
		bool isconst = false;
		if(Qualifier *q = dcast_Qualifier(ef.left().rtype()))
			if(q->is_const())
				isconst = true;
		emf = findfn(emf.decl(), ela, isconst);
		ef = Expr(ef.op(), ef.left(), emf);
		ef.set_rtype(emf.rtype());
	}

	ExprCall e(ef, ela);
	type_t ft = ef.rtype();
	type_t rt = ft;

	if(PtrTo *ptr = dcast_PtrTo(ft))
		ft = ptr->rtype();

	Func *f1;
	if(Method *meth = dcast_Method(ft))
		f1 = dcast_Func(meth->ftype());
	else
		f1 = dcast_Func(ft);

	if(f1) {
		rt = f1->rtype();
	} else {
		sz buf[MAXTYPELEN];
		buf[0] = '\0';
		error("`%s' is not a function", type_string(ft, buf));
	}
	e.set_rtype(realtype(rt));
	uExpr(r) = e;
	r->token = EXPR;
}

void
CAction::CtorCall(Item *r, Item_ cc, Item_ a, Item_ scope)
{
	PRINT_ACTION(r, cc, a, none);
	Item cf = cc;
	cf.token = VAL_ID;
	assert(a.token == EXPRLIST);
	ExprList al = ExprList(exprof(a));
	if(cc.token == VAL_ID) {
		uExpr(r) = ExprCall(Expr(uValId(cc).decl), al);
		r->token = EXPR;
		return;
	}
	if(cc.token == AGGR_ID) {
		Decl *d = uValId(cc).decl;
		Aggr *aggr = dcast_Aggr(d->rtype);
		assert(aggr);
		sz buf[MAXIDLEN];
		sprintf(buf, " %s", cstr(aggr->name()));
		Scope s = aggr->scope();
		if(!s)
			error("not yet defined");
		else if(!(uValId(&cf).decl = s.get(strid(buf))))
			error("no constructor");
		else {
			Call(r, cf, a, scope);
			uExpr(r) = ExprCall(Expr(uValId(cc).decl),
				ExprList(exprof(a))).set_rtype(d->rtype);
			r->token = EXPR;
		}
	} else if(al.len() == 1) {
		Item tl;
		Item v;
		BuildTypeList(&tl, none, cc, none);
		v.token = EXPR;
		uExpr(&v) = al.first();
		Cast(r, v, tl, none);
	} else
		error("no constructor");
}

void
CAction::Goto(Item *r, Item_ type, Item_ name, Item_ )
{
	PRINT_ACTION(r, type, name, none);
	switch(type.token) {
	case KW_BREAK: uStmt(r) = StmtBreak(); break;
	case KW_CONTINUE: uStmt(r) = StmtContinue(); break;
	case NONE: uStmt(r) = StmtGoto(exprof(name)); break;
	default:
		INTERNAL_ERROR;
	}
	r->token = STMT;
}


void
CAction::print_syms()
{
	ccout.print_decls(global());
	ccout.print_defs(global());
	if(_hdr)
		ccout.print_reqs(global());
}

