/*  $Id: CAction.hh,v 1.55 1998/08/22 03:35:09 shaggy Exp $
 *      Default C/C++ Parsing Actions
 *      
 *      relative path: .//lib/cparse
 *  
 *      Shag/OS: an Object-Oriented MicroKernel Operating System
 *       Libraries
 *        C/C++ Parsing classes
 *  
 *  Copyright (c) 1996,1997 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the Shag/OS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare Non-profit Public License (SNPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare Non-profit Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

//  $Log: CAction.hh,v $
//  Revision 1.55  1998/08/22 03:35:09  shaggy
//  header update
//
//  Revision 1.54  1998/08/17 06:38:15  shaggy
//  more fixes for forward refs; __attribute__ support in aggrs
//
//  Revision 1.53  1998/08/16 22:17:57  shaggy
//  improved header output
//
//  Revision 1.52  1998/08/13 04:26:34  shaggy
//  bug fixes and dynamic sized arrays
//
//  Revision 1.51  1998/08/12 15:18:31  shaggy
//  better typing; fixed some type coercions;
//  cleanup for #include's, and fixed some constructor call bugs
//
//  Revision 1.50  1998/08/11 02:05:21  shaggy
//  deco can finally parse itself
//
//  Revision 1.49  1998/08/09 03:57:33  shaggy
//  lots of work on output dependency checking
//
//  Revision 1.48  1998/08/03 15:00:07  shaggy
//  conversion operators, constructor conversions, default params
//
//  Revision 1.47  1998/08/03 00:12:55  shaggy
//  tons of work-- friends, class scope fixes, tons of bugfixes,
//  lots more cleanup, more DC++ work, etc
//
//  Revision 1.46  1998/08/02 04:29:39  shaggy
//  more fixes
//
//  Revision 1.45  1998/08/01 21:02:06  shaggy
//  much more cleanup; many bugfixes; DC++ progress
//
//  Revision 1.44  1998/08/01 14:56:45  shaggy
//  about to restructure deco
//
//  Revision 1.43  1998/07/31 23:44:24  shaggy
//  running out of time
//
//  Revision 1.42  1998/07/31 04:15:25  shaggy
//  need sleep
//
//  Revision 1.41  1998/07/30 04:40:29  shaggy
//  much more dtype and dclass work
//
//  Revision 1.40  1998/07/29 03:35:10  shaggy
//  dtype support nearly complete, initial dclass support
//
//  Revision 1.39  1998/07/28 19:55:25  shaggy
//  moved most of generic parsing from Deco into CAction
//
//  Revision 1.38  1998/07/28 16:43:16  shaggy
//  moved lots of Deco into generic CParse lib
//  CAction split into vCAction and CAction
//  printaction -> part of CAction
//  ccint -> CRep (C Representation)
//  ccout -> COut (C output)
//
//  Revision 1.37  1998/07/19 22:33:32  shaggy
//  part 2 of major overhaul (needed for type-safty and to start
//  being more consistent)
//
//  Revision 1.36  1998/07/19 06:46:44  shaggy
//  1st stage of Item cleanup
//
//  Revision 1.35  1998/07/11 19:02:56  shaggy
//  declarations in for() work now (even multiple decls)
//
//  Revision 1.34  1998/07/05 20:32:54  shaggy
//  tons of bugfixes, enhancements and parser improvements
//
//  Revision 1.33  1998/07/04 20:38:56  shaggy
//  lots of bug fixes; fn, aggr, etc improvements, new/delete
//
//  Revision 1.32  1998/06/29 14:14:15  shaggy
//  more DECO conversions, support for 'this', etc
//
//  Revision 1.31  1998/06/29 01:36:51  shaggy
//  fixed inline attr, output of pure virtuals, etc...
//
//  Revision 1.30  1998/06/02 14:50:39  shaggy
//  in the midst of more full parsing updates
//
//  Revision 1.29  1998/06/01 13:28:58  shaggy
//  switchable code save/parse
//
//  Revision 1.28  1998/05/26 05:02:52  shaggy
//  parameter scoping works correctly
//  most basic C/C++ statements now work
//
//  Revision 1.27  1998/05/26 01:39:45  shaggy
//  params almost working right
//
//  Revision 1.26  1998/05/25 23:28:41  shaggy
//  major rework of func/param info
//
//  Revision 1.25  1998/05/17 23:14:54  shaggy
//  expressions, expression lists, etc
//
//  Revision 1.24  1998/05/15 00:17:24  shaggy
//  ctor initializers work, and enums are fixed so declarations
//  in an enum can be immediately re-used
//
//  Revision 1.23  1998/05/14 18:02:03  shaggy
//  fixed constant handling, enums, function cv tags, etc
//
//  Revision 1.22  1998/05/14 05:08:39  shaggy
//  inheritence works and lots of bugs fixed
//
//  Revision 1.21  1998/05/13 00:17:40  shaggy
//  scoping works and enums fixed (mostly)
//
//  Revision 1.20  1998/05/12 15:53:20  shaggy
//  scoping and class tags almost work
//
//  Revision 1.19  1998/05/12 05:00:12  shaggy
//  more scope work
//
//  Revision 1.18  1998/05/10 18:53:39  shaggy
//  just about ready to start fixing scoping
//
//  Revision 1.17  1998/05/10 00:03:30  shaggy
//  string concatenation works, and comments are being
//  added to get ready to fix scoping
//
//  Revision 1.16  1998/05/04 03:09:18  shaggy
//  support for ctors, dtors, conversion operators, operator overloading, etc
//
//  Revision 1.15  1998/05/03 02:47:07  shaggy
//  added inheritence to parser
//
//  Revision 1.14  1998/04/26 14:08:56  shaggy
//  cleanup
//
//  Revision 1.13  1998/04/22 13:06:38  shaggy
//  better enum handling
//
//  Revision 1.12  1998/04/22 12:02:03  shaggy
//  Changed ID to Id to be consistent with naming conventions
//
//  Revision 1.11  1998/04/16 13:35:43  shaggy
//  access control
//
//  Revision 1.10  1998/04/15 13:39:43  shaggy
//  added types to constants
//
//  Revision 1.9  1998/04/12 03:19:59  shaggy
//  improvments for function output
//
//  Revision 1.8  1998/04/06 14:28:06  shaggy
//  got enums almost working
//
//  Revision 1.7  1998/02/08 19:07:05  shaggy
//  about to start splitting up everything
//
//  Revision 1.6  1998/01/12 04:12:48  shaggy
//  got dependencies working when re-generating code
//
//  Revision 1.5  1998/01/11 05:08:24  shaggy
//   about to switch to standard ref/ptr semantics
//
//  Revision 1.4  1997/12/09 16:43:19  shaggy
//  param parsing fixed
//
//  Revision 1.3  1996/10/19  21:46:46  shaggy
//  moved non-lex/parse tokens here
//
//  Revision 1.2  1996/10/19  16:42:38  shaggy
//  about to move around token declarations
//
//  Revision 1.1  1996/04/10  12:27:55  shaggy
//  Initial revision
//


#include <CLex>
#include <vStrTab>
#include <vSymTab>
#include <vCAction>
#include <CRep>
#include <COut>


#define INTERNAL_ERROR  { internal_error(__FILE__, __LINE__, __FUNCTION__); \
			  for(;;); }

struct Param {
	type_t		type;		// type of parameter
	vStrTab::Id	name;		// formal name of parameter
	vCAction::Item	defval;		// default value
	Param		*next;	
	void clear() { memset(this, 0, sizeof *this); }
};

struct ParamList {
	Param	*head;
	uint8	nparam;
	bool	variable:1;
};

#define GEN_UITEM(t, n) \
	inline t const & n(vCAction::Item_ i) \
		{ return (t const &)i.val; } \
	inline t& n(vCAction::Item *i) \
		{ return (t&)i->val; }

// Things we wish were declared in the union in vCAction::Item,
// and inline conversions to get to those things:
struct ItemDecl {
	vStrTab::Id	id;
	Decl		*decl;
};
GEN_UITEM(ItemDecl, uDecl)
GEN_UITEM(ItemDecl, uValId)
GEN_UITEM(ItemDecl, uAggrId)
GEN_UITEM(ItemDecl, uEnumId)
GEN_UITEM(Expr, uExpr)
GEN_UITEM(ExprList, uExprList)
GEN_UITEM(Stmt, uStmt)
GEN_UITEM(StmtList, uStmtList)
struct ItemConvId {
	vStrTab::Id	id;
	type_t		type;
};
GEN_UITEM(ItemConvId, uConvId)
struct ItemOperId {
	vStrTab::Id	id;
};
GEN_UITEM(ItemOperId, uOperId)
struct ItemCtorId {
	vStrTab::Id	id;
	vStrTab::Id	aggrid;
};
GEN_UITEM(ItemCtorId, uCtorId)
GEN_UITEM(ItemCtorId, uDtorId)
GEN_UITEM(Param*, uParm)
GEN_UITEM(TDecl, uTypeList)
GEN_UITEM(TDecl, uScope)




class CAction : public vCAction {
protected:
	CLex &_clex;
private:
	vStrTab &_strtab;
	vSymTab &_global_scope;
public:
	typedef vStrTab::Id Id;

	CAction(CLex &clp, vStrTab &strtab, vSymTab &globlsym);
	void accept();
	void eol();
	csz* cstr(Token token) const;
	csz* cstr(vStrTab::Id id) const;
	sz * cstr(Item_ item, sz *buf, size_t buflen,
				bool show_type = false) const;
	Id strid(Token, csz *s) { return _strtab.id(s); }
	Id strid(csz *s) { return _strtab.id(s); }

	void eprintf(csz *fmt, ...);
	void mprintf(csz *prefix, csz *fmt, ...);
	void vmprintf(csz *prefix,
				csz *fmt, va_list args);
	void _assertfail(csz *cond, csz *filename,
				int line);

	void error(csz *fmt, ...);
	void warning(csz *fmt, ...);
	void abort(csz *fmt, ...);
	void internal_error(csz *file, int line,
				csz *module = 0);
	
	void mprint_action(csz *prefix, csz *name,
			Item_ p1, Item_ p2, Item_ p3);
	void mprint_result(csz *prefix, Item_ r);
	void unsupported(csz *name, Item_ p1, Item_ p2, Item_ p3);

	// vSymTab* get_symtab(Item_ id, Item_ scope);

	action Test, Undeclared;
	action Index, DoWhile;
	action Break, Continue;
	action NewAggr, NewNspace, NewLocal;
	action FreeScope;

	virtual void abort_exit() = 0;
	virtual void internal_error_exit() = 0;

private:
	int _inclevel;
	bool _cxx, _dcxx, _hdr;
protected:
	int _anon;
	bool _realtypes;
	bool _warnerr;
	bool _nowarn;
	bool _nooutput;
	Id _id_this;
	ScopeList _allscopes;
	Decl::OutType _otype;
public:
	CCOut ccout;
	bool _notedecl, _notedef;

	// Id string_id(Token, csz *str);
	// void abort_exit();
	// void internal_error_exit();
	// void veprintf(csz *fmt, va_list args);

	void realtypes(bool tf=true) { _realtypes = tf; }
	void warnerr(bool tf=true) { _warnerr = tf; }
	void nowarnings(bool tf=true) { _nowarn = tf; }
	void nooutput(bool tf) { _nooutput = tf; }
	bool nooutput() { return _nooutput || errors() ||
			   (_warnerr && warnings()); }
	Id new_anon(Scope, csz *tname);

	// void print_type(const type_t type);
	// sz* type_string(const type_t type, sz *s);

	type_t decl_type(Item_ item, bool impwarn = true);
	type_t decl_type(Item *r, Item_ item, bool impwarn = true);
	int type_sizeof(type_t type);
	int type_alignment(type_t type);

	Scope global() const { return &_global_scope; }

	void print_syms();
	void dump_scopes()
		{ ::dump_scopes(_allscopes); }

	Decl *forwref(TDecl::DataType type, Id id, Scope = 0, bool search = false);
	static Token decltoken(Decl *decl);

	Scope newscope(Scope parent, Decl *);
	Scope newaggrscope(Decl *);
	Scope getaggrscope(Decl *);
	vSymTab* get_symtab(Decl *scope);
	vSymTab* get_symtab(Item_ scope);
	vSymTab* get_symtab(Item_ id, Item_ scope);

	action Literal, Primary;
	action UnaryOp, BinaryOp, CondExpr, OpNew;
	action DeclSym, DeclParam;
	action DeclPtr, DeclArray, DeclFunc;
	action SetAttr;
	action Define, DefParam;
	action BuildTypeList, BuildParamList;
	action AggrRef, AggrDecl, EnumRef;
	action NewScope, FuncScope, SetAccess, Inherit;
	action EndAggr;
	action NewEnum, EnumNext;
	action FindSym, SpecialId, SpecialStr;

	action BuildExprList, BuildStmtList, BuildArray;
	action Cast, Call, CtorCall, Goto;
	action CtorInit;

	action IfElse, While, For, ForExpr, Switch;
	action Label, Return;

	long check_repeat(long a, long b, csz *name);

	void parse_dcxx()
		{ _hdr = false; _dcxx = _cxx = true; _clex.parse_dcxx(); ccout.hdr(_hdr); }
	void parse_dhxx()
		{ _hdr = _dcxx = _cxx = true; _clex.parse_dcxx(); ccout.hdr(_hdr); }
	void parse_cxx()
		{ _hdr = _dcxx = false; _cxx = true; _clex.parse_cxx(); ccout.hdr(_hdr); }
	void parse_hxx()
		{ _hdr = _cxx = true; _dcxx = false; _clex.parse_cxx(); ccout.hdr(_hdr); }
	void parse_c()
		{ _hdr = _dcxx = _cxx = false; _clex.parse_c(); ccout.hdr(_hdr); }
	void parse_h()
		{ _hdr = true; _dcxx = _cxx = false; _clex.parse_c(); ccout.hdr(_hdr); }
	bool dcxx() const { return _dcxx; }
	bool cxx() const { return _cxx; }
	bool hdr() const { return _hdr; }

	Decl* newdecl(Id id, Scope parent,
			TDecl::DataType = TDecl::dt_default, bool err = true);
	Decl* newdecl(Id id, Scope parent, type_t type,
			TDecl::DataType = TDecl::dt_default, bool err = true);
	// Decl* newdecl_item(Item_ type, Item_ sym, Item_ scope, bool err);
	Decl* parentdecl(Decl *);
	Decl* parentaggrdecl(Decl *);
	void chaindecl(Decl *);
	void unchaindecl(Decl *);
	csz* printable_scope_string(Scope);
	void error_id(csz *what, Id id, Scope, bool showtype);
	void error_dupdecl(Id id, Scope s)
		{ error_id("already declared", id, s, true); }
	void error_dupdef(Id id, Scope s)
		{ error_id("already defined", id, s, true); }
	void error_dupinit(Id id, Scope s)
		{ error_id("already initialized", id, s, false); }
	void error_notdecl(Id id, Scope s)
		{ error_id("not declared", id, s, false); }
	void error_notdef(Id id, Scope s)
		{ error_id("not defined", id, s, false); }
	void error_notinit(Id id, Scope s)
		{ error_id("not initialized", id, s, false); }
	void error_notfunc(Id id, Scope s)
		{ error_id("is not a function", id, s, false); }
	void error_shadows(Id id, Scope s)
		{ error_id("shadows parameter", id, s, false); }
	void error_noderef(Id id, Scope s)
		{ error_id("cannot be dereferenced", id, s, false); }
	void error_notaggr(Id id, Scope s)
		{ error_id("is not an aggregate", id, s, false); }
	csz* opname(Token t);
	bool makeconst(Item *r, Item_ arg);
	Expr exprof(Item_ item);
	Stmt stmtof(Item_ item);
	StmtList stmtlistof(Item_ item);
	Expr dotexpr(type_t t, Item_ m, type_t *rt = 0);
	Expr ptrexpr(type_t t, Item_ m, type_t *rt = 0);
	Decl* addconv(Scope, type_t, Decl*);
	Decl* addctor(Scope, Decl*);
	void inherit(Scope, Decl *base, Decl::Access acc = Decl::ac_public);
	void inherit(Decl *, Decl *base, Decl::Access acc = Decl::ac_public);

	void notesymdecl(Decl *decl);
	void notesymdef(Decl *decl);
	sz * scoped_string(Scope sc, sz *s)
		{ return ccout.scoped_string(sc, s); }
	sz * scoped_string(Decl *d, sz *s)
		{ return ccout.scoped_string(d, s, true); }
	sz * scoped_string_notag(Decl *d, sz *s)
		{ return ccout.scoped_string(d, s, false); }
	sz * type_string(type_t t, sz *s)
		{ return ccout.type_string(t, s); }
	Decl* addparam(int i, Scope, const Param &p);
	Decl* addthis(Decl *decl);
	Decl* getthis(Decl *decl) { return addthis(decl); }
	Decl *findfn(Decl *df, ExprList arg, bool thisconst = false);
	int _typematch(type_t to, type_t from);
	int typematch(type_t to, type_t from);
	int typematchctors(type_t to, type_t from);
	int fnmatch(type_t ft, ExprList arg, bool thisconst = false);

	type_t rtypeof(Item_ val);
	type_t deref(type_t type);

	void mergedefs(Func *to, Func *from);

	int inclevel() const { return _inclevel; }
	void inclevel(int i) { _inclevel = i; }
};

// XXX - probably ought to be static member fns:
type_t realtype(type_t ptype);
TDecl::DataType datatypeof(vCAction::Token tok);
type_t type_ptrto(type_t t, bool is_const = false, bool is_volatile = false);
type_t type_refto(type_t t, bool is_const = false, bool is_volatile = false);
type_t type_qualifier(type_t t, bool is_const, bool is_volatile);
type_t type_const(type_t t);
type_t type_declref(Decl *d);
type_t type_array(type_t t, uint size, void *dsize = 0);
type_t type_aggr(vStrTab::Id id, Scope sc);
type_t type_enum(Decl *d);
type_t type_psig(uint np, const type_t *p, bool var = false);
type_t type_func(type_t rt, type_t p, vSymTab *s);
type_t type_method(type_t o, type_t f);
type_t realtype(type_t ptype);


#ifdef CACTION_DEBUG

extern bool debugaction;

#define PRINT_ACTION(r,p1,p2,p3) \
	PrintAction _p_a_ (this, __FUNCTION__, r, p1, p2, p3)

class PrintAction {
	CAction *_caction;
	CAction::Item *_rval;
public:
	PrintAction(CAction *caction, csz *name,
		CAction::Item *rval,
		CAction::Item_ p1,
		CAction::Item_ p2,
		CAction::Item_ p3)
			: _caction(caction), _rval(rval)
		{ if(debugaction)
			_caction->mprint_action("", name, p1, p2, p3);
		}
	~PrintAction()
		{ if(debugaction)
			_caction->mprint_result("", *_rval);
		}
};

#else
#define PRINT_ACTION(r,p1,p2,p3) 	
#endif


// CAction Item types (XXX - majorly out of date)
//	(see vCAction.hh for CParse Item types
//
//	AnyId:		any identifier (might not be declared)
//		= DeclId | NewId
//
//	AnyDeclId:	any declared indentifier
//		= AnyTypeId | ValId | NspaceId | TplateId
//
//	AnyTypeId:	identifier for any type descriptor
//		= AggrId | EnumId | TypeId
//
//	AnyValId:	identifier for any symbol with a value
//		= ValId
//	    destructor:
//		token= DTOR_ID
//		id= string id (~classname)
//		tuple.right= original string id (without the ~)
//	    operator:
//		token= OPER_ID
//		id= string id (full operator name)
//	    conversion:
//		token= CONV_ID
//		id= string id (full type name)
//		tuple.right= type ptr
//
//	AnyScopeId:	identifier for any scope
//		= AggrId | NspaceId
//
//	AggrId:		identifier for an aggregate (class/struct/union)
//		token= AGGR_ID
//		id= string id
//		tuple.right = Decl*
//
//	EnumId:		identifier for an enumeration
//		token= ENUM_ID
//		id= string id
//		tuple.right = Decl*
//
//	TypeId:		identifier for a type (not enum/aggr)
//		token= TYPE_ID
//		id= string id
//		tuple.right = Decl*
//
//	ValId:		identifier for a symbol with a value
//		token= VAL_ID
//		id= string id
//		tuple.right = Decl*
//
//	NewId:		undefined identifier
//		token= ID
//		id= string id
//
//	NspaceId:
//		token= NSPACE_ID | ASPACE_ID
//		id= string id
//
//	TypeList:	compound list of identifiers to describe a type
//		token= [TYPELIST]
//
//	Lit:		literal (and synthesized literals, which
//				are essentially constants)
//		token= [LITERAL] | NONE
//
//	Expr:		compound expression
//
//

