/*  $Id: CLex.hh,v 1.22 1998/08/22 03:35:10 shaggy Exp $
 *      C/C++ Lexical Analysis
 *      
 *      relative path: .//lib/cparse
 *  
 *      Shag/OS: an Object-Oriented MicroKernel Operating System
 *       Libraries
 *        C/C++ Parsing classes
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the Shag/OS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare Non-profit Public License (SNPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare Non-profit Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

//  $Log: CLex.hh,v $
//  Revision 1.22  1998/08/22 03:35:10  shaggy
//  header update
//
//  Revision 1.21  1998/08/12 15:18:32  shaggy
//  better typing; fixed some type coercions;
//  cleanup for #include's, and fixed some constructor call bugs
//
//  Revision 1.20  1998/08/11 12:50:42  shaggy
//  getting conditional assignment to work
//
//  Revision 1.19  1998/07/30 16:16:29  shaggy
//  switched to new vsn of shacc and removed useless underscore prefixes
//  as well as lot of extra configuration code that is no longer needed
//
//  Revision 1.18  1998/07/25 21:24:59  shaggy
//  minor parsing fixes
//
//  Revision 1.17  1998/07/05 20:32:55  shaggy
//  tons of bugfixes, enhancements and parser improvements
//
//  Revision 1.16  1998/07/04 20:38:57  shaggy
//  lots of bug fixes; fn, aggr, etc improvements, new/delete
//
//  Revision 1.15  1998/06/29 14:14:16  shaggy
//  more DECO conversions, support for 'this', etc
//
//  Revision 1.14  1998/05/17 23:49:20  shaggy
//  can handle GNU-style __attribute__ clause (so we can parse the
//  standard headers)
//
//  Revision 1.13  1998/05/16 23:30:39  shaggy
//  added CTOR_ID and some rules to help fix ctor ambiguity problem
//  output turned off for #include'ed files
//  better error checking, and ability to list locations
//  of symbol declaration and definition
//
//  Revision 1.12  1998/05/14 18:02:03  shaggy
//  fixed constant handling, enums, function cv tags, etc
//
//  Revision 1.11  1998/05/10 00:03:30  shaggy
//  string concatenation works, and comments are being
//  added to get ready to fix scoping
//
//  Revision 1.10  1998/05/07 14:46:37  shaggy
//  const/volatile partially working
//
//  Revision 1.9  1998/04/26 14:08:57  shaggy
//  cleanup
//
//  Revision 1.8  1998/04/15 13:39:43  shaggy
//  added types to constants
//
//  Revision 1.7  1998/04/12 03:20:00  shaggy
//  improvments for function output
//
//  Revision 1.6  1998/01/12 04:12:48  shaggy
//  got dependencies working when re-generating code
//
//  Revision 1.5  1998/01/11 05:11:03  shaggy
//   about to switch to standard ref/ptr semantics
//
//  Revision 1.4  1996/11/08  22:22:54  shaggy
//  header update
//
//  Revision 1.3  1996/10/26  19:01:30  shaggy
//  header update
//
//  Revision 1.2  1996/10/26  18:52:00  shaggy
//  header update
//
//  Revision 1.1  1996/04/07  18:31:44  shaggy
//  Initial revision
//


#include <Lexer>


class CLex : public Lexer {
	enum State {
		S_NEXT, S_SKIP, S_SKIP_NEXT, S_SPACE, S_SPACE_EOL,
		S_EOL, S_NUM, S_BIN, S_OCT, S_HEX,
		S_NUM_START, S_NUM_END, S_FLOAT, S_ID, S_SQUOTE,
		S_DQUOTE, S_DQUOTE2, S_BQUOTE, S_AQUOTE, S_QUOTE_ESC,
		S_QUOTE_OCT, S_QUOTE_HEX, S_OP_ADD, S_OP_SUB, S_OP_MUL,
		S_OP_DIV, S_OP_MOD, S_OP_AND, S_OP_OR, S_OP_XOR,
		S_OP_LNOT, S_OP_ASN, S_OP_LSS, S_OP_GTR, S_OP_LSHIFT,
		S_OP_RSHIFT, S_COLON, S_DOT, S_DOT2, S_PTRTO,
		S_QMARK, S_CPP, S_NOCPP, S_CPP_NUM, S_CPP_NONUM,
		S_CPP_NAME, S_CPP_NONAME, S_CPP_OPTNUM,S_CPP_NOOPTNUM,S_CPP_CMD,
		S_CPP_ARG1, S_COMMENT, S_COMMENT2, S_EAT, S_EAT_STR,
		S_LINE, S_LINE_STR
	} state;
	enum State old_state;
	char *str;		// string to store tokens in
	char *strp;		// current pointer in 'str'
	size_t str_size;	// total size of 'str'
	char *buf;		// buffer to store translations in
	char *bufp;		// current pointer in 'buf'
	size_t buf_size;	// total size of 'buf'
	char *name;		// buffer to store source name in
	char *namep;		// current pointer in 'name'
	size_t name_size;	// total size of 'name'
	uint64 num;	// numeric value of last token
	double fnum;	// floating point value of last token
	size_t size;	// size of last token
	bool _ws;	// enable whitespace recognition
	bool _indent;	// enable indent whitespace recognition
	bool _kw;	// enable keyword lookup
	bool _cpp;	// return cpp tokens
	bool _cxx;	// parse C++ tokens and keywords
	bool _dcxx;	// parse DC++ tokens and keywords
	bool _eol;	// return EOL tokens
	bool _in_cpp;	// currently in the middle of a cpp line
	bool _qwrap;	// wrap quotes until complete
	bool _cpp_first; // first argument after '#'
	bool _first;	// first token in line
	bool _signed;	// last num was signed (unused)
	int _long;	// last num was long (unused)
public:
	enum Token {
		NONE = 0,
		EOF_COMPLETE,
		EOF_INCOMPLETE,
		INCOMPLETE, INVALID,

#include "CParse.tab.h"

		NUM_TOKENS,

		START_CPP = CPP_DEFINE,
		END_CPP = CPP__STDCVER,
		START_C = KW_ATTRIBUTE,
		END_C = KW__PFUNC,
		START_CXX = KW_OPERATOR,
		END_CXX = OP_DELETE,
		START_DCXX = KW_DYNAMIC,
		END_DCXX = KW_DYNAMIC
	};
private:
	static csz *token_table[];
	int _include_level;
	Token token;
public:
	CLex(char *buf, size_t str_size);
	CLex& set_buf(char *ptr, size_t size)
		{ buf = ptr; buf_size = size; return *this; }
	CLex& set_tok(char *ptr, size_t size)
		{ str = ptr; str_size = size; return *this; }
	CLex& set_name(char *ptr, size_t size)
		{ name = ptr; name_size = size; return *this; }
	CLex& parse_space(bool tf = true) { _ws = tf; return *this; }
	CLex& parse_indent(bool tf = true) { _indent = tf; return *this; }
	CLex& parse_keywords(bool tf = true) { _kw = tf; return *this; }
	CLex& parse_c() { _cxx = _dcxx = false; return *this; }
	CLex& parse_cpp(bool tf = true) { _cpp = tf; return *this; }
	CLex& parse_cxx() { _cxx = true; _dcxx = false; return *this; }
	CLex& parse_dcxx() { _dcxx = true; _cxx = true; return *this; }
	CLex& parse_eol(bool tf = true) { _eol = tf; return *this; }
	void get_line() { strp = str; state = S_LINE; }
	Token next_token();
	csz* token_string();
	csz* token_string(Token t) const;
	csz* token_string(Token t, bool tf) const;
	csz* token_string(Token t, uint64 val,
				bool is_signed = true, int is_long = 0);
	csz* token_string(Token t, csz *str, size_t len = NPOS);
	csz* token_spacing(Token prev, Token next) const;
	Token last_token() const { return token; }
	uint64 last_num() const { return num; }
	uint last_char() const;
	bool last_bool() const { return (bool)num; }
	bool last_signed() const { return _signed; }
	int last_long() const { return _long; }
	double last_float() const { return fnum; }
	csz* last_str() const { return str; }
	csz* last_id() const { return str; }
	size_t last_size() const { return size; }
	csz* source() const { return name; }
	void source(csz* str) { strcpy(name, str); }
	int include_level() { return _include_level; }
	bool in_string() const { return state == S_DQUOTE; }
	bool in_char() const { return state == S_SQUOTE; }
	bool in_bquote() const { return state == S_BQUOTE; }
	void set_token(Token t) { token = t; }
};

