/*  $Id: CParse.cc,v 1.33 1998/08/22 03:35:13 shaggy Exp $
 *      C/C++ Parser (yacc grammar)
 *      
 *      relative path: .//lib/cparse
 *  
 *      Shag/OS: an Object-Oriented MicroKernel Operating System
 *       Libraries
 *        C/C++ Parsing classes
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the Shag/OS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare Non-profit Public License (SNPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare Non-profit Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

/*  
 *  $Log: CParse.cc,v $
 *  Revision 1.33  1998/08/22 03:35:13  shaggy
 *  header update
 *
 *  Revision 1.32  1998/08/12 15:18:34  shaggy
 *  better typing; fixed some type coercions;
 *  cleanup for #include's, and fixed some constructor call bugs
 *
 *  Revision 1.31  1998/08/01 14:56:46  shaggy
 *  about to restructure deco
 *
 *  Revision 1.30  1998/07/30 16:16:29  shaggy
 *  switched to new vsn of shacc and removed useless underscore prefixes
 *  as well as lot of extra configuration code that is no longer needed
 *
 *  Revision 1.29  1998/07/28 16:43:18  shaggy
 *  moved lots of Deco into generic CParse lib
 *  CAction split into vCAction and CAction
 *  printaction -> part of CAction
 *  ccint -> CRep (C Representation)
 *  ccout -> COut (C output)
 *
 *  Revision 1.28  1998/07/22 14:52:32  shaggy
 *  some namespace support
 *
 *  Revision 1.27  1998/07/19 06:46:44  shaggy
 *  1st stage of Item cleanup
 *
 *  Revision 1.26  1998/07/11 19:02:57  shaggy
 *  declarations in for() work now (even multiple decls)
 *
 *  Revision 1.25  1998/07/10 13:45:05  shaggy
 *  working on decls in conditions
 *
 *  Revision 1.24  1998/07/09 14:34:35  shaggy
 *  first phase of fixes for decl's in if(), while(), etc
 *
 *  Revision 1.23  1998/07/07 13:41:39  shaggy
 *  enum cleanup, and can define types inside functions now
 *
 *  Revision 1.22  1998/07/05 20:32:55  shaggy
 *  tons of bugfixes, enhancements and parser improvements
 *
 *  Revision 1.21  1998/07/04 20:38:57  shaggy
 *  lots of bug fixes; fn, aggr, etc improvements, new/delete
 *
 *  Revision 1.20  1998/06/30 18:19:00  shaggy
 *  more fixes for member pointers
 *
 *  Revision 1.19  1998/06/29 01:36:52  shaggy
 *  fixed inline attr, output of pure virtuals, etc...
 *
 *  Revision 1.18  1998/06/05 03:34:17  shaggy
 *  parser fixes
 *
 *  Revision 1.17  1998/06/04 13:40:37  shaggy
 *  more fixes
 *
 *  Revision 1.16  1998/06/04 12:52:20  shaggy
 *  reparsing seems to work now
 *
 *  Revision 1.15  1998/06/03 13:39:12  shaggy
 *  function re-parsing almost works (for within classes)
 *
 *  Revision 1.14  1998/06/02 14:50:39  shaggy
 *  in the midst of more full parsing updates
 *
 *  Revision 1.13  1998/06/01 13:28:58  shaggy
 *  switchable code save/parse
 *
 *  Revision 1.12  1998/05/25 23:28:42  shaggy
 *  major rework of func/param info
 *
 *  Revision 1.11  1998/05/16 23:30:40  shaggy
 *  added CTOR_ID and some rules to help fix ctor ambiguity problem
 *  output turned off for #include'ed files
 *  better error checking, and ability to list locations
 *  of symbol declaration and definition
 *
 *  Revision 1.10  1998/05/15 07:03:35  shaggy
 *  dependency ordering fixed; constructor initializers; lots of bug fixes, etc
 *
 *  Revision 1.9  1998/05/14 05:08:39  shaggy
 *  inheritence works and lots of bugs fixed
 *
 *  Revision 1.8  1998/05/12 05:00:12  shaggy
 *  more scope work
 *
 *  Revision 1.7  1998/05/10 18:53:40  shaggy
 *  just about ready to start fixing scoping
 *
 *  Revision 1.6  1998/05/10 00:03:31  shaggy
 *  string concatenation works, and comments are being
 *  added to get ready to fix scoping
 *
 *  Revision 1.5  1998/04/15 13:39:44  shaggy
 *  added types to constants
 *
 *  Revision 1.4  1998/02/08 19:07:05  shaggy
 *  about to start splitting up everything
 *
 *  Revision 1.3  1998/01/12 04:12:48  shaggy
 *  got dependencies working when re-generating code
 *
 *  Revision 1.2  1998/01/11 05:06:33  shaggy
 *  about to switch to standard ref/ptr semantics
 *
 */


#define IMPLEMENTATION_CParse

#include <defines>
#include <inttypes.h>
#include <posix/stdarg.h>

#include "CParse.hh"

#define YYSTYPE vCAction::Item


// #define none vCAction::none


void
CParse::_internal_error(vErrorReport *er, csz *file, int line)
{ er->internal_error(file, line, "CParse"); }


CParse::CParse(CLex &clex, vCAction &caction) 
		: _clex(&clex), _caction(&caction)
{
	_prev_yychar = 0;
	_havenext = false;
	_current_scope = CLex::NONE;
	_stmtlist = CLex::NONE;
	_debug = false;
	_yydebug = false;
	_rawid = false;
	_gsavecode = _savecode = false;
	_deftok = 0;
	_saveline = 0;
}

void
CParse::inject(TokenList *tok)
{
	if(_toksrc)
		tok->append_delete(_toksrc);
	_toksrc = tok;
}

void
CParse::newdecl(Item *list, Item_ decl)
{
	if(list->token != T(NONE) && list->scope == _current_scope.scope) {
		Item stmt = decl;
		stmt.token = T(DECL);
		do3(*list, BuildStmtList, *list, stmt, _current_scope);
	}
} 


// XXX - some of this functionality probably needs to be offloaded
//	to CAction, to allow greater flexibility
//	since it's not truly part of the parser

int
CParse::_lex(YYSTYPE *yylval)
{
	while(true) {
		if(_havenext) {
			*yylval = _next_yylval;
			_havenext = false;
			return _next_yychar;
		} else if(_havenextnext) {
			*yylval = _nextnext_yylval;
			_havenextnext = false;
			return _nextnext_yychar;
		} 

		if(_toksrc) {
			if(!_saveline)
				_saveline = _clex->line_num();
			if(_toksrc->next(yylval)) {
				_clex->set_line_num(yylval->line);
				return yylval->token;
			} else {
				delete _toksrc;
				_toksrc = 0;
				_clex->set_line_num(_saveline);
				_saveline = 0;
			}
		}

		yylval->val = 0;
		vCAction::Token token = (vCAction::Token)_clex->next_token();
		yylval->token = token;
		switch(token) {
		case T(NONE):
			_caction->get_input(false); continue;
		case T(INCOMPLETE):
			_caction->get_input(true); continue;
		case T(EOF_COMPLETE):
			return 0;
		case T(EOF_INCOMPLETE):
			if(_clex->in_string())
				_caction->error("unterminated string");
			else if(_clex->in_char())
				_caction->error("unterminated character constant");
			else if(_clex->in_bquote())
				_caction->error("unterminated backquote");
			else
				_caction->error("incomplete final token");
			return 0;
		case T(INVALID):
			_caction->error("invalid character"); continue;
		case T(NUM):
			yylval->Num = _clex->last_num();
			yylval->c.is_signed = _clex->last_signed();
			yylval->c.is_long = _clex->last_long();
			break;
		case T(BOOL):
			yylval->Bool = _clex->last_bool(); break;
		case T(KW_TRUE):
			token = yylval->token = T(BOOL);
			yylval->Bool = true;
			break;
		case T(KW_FALSE):
			token = yylval->token = T(BOOL);
			yylval->Bool = false; 
			break;
		case T(STR):
		case T(BACKSTR):
		case T(ANGSTR):
		case T(ID):
			yylval->id = _caction->strid(
				token, _clex->last_str()); break;
		case T(CHAR):
			yylval->Char = _clex->last_char(); break;
		case T(FLOAT):
			yylval->Float = _clex->last_float(); break;
		case T(EOL):
			_caction->eol(); continue;
		case T(KW__FUNC):
		case T(KW__PFUNC):
			do2(*yylval, SpecialStr, *yylval, _current_scope);
			token = yylval->token;
			break;
		default: break;
		}
		yylval->line = _clex->line_num();
		// printf("sending yychar %d\n", (int)token);
		return (int)token;
	}
}


#define set_next_yy(c, l) \
	{ _next_yychar = (c); _next_yylval = (l); _havenext = true; }
#define set_nextnext_yy(c, l) \
	{ _nextnext_yychar = (c); _nextnext_yylval = (l); _havenextnext = true; }


// peekahead_lex() allows peaking ahead in the token stream
// for complex sequences.
// It handles scoping with the '::' operator,
// as well as looking up identifiers to find their types and
// aid the parser

int
CParse::_peekahead_lex(YYSTYPE *yylval)
{
	Item scope;
	bool is_dtor;
	int yychar = _lex(yylval);

	if(_rawid)
		goto ReturnToken;

	// printf("got %d,%d\n", yychar, T(KW_TEMPLATE));

	// We can't support templates yet...
	if(yychar == T(KW_TEMPLATE))
		_caction->abort("templates are not yet supported");

	// We can't support exception handling yet
	if(yychar == T(KW_THROW) || yychar == T(KW_CATCH)
			|| yychar == T(KW_TRY))
		_caction->abort("exception handling not yet supported");

	if(yychar == T(OP_NEW) || yychar == T(OP_DELETE)) {
		Item nyylval;
		int nyychar = _lex(&nyylval);
		if(nyychar == T(LBRACK)) {
			Item nnyylval;
			int nnyychar = _lex(&nnyylval);
			if(nnyychar == T(RBRACK)) {
				yylval->token = yylval->token == T(OP_NEW)
					? T(OP_NEWA) : T(OP_DELETEA);
				yychar = yylval->token;
				goto ReturnToken;
			}
			set_nextnext_yy(nnyychar, nnyylval);
		} 
		set_next_yy(nyychar, nyylval);
	}

	if((is_dtor = (yychar == T(OP_NOT)))) {
		yychar = _lex(yylval);
		if(yychar != T(ID)) {
			set_next_yy(yychar, *yylval);
			yychar = T(OP_NOT); *yylval = T(OP_NOT);
		}
	}

	scope.token = T(SCOPE);
	scope.scope = 0;
	if(yychar == T(OP_SCOPE)) {
		yychar = _lex(yylval);
		do3(scope, FindSym, T(SCOPE), T(NONE), none);
		if(yychar != T(ID)) {
			set_next_yy(yychar, *yylval);
			*yylval = scope;
			yychar = T(SCOPE);
		}
	}

	/* scope field of ID should be null unless a scope
 	 * is explicitly stated, in which case it should point
	 * to that symtab 
			
	 */
	if(yychar == T(ID)) {
		while((_next_yychar = _lex(&_next_yylval)) == T(OP_SCOPE)) {
			if(is_dtor) {
				set_nextnext_yy(_next_yychar, _next_yylval);
				set_next_yy(yychar, *yylval);
				yychar = T(OP_NOT); *yylval = T(OP_NOT);
				goto ReturnToken;
			}
			yylval->scope = scope.scope;
			do3(scope, FindSym, T(SCOPE), *yylval, _current_scope);
			yychar = _lex(yylval);
			if(scope.token == T(NONE)) {
				_caction->error("undefined scope");
				scope.scope = 0;
			}
			if(yychar != T(ID)) {
				set_next_yy(yychar, *yylval);
				*yylval = scope;
				yychar = T(SCOPE);
				goto ReturnToken;
			}
		}

		_havenext = true;

		Item def;
		Item type;

		if(_prev_yychar == T(KW_CLASS) || _prev_yychar == T(KW_STRUCT)
		 || _prev_yychar == T(KW_UNION) || _prev_yychar == T(KW_ENUM))
			type = (CLex::Token)(_prev_yychar);
		else
			type = T(NONE);

		yylval->scope = scope.scope;	
		if(yychar != T(ID))
			goto ReturnToken;
		do3(def, FindSym, type, *yylval, _current_scope);

		/* XXX - hack */
		if(def.token == T(CTOR_ID) && _next_yychar != T(LPAREN))
			def.token = T(AGGR_ID);
	
		switch(def.token) {
		case T(TYPE_ID):
		case T(ENUM_ID):
		case T(NSPACE_ID):
		case T(ASPACE_ID):
		case T(TPLATE_ID):
		case T(AGGR_ID):
		case T(VAL_ID):
			yychar = def.token;
			*yylval = def;
			break;
		case T(CTOR_ID):
			yychar = is_dtor ? T(DTOR_ID) : def.token;
			*yylval = def;
			break;
		default:
			break;
		}

		if(is_dtor && yychar != T(DTOR_ID)) {
			set_nextnext_yy(_next_yychar, _next_yylval);
			set_next_yy(yychar, *yylval);
			yychar = T(OP_NOT); *yylval = T(OP_NOT);
		}
	}

ReturnToken:		 
	if(_debug)
		_caction->eprintf("next token = %s\n",
				_caction->cstr((vCAction::Token)yychar));
	_prev_yychar = yychar;

	if(_savecode) {
		if(yychar == T(LBRACE))
			yychar = T(LBRACE_SAVE);
		else if(yychar == T(COLON))
			yychar = T(COLON_SAVE);
	}
	return yychar;
}

void
CParse::_warn_unused_type(Item_ t)
{
	_caction->warning("unused type declaration: %s\n", _caction->cstr(t.token));
}

int
CParse::parse()
{
	if(_do_parse()) {
		_caction->abort("cannot recover from earlier errors");
		return 0;
	} else {
		_caction->accept();
		return 1;
	}
}

