/*  $Id: CParse.hh,v 1.20 1998/08/22 03:35:13 shaggy Exp $
 *      C/C++ Parser
 *      
 *      relative path: .//lib/cparse
 *  
 *      Shag/OS: an Object-Oriented MicroKernel Operating System
 *       Libraries
 *        C/C++ Parsing classes
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the Shag/OS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare Non-profit Public License (SNPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare Non-profit Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

//  $Log: CParse.hh,v $
//  Revision 1.20  1998/08/22 03:35:13  shaggy
//  header update
//
//  Revision 1.19  1998/08/12 15:18:34  shaggy
//  better typing; fixed some type coercions;
//  cleanup for #include's, and fixed some constructor call bugs
//
//  Revision 1.18  1998/07/28 16:43:18  shaggy
//  moved lots of Deco into generic CParse lib
//  CAction split into vCAction and CAction
//  printaction -> part of CAction
//  ccint -> CRep (C Representation)
//  ccout -> COut (C output)
//
//  Revision 1.17  1998/07/11 19:02:58  shaggy
//  declarations in for() work now (even multiple decls)
//
//  Revision 1.16  1998/07/10 13:45:06  shaggy
//  working on decls in conditions
//
//  Revision 1.15  1998/07/09 14:34:37  shaggy
//  first phase of fixes for decl's in if(), while(), etc
//
//  Revision 1.14  1998/07/07 13:41:39  shaggy
//  enum cleanup, and can define types inside functions now
//
//  Revision 1.13  1998/07/04 20:38:57  shaggy
//  lots of bug fixes; fn, aggr, etc improvements, new/delete
//
//  Revision 1.12  1998/06/03 13:39:12  shaggy
//  function re-parsing almost works (for within classes)
//
//  Revision 1.11  1998/06/01 13:28:58  shaggy
//  switchable code save/parse
//
//  Revision 1.10  1998/05/25 23:28:42  shaggy
//  major rework of func/param info
//
//  Revision 1.9  1998/05/24 18:41:58  shaggy
//  improved regression testing
//  proper scoping for initializers
//  decls available for their initializer
//
//  Revision 1.8  1998/05/16 23:30:40  shaggy
//  added CTOR_ID and some rules to help fix ctor ambiguity problem
//  output turned off for #include'ed files
//  better error checking, and ability to list locations
//  of symbol declaration and definition
//
//  Revision 1.7  1998/05/15 00:17:24  shaggy
//  ctor initializers work, and enums are fixed so declarations
//  in an enum can be immediately re-used
//
//  Revision 1.6  1998/04/06 14:28:07  shaggy
//  got enums almost working
//
//  Revision 1.5  1998/03/21 00:23:49  shaggy
//  All kinds of new changes...
//  and it works!!!
//
//  Revision 1.4  1998/02/08 19:07:05  shaggy
//  about to start splitting up everything
//
//  Revision 1.3  1998/01/12 04:12:48  shaggy
//  got dependencies working when re-generating code
//
//  Revision 1.2  1998/01/11 05:11:03  shaggy
//   about to switch to standard ref/ptr semantics
//
//  Revision 1.1  1996/04/07  00:11:33  shaggy
//  Initial revision
//


#define CParse_VERSION_MAJOR	0
#define CParse_VERSION_MINOR	1
#define CParse_VERSION_STRING	"0.01" 

#include <CLex>
#include <vCAction>
#include <TokenList>


class ActionStack;


class CParse {
public:
	typedef vCAction::Item Item;
	typedef const vCAction::Item &Item_;
private:
	CLex *_clex;
	Item _next_yylval;
	Item _nextnext_yylval;
	short _next_yychar;
	short _nextnext_yychar;
	short _prev_yychar;
	bool _havenext;
	bool _havenextnext;
	bool _debug;
	TokenList *_toksrc;	/* tokens to use as input */
	int _saveline;
protected:	/* so auto-generated parser can access them */
	vCAction *_caction;
	bool _yydebug;
	Item _current_scope;
	Item _enum_val, _enum_type;
	Item _stmtlist;
	bool _savecode;
	bool _gsavecode; /* global savecode variable */
	bool _rawid;	/* if true, don't preprocess ID's for scoping, etc */
	TokenList *_deftok;	/* deferred tokens */
public:
	CParse(CLex &clex, vCAction &caction);
	int parse();
	bool debug() { return _debug; }
	void debug(bool d = true) { _debug = d; }
	bool yydebug() { return _yydebug; }
	void yydebug(bool d = true) { _yydebug = d; }
	void savecode(bool b = true) { _gsavecode = _savecode = b; }
	void inject(TokenList *tok);

#ifdef IMPLEMENTATION_CParse
private:
	int _do_parse();
	int _lex(Item *yylval);
	static void _internal_error(vErrorReport *er,
						csz *file, int line);

protected:	/* so auto-generated parser can access them */
	int _peekahead_lex(Item *yylval);
	void _warn_unused_type(Item_ t);
	void newdecl(Item*, Item_);

#define T(x)	vCAction::x
#define A(x)	vCAction::x

#define INTERNAL_ERROR CParse::_internal_error(_caction, __FILE__, __LINE__)

// *Sigh*... if only cpp macros supported variable #'s of arguments...

#define do(r,a) _caction->a(&r,none,none,none)
#define do1(r,a,p1) _caction->a(&r,p1,none,none)
#define do2(r,a,p1,p2) _caction->a(&r,p1,p2,none)
#define do3(r,a,p1,p2,p3) _caction->a(&r,p1,p2,p3)

#define none vCAction::none


#endif

};

