/*  $Id: CParse.y,v 1.110 1998/08/22 03:35:14 shaggy Exp $
 *      C/C++ Parser (shacc grammar)
 *      
 *      relative path: .//lib/cparse
 *  
 *      Shag/OS: an Object-Oriented MicroKernel Operating System
 *       Libraries
 *        C/C++ Parsing classes
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the Shag/OS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare Non-profit Public License (SNPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare Non-profit Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

/*  
 *  $Log: CParse.y,v $
 *  Revision 1.110  1998/08/22 03:35:14  shaggy
 *  header update
 *
 *  Revision 1.109  1998/08/17 06:38:16  shaggy
 *  more fixes for forward refs; __attribute__ support in aggrs
 *
 *  Revision 1.108  1998/08/12 15:18:35  shaggy
 *  better typing; fixed some type coercions;
 *  cleanup for #include's, and fixed some constructor call bugs
 *
 *  Revision 1.107  1998/08/11 12:50:43  shaggy
 *  getting conditional assignment to work
 *
 *  Revision 1.106  1998/08/09 21:49:53  shaggy
 *  DECO *finally* converts all of my original testcases automatically
 *
 *  Revision 1.105  1998/08/04 00:45:34  shaggy
 *  lots of fixes so it builds under Solaris again
 *
 *  Revision 1.104  1998/08/03 00:12:57  shaggy
 *  tons of work-- friends, class scope fixes, tons of bugfixes,
 *  lots more cleanup, more DC++ work, etc
 *
 *  Revision 1.103  1998/07/31 04:15:26  shaggy
 *  need sleep
 *
 *  Revision 1.102  1998/07/30 16:16:30  shaggy
 *  switched to new vsn of shacc and removed useless underscore prefixes
 *  as well as lot of extra configuration code that is no longer needed
 *
 *  Revision 1.101  1998/07/30 04:40:31  shaggy
 *  much more dtype and dclass work
 *
 *  Revision 1.100  1998/07/29 03:35:11  shaggy
 *  dtype support nearly complete, initial dclass support
 *
 *  Revision 1.99  1998/07/28 16:43:19  shaggy
 *  moved lots of Deco into generic CParse lib
 *  CAction split into vCAction and CAction
 *  printaction -> part of CAction
 *  ccint -> CRep (C Representation)
 *  ccout -> COut (C output)
 *
 *  Revision 1.98  1998/07/22 14:52:33  shaggy
 *  some namespace support
 *
 *  Revision 1.97  1998/07/20 17:51:45  shaggy
 *  added bitfield support
 *
 *  Revision 1.96  1998/07/20 00:45:35  shaggy
 *  fixed to allow cvqual after type in type-id
 *
 *  Revision 1.95  1998/07/19 22:33:34  shaggy
 *  part 2 of major overhaul (needed for type-safty and to start
 *  being more consistent)
 *
 *  Revision 1.94  1998/07/19 06:46:45  shaggy
 *  1st stage of Item cleanup
 *
 *  Revision 1.93  1998/07/17 18:55:23  shaggy
 *  fixed cpp support; added some parsing extensions
 *
 *  Revision 1.92  1998/07/16 23:21:42  shaggy
 *  more anon struct/union support
 *
 *  Revision 1.91  1998/07/16 13:44:51  shaggy
 *  parser ambiguity cleanup
 *
 *  Revision 1.90  1998/07/11 23:06:46  shaggy
 *  fixed labels and goto
 *
 *  Revision 1.89  1998/07/11 19:02:58  shaggy
 *  declarations in for() work now (even multiple decls)
 *
 *  Revision 1.88  1998/07/10 13:45:07  shaggy
 *  working on decls in conditions
 *
 *  Revision 1.87  1998/07/09 14:34:38  shaggy
 *  first phase of fixes for decl's in if(), while(), etc
 *
 *  Revision 1.86  1998/07/07 13:41:40  shaggy
 *  enum cleanup, and can define types inside functions now
 *
 *  Revision 1.85  1998/07/06 13:39:03  shaggy
 *  many more changes/fixes for first pre-release
 *
 *  Revision 1.84  1998/07/05 20:32:55  shaggy
 *  tons of bugfixes, enhancements and parser improvements
 *
 *  Revision 1.83  1998/07/04 20:38:58  shaggy
 *  lots of bug fixes; fn, aggr, etc improvements, new/delete
 *
 *  Revision 1.82  1998/07/04 01:23:29  shaggy
 *  anonymous unions work now
 *
 *  Revision 1.81  1998/06/30 18:19:01  shaggy
 *  more fixes for member pointers
 *
 *  Revision 1.80  1998/06/30 12:22:31  shaggy
 *  many fixes for "this"
 *
 *  Revision 1.79  1998/06/29 14:14:16  shaggy
 *  more DECO conversions, support for 'this', etc
 *
 *  Revision 1.78  1998/06/29 01:36:52  shaggy
 *  fixed inline attr, output of pure virtuals, etc...
 *
 *  Revision 1.77  1998/06/08 04:07:30  shaggy
 *  more fixes
 *
 *  Revision 1.76  1998/06/07 15:45:10  shaggy
 *  realtype() works, and . and -> now work
 *
 *  Revision 1.75  1998/06/05 03:34:18  shaggy
 *  parser fixes
 *
 *  Revision 1.74  1998/06/04 13:40:38  shaggy
 *  more fixes
 *
 *  Revision 1.73  1998/06/04 12:52:21  shaggy
 *  reparsing seems to work now
 *
 *  Revision 1.72  1998/06/03 13:39:13  shaggy
 *  function re-parsing almost works (for within classes)
 *
 *  Revision 1.71  1998/06/02 14:50:40  shaggy
 *  in the midst of more full parsing updates
 *
 *  Revision 1.70  1998/06/01 13:28:59  shaggy
 *  switchable code save/parse
 *
 *  Revision 1.69  1998/06/01 12:28:58  shaggy
 *  Fixed dangling-else problem (for IF/IF/ELSE)
 *
 *  Revision 1.68  1998/06/01 02:10:34  shaggy
 *  more code parsing working
 *
 *  Revision 1.67  1998/05/31 20:58:01  shaggy
 *  got reduction precedence working
 *
 *  Revision 1.66  1998/05/26 05:02:53  shaggy
 *  parameter scoping works correctly
 *  most basic C/C++ statements now work
 *
 *  Revision 1.65  1998/05/26 01:39:46  shaggy
 *  params almost working right
 *
 *  Revision 1.64  1998/05/25 23:28:43  shaggy
 *  major rework of func/param info
 *
 *  Revision 1.63  1998/05/24 18:41:59  shaggy
 *  improved regression testing
 *  proper scoping for initializers
 *  decls available for their initializer
 *
 *  Revision 1.62  1998/05/17 23:49:21  shaggy
 *  can handle GNU-style __attribute__ clause (so we can parse the
 *  standard headers)
 *
 *  Revision 1.61  1998/05/17 23:14:55  shaggy
 *  expressions, expression lists, etc
 *
 *  Revision 1.60  1998/05/16 23:30:41  shaggy
 *  added CTOR_ID and some rules to help fix ctor ambiguity problem
 *  output turned off for #include'ed files
 *  better error checking, and ability to list locations
 *  of symbol declaration and definition
 *
 *  Revision 1.59  1998/05/15 07:03:36  shaggy
 *  dependency ordering fixed; constructor initializers; lots of bug fixes, etc
 *
 *  Revision 1.58  1998/05/15 00:17:25  shaggy
 *  ctor initializers work, and enums are fixed so declarations
 *  in an enum can be immediately re-used
 *
 *  Revision 1.57  1998/05/14 18:02:03  shaggy
 *  fixed constant handling, enums, function cv tags, etc
 *
 *  Revision 1.56  1998/05/14 05:08:40  shaggy
 *  inheritence works and lots of bugs fixed
 *
 *  Revision 1.55  1998/05/13 00:21:15  shaggy
 *  fixed enums (mostly)
 *
 *  Revision 1.54  1998/05/12 05:00:12  shaggy
 *  more scope work
 *
 *  Revision 1.53  1998/05/10 18:53:41  shaggy
 *  just about ready to start fixing scoping
 *
 *  Revision 1.52  1998/05/10 00:03:31  shaggy
 *  string concatenation works, and comments are being
 *  added to get ready to fix scoping
 *
 *  Revision 1.51  1998/05/07 14:46:38  shaggy
 *  const/volatile partially working
 *
 *  Revision 1.50  1998/05/05 13:00:41  shaggy
 *  better function parameter matching
 *
 *  Revision 1.49  1998/05/04 03:09:18  shaggy
 *  support for ctors, dtors, conversion operators, operator overloading, etc
 *
 *  Revision 1.48  1998/05/03 02:47:07  shaggy
 *  added inheritence to parser
 *
 *  Revision 1.47  1998/04/27 04:36:47  shaggy
 *  started auto-generation of code
 *
 *  Revision 1.46  1998/04/26 15:21:23  shaggy
 *  better implicit int handling
 *
 *  Revision 1.45  1998/04/26 14:08:57  shaggy
 *  cleanup
 *
 *  Revision 1.44  1998/04/23 13:47:50  shaggy
 *  changes to make enums and constants work better
 *
 *  Revision 1.42  1998/04/22 13:06:38  shaggy
 *  better enum handling
 *
 *  Revision 1.41  1998/04/16 13:35:44  shaggy
 *  access control
 *
 *  Revision 1.40  1998/04/15 14:48:11  shaggy
 *  array parsing works
 *
 *  Revision 1.39  1998/04/15 13:39:44  shaggy
 *  added types to constants
 *
 *  Revision 1.38  1998/04/14 13:19:16  shaggy
 *  improved declaration parsing
 *
 *  Revision 1.37  1998/04/14 03:57:19  shaggy
 *  forward reference fix
 *
 *  Revision 1.36  1998/04/13 13:16:47  shaggy
 *  added %uniq and reorged tasks
 *
 *  Revision 1.35  1998/04/13 03:15:26  shaggy
 *  various updates
 *
 *  Revision 1.34  1998/04/12 03:20:00  shaggy
 *  improvments for function output
 *
 *  Revision 1.33  1998/04/11 21:49:27  shaggy
 *  added token saving (for regurgitation)
 *
 *  Revision 1.32  1998/04/11 01:47:40  shaggy
 *  deco uses cpp
 *  tests fixed for dc++
 *  dc++ grammar fixed
 *
 *  Revision 1.31  1998/04/06 14:28:07  shaggy
 *  got enums almost working
 *
 *  Revision 1.29  1998/03/23 14:17:02  shaggy
 *  lots more changes
 *
 *  Revision 1.28  1998/03/22 03:14:56  shaggy
 *  almost there...
 *
 *  Revision 1.27  1998/03/21 00:23:49  shaggy
 *  All kinds of new changes...
 *  and it works!!!
 *
 *  Revision 1.26  1998/03/20 21:52:21  shaggy
 *  most skeleton code now added (except templates, exceptions)
 *
 *  Revision 1.25  1998/03/20 17:36:26  shaggy
 *  all notes from Appendix A are in
 *
 *  Revision 1.24  1998/03/19 14:00:23  shaggy
 *  switching over to new parser
 *
 *  Revision 1.1  1998/02/23 14:33:52  shaggy
 *  working on "shacc"
 *
 *
 */


%{//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

#pragma Make Exclusive

#include <defines>
#include <inttypes.h>
#include <posix/stdarg.h>
#include <posix/string.h>
#include "ActionStack.h"
#include <TokenList>

#define IMPLEMENTATION_CParse
#include "CParse.hh"


#define YYSTYPE CParse::Item


#define yyassert(c)	{ if(!(c)) INTERNAL_ERROR; }
#define yyerror 	_caction->error
#define yywarning 	_caction->warning

#define YYCLASS		CParse
#define YYPARSE		_do_parse
#define YYLEX		_peekahead_lex
#define YYDEBUGFLAG	_yydebug
#define YYEPRINTF	_caction->eprintf
#define YYPARSE_VARS	\
	ActionStack actionstack(_caction);

//#define YYPDEBUG


// XXX - eventually, all of the constant strings (for errors/warnings)
// should be removed and replaced with error ID's
// that can be passed to CAction where they can have defaults
// and overrides.  This will also aid internationalization.
// The position specifier needs to be added to printf for this
// to entirely work correctly.



#define swap_items(a,b)	{ Item tmp = a; a = b; b = tmp; }


#define eval_decl(r, a, t) \
	{ actionstack.pop_lifo(&(r), (a), decl_actions, (t)); }
#define eval_pdecl(r, a, t) \
	{ actionstack.pop_lifo(&(r), (a), pdecl_actions, (t)); }

#define declSym(r, s, d) \
	{ actionstack.push(&(r), none, ACTION_Decl, T(TBC), (d), (s)); }
#define declSym2(r, a, s, d) \
	{ actionstack.push(&(r), (a), ACTION_Decl, T(TBC), (d), (s)); }
#define defineSym(r, s, v) \
	{ actionstack.push(&(r), none, ACTION_Define, (s), (v), \
						none); }
#define declPtr(r, op, d) \
	{ actionstack.push(&(r), (d), ACTION_Ptr, (op), T(TBC)); }
#define declArray(r, d, i) \
	{ actionstack.push(&(r), (d), ACTION_Array, T(TBC), i, 0); }
#define declFunc(r, d, p, c) \
	{ actionstack.push(&(r), (d), ACTION_Func, T(TBC), (p), (c)); }
#define declInit(r, d, v) \
	{ actionstack.push(&(r), (d), ACTION_Init, T(TBC), (p)); }
#define buildParamList(r, l, i) \
	{ actionstack.push(&(r), none, ACTION_BuildList, (l), (i)); }
#define buildDeclList(r, l, i) \
	{ actionstack.push(&(r), none, ACTION_None, (l), (i)); }

#define declInherit(r, t, b) \
	{ actionstack.push(&(r), none, ACTION_Inherit, T(TBC), (t), (b)); }
#define setAttr(r, d, a, v) \
	{ actionstack.push(&(r), (d), ACTION_SetAttr, T(TBC), (a), (v)); }


#define UNSUPPORTED yyerror("unsupported syntax (%d)", __LINE__);
#define NOSCOPE(s) if(s.scope) yyerror("scope not allowed");



enum ActionIndex {
	ACTION_None = 0,
	ACTION_Decl,
	ACTION_Define,
	ACTION_Ptr,
	ACTION_Array,
	ACTION_Func,
	ACTION_BuildList,
	ACTION_Inherit,
	ACTION_SetAttr,
};


static const vCAction::Action decl_actions[] = {
	0,		/* none */
	&A(DeclSym),	/* declarator */
	&A(Define),	/* define initial value */
	&A(DeclPtr),	/* pointer (*) or reference (&) declaration */
	&A(DeclArray),	/* array declaration ([n]) */
	&A(DeclFunc),	/* function declaration */
	&A(BuildParamList), /* parameter list (,) */
	&A(Inherit),	/* inherit from a base class */
	&A(SetAttr),	/* set attributes */
};

static const vCAction::Action pdecl_actions[] = {
	0,		/* none */
	&A(DeclParam),	/* declarator */
	&A(DefParam),	/* default value */
	&A(DeclPtr),	/* pointer (*) or reference (&) declaration */
	&A(DeclArray),	/* array declaration ([n]) */
	&A(DeclFunc),	/* function declaration */
	&A(BuildParamList), /* parameter list (,) */
	0,		/* none */
	0,		/* none */
};


/* Note: when defining parameter lists, the current scope,
	which is used for each parameter, can either be some arbitrary
	undefined scope, for arbitrary declarations,
	or the scope of the function, for the header on the definition */

/* Need a 'kill_scope' to get rid of unneeded scopes later, and
   reference counts should probably be used on all items so
   that those items referenced in a scope which are still needed 
   will be available */


#define curscope	_current_scope
#define savescope_new(s,f,t,i)	{ s = _current_scope; \
		do3(_current_scope, f, t, i, s); }
#define savescope_newlocal(s)	{ s = _current_scope; \
		do3(_current_scope, NewLocal, none, none, s); }
#define savescope_newfunc(s,d)	{ s = _current_scope; \
		do2(_current_scope, FuncScope, d, s); }
#define savescope_set(s, n)	{ s = _current_scope; _current_scope = n; }
#define savescope_swap(s)	{ Item t = s; s = _current_scope; _current_scope = t; }
#define savescope(s)		{ s = _current_scope; }
#define getscope_restore(g, r)	savescope_set(g, r)
#define restorescope(s) { _current_scope = s; }

%}//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


// The exact sequence of the following tokens must
// exactly match the sequence in token_table[] in CLex.C 
// Any additional tokens should be added above
// this line, or added appropriately to CLex.C

// Generic tokens:

%token ID		// identifier
%token NUM		// integer-literal
%token CHAR		// character-literal
%token FLOAT		// floating-literal
%token STR		// string-literal
%token CSTR		// constant string (for internal use)
%token BOOL		// boolean-literal
%token BACKSTR		// `string`
%token ANGSTR		// <string>
%token LINE		// % line
%token SPACE		// whitespace
%token EOL		// end of line

// CPP tokens:

%token CPP		// #
%token CPP_JOIN		// ##
%token CPP_DEFINED 	// defined

// CPP keywords:

%token CPP_DEFINE	// #define
%token CPP_UNDEF	// #undef
%token CPP_IF		// #if
%token CPP_IFDEF	// #ifdef
%token CPP_IFNDEF	// #ifndef
%token CPP_ENDIF	// #endif
%token CPP_ELIF		// #elif
%token CPP_INCLUDE	// #include
%token CPP_PRAGMA	// #pragma
%token CPP_ASSERT	// #assert
%token CPP_UNASSERT	// #unassert
%token CPP_LINE		// #line
%token CPP_ERROR	// #error
%token CPP_WARNING	// #warning
%token CPP_IDENT	// #ident

// CPP special substitution tokens
%token CPP__FILE	// __FILE__
%token CPP__LINE	// __LINE__
%token CPP__DATE	// __DATE__
%token CPP__TIME	// __TIME__
%token CPP__STDC	// __STDC__
%token CPP__STDCVER	// __STDC_VERSION__

// standard C punctuation:

%token EOS COLON COMMA		// ; : ,
%token QMARK POUND LBRACE RBRACE	// ? # { }
%token LBRACK RBRACK LPAREN RPAREN	// [ ] ( )

// extended punctuation:

%token QLINE DOTDOT AT BSLASH	// % .. @ \ 

// standard C operators:

%token OP_ADD OP_SUB			// + -
%token OP_MUL OP_DIV OP_MOD		// * / %
%token OP_NOT OP_AND OP_OR OP_XOR	// ~ & | ^
%token OP_LNOT OP_LAND OP_LOR OP_LXOR // ! && || ^^
%token OP_LSS OP_GTR OP_LEQ OP_GEQ	// < > <= >=
%token OP_EQU OP_NEQ OP_INC OP_DEC	// == != ++ --
%token OP_LSHIFT OP_RSHIFT		// << >>
%token OP_ASN				// =
%token OP_ASNADD OP_ASNSUB		// += -=
%token OP_ASNMUL OP_ASNDIV OP_ASNMOD	// *= /= %=
%token OP_ASNAND OP_ASNOR OP_ASNXOR	// &= |= ^=
%token OP_ASNLSHIFT OP_ASNRSHIFT	// <<= >>=
%token OP_DOT OP_PTRTO		// . ->

// C++ operators:

%token OP_SCOPE ELIPSIS OP_MDOT OP_MPTRTO	// :: ... .* ->*

// DC++ operators:

%token OP_CONDASN 			// ?=

// standard C keywords:

%token KW_ATTRIBUTE			// GNU-ish extension
%token KW_CONST KW_VOLATILE
%token KW_AUTO KW_STATIC KW_EXTERN KW_REGISTER
%token KW_LONG KW_SHORT KW_SIGNED KW_UNSIGNED
%token KW_INT KW_CHAR KW_FLOAT KW_DOUBLE KW_VOID
%token KW_IF KW_ELSE KW_DO KW_WHILE KW_FOR
%token KW_SWITCH KW_CASE KW_DEFAULT
%token KW_BREAK KW_CONTINUE KW_GOTO
%token KW_RETURN
%token KW_ENUM KW_UNION KW_STRUCT KW_TYPEDEF
%token OP_SIZEOF

// special C substitution strings:

%token KW__FUNC	// __FUNCTION__
%token KW__PFUNC	// __PRETTY_FUNCTION__

// C++ keywords:

%token KW_OPERATOR 
%token KW_CLASS KW_THIS
%token KW_TYPEID KW_TYPENAME
%token KW_NAMESPACE KW_USING
%token KW_TEMPLATE
%token KW_PUBLIC KW_PRIVATE KW_PROTECTED KW_FRIEND
%token KW_VIRTUAL KW_MUTABLE KW_EXPLICIT
%token KW_BOOL KW_TRUE KW_FALSE
%token KW_WCHAR
%token KW_THROW KW_CATCH KW_TRY
%token KW_SCAST KW_DCAST KW_CCAST KW_RCAST 
%token KW_INLINE KW_ASM
%token OP_TYPEOF
%token OP_NEW OP_DELETE
%token OP_NEWA OP_DELETEA	// new[] delete[]

// extended (DC++) keywords:

%token KW_DYNAMIC

/////////////////////////////////////////////////////////

// Additional tokens not generated by original lexer,
// but substituted in, and used for semantic parsing:

%token VAL_ID		// value (not type) identifier
%token AGGR_ID		// aggregate identifier (class,struct,union)
%token ENUM_ID		// enumerator identifier
%token TYPE_ID		// typedef'ed identifier
%token CTOR_ID		// constructor ID (matches enclosing scope)
%token DTOR_ID		// destructor ID ( OP_NOT AGGR_ID )
%token OPER_ID		// operator ID	(operator*  , etc...)
%token CONV_ID		// conversion ID  (operator int  , etc...)
%token NSPACE_ID	// namespace identifier
%token ASPACE_ID	// alias namespace identifier
%token TPLATE_ID	// template identifier
%token SCOPE		// scope specification
%token TOKENLIST	// a list of tokens saved for parsing later

/////////////////////////////////////////////////////////

// Additional tokens not generated by lexer, and not used for parsing:


// unary C/C++ operators

%token OP_REF				// &
%token OP_PTR				// *
%token OP_MPTR				// ::*
%token OP_UMINUS			// -
%token OP_UPLUS			// +
%token OP_INCP				// ++
%token OP_DECP				// --
%token OP_TSIZEOF			// sizeof(type)

// other C/C++ operators

%token OP_COMMA	// ,
%token OP_FUNC		// ()
%token OP_INDEX	// []
%token OP_CAST		// (old_style_cast)()

// various compound keywords

%token KW_DCLASS	// dynamic class (dclass)
%token KW_DTYPE	// dynamic virtual class (dtype)

// special tokens

%token NEXT_INSN	// use next instruction in sequence
%token BACKPATCH	// this will get backpatched later
%token ACTION		// used to maintain internal action lists
%token TBC		// to be computed (later)

// special tokens for parse control
%token LBRACE_SAVE
%token COLON_SAVE
%token PARSE_COMPOUND	// to parse a compound statement (function body)
%token DEFINE

%%//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


// For the syntax descriptions,
// the section numbers refer to "The C++ Programming Language" Third Edition
// by Bjarne Stroustrup. (1997)

// unquoted names represent other syntactical parts
// (which are italicized in the document), and quoted names
// represent literal text (keywords)
// any item followed by `(opt)' is optional.

// suffixes:
//	_opt		optional
//	_nodecl 	cannot look like a declarator
//	_nosdec		cannot look like a simple_declaration
//	_notype		cannot look like a type
//	_noptype	cannot have a type in parenthesis

//////////////////////////////////////////////////////////////////////////////
// PROGRAMS (A.4)
//
// translation-unit:
//	declaration-seq(opt)

translation_unit:
	  declaration_seq_opt
	| PARSE_COMPOUND compound_statement
		{ /* not used yet */ }
	;

//////////////////////////////////////////////////////////////////////////////
// DECLARATIONS (A.7)
//
// declaration-seq:
//	declaration
//	declaration-seq declaration

declaration_seq:
	  declaration_seq_opt { _caction->complete(false); } declaration
		{ _caction->complete(true);
		  if(_deftok && !_savecode) {
			inject(_deftok);
			_deftok = 0;
		  } }
	;

declaration_seq_opt: none | declaration_seq ;

// declaration:
//	block-declaration
//	function-definition
//	template-declaration
//	explicit-instantiation
//	explicit-specialization
//	linkage-specification
//	namespace-definition

declaration:
	  block_declaration %rprec 1 %uniq
		{}
	| none init_declarator_list EOS %rprec 2 %uniq
		{}
	| function_definition 
//x	| template_declaration 
//x	| explicit_instantiation 
//x	| explicit_specialization 
	| linkage_specification 
	| namespace_definition 	
	| namespace_definition EOS	
	| DEFINE anyscope id_or_ctor 
		{ savescope_swap($2); $$ = $3; } // $4
	    ctor_initializer_opt function_body
		{ if($5.token != T(NONE))
			do2($6, CtorInit, $5, $6);
		  do2($$, Define, $3, $6);
		  restorescope($2); }
	;

id_or_ctor: identifier | CTOR_ID;

anyscope: AGGR_ID | NSPACE_ID | ASPACE_ID | SCOPE;
 

// block-declaration:
//	simple-declaration
//	asm-definition
//	namespace-alias-definition
//	using-declaration
//	using-directive

block_declaration:
	  simple_declaration
	| asm_directive
	| namespace_alias_definition
	| using_declaration
	| using_directive
	;

// simple-declaration:
//	decl-specifier-seq(opt) init-declarator-list(opt) ;

simple_declaration:
	  decl_specifier_seq init_declarator_list EOS %rprec 1 
	| decl_specifier_seq EOS %rprec 2 
		{ do3($$, DeclSym, $1, none, curscope); }
	;

// decl-specifier:
//	storage-class-specifier
//	type-specifier
//	function-specifier
//	"friend"
//	"typedef"

decl_specifier_notypename:
	  storage_class_specifier
	| type_specifier_notypename
	| function_specifier
	| KW_FRIEND
	| KW_TYPEDEF
	;

// decl-specifier-seq:
//	decl-specifier-seq(opt) decl-specifier

decl_specifier_notypename_seq:
	  decl_specifier_notypename
		{ do2($$, BuildTypeList, none, $1); }
	| decl_specifier_notypename_seq decl_specifier_notypename
		{ do2($$, BuildTypeList, $1, $2); }
	;

decl_specifier_seq:
	  decl_specifier_notypename_seq
	| type_name
		{ do2($$, BuildTypeList, none, $1); }
	| decl_specifier_notypename_seq type_name
		{ do2($$, BuildTypeList, $1, $2); }
	| type_name decl_specifier_notypename_seq
		{ do2($$, BuildTypeList, $2, $1); }
	| decl_specifier_notypename_seq type_name decl_specifier_notypename_seq
		{ do2($1, BuildTypeList, $1, $2); 
		  do2($$, BuildTypeList, $1, $3); }
	;

decl_specifier_seq_opt: none | decl_specifier_seq ;

// storage-class-specifier:
//	"auto"
//	"register"
//	"static"
//	"extern"
//	"mutable"

storage_class_specifier:
	  KW_AUTO
	| KW_REGISTER
	| KW_STATIC
	| KW_EXTERN
	| KW_MUTABLE
	;

// function-specifier:
//	"inline"
//	"virtual"
//	"explicit"

function_specifier:
	  KW_INLINE
	| KW_VIRTUAL
	| KW_EXPLICIT
	;

// type-specifier:
//	simple-type-specifier
//	class-specifier
//	enum-specifier
//	elaborated-type-specifier
//	cv-qualifier

type_specifier_notypename:
	  simple_type_specifier_notypename
	| class_specifier
	| enum_specifier
	| elaborated_type_specifier
	| cv_qualifier
	;

// simple-type-specifier:
//	::(opt) nested-name-specifier(opt) type-name
//	"char"
//	"wchar_t"
//	"bool"
//	"short"
//	"int"
//	"long"
//	"signed"
//	"unsigned"
//	"float"
//	"double"
//	"void"

simple_type_specifier_notypename:
	  KW_CHAR
	| KW_WCHAR
	| KW_BOOL
	| KW_SHORT
	| KW_INT
	| KW_LONG
	| KW_SIGNED
	| KW_UNSIGNED
	| KW_FLOAT
	| KW_DOUBLE
	| KW_VOID
	;

// type-name:
//	class-name
//	enum-name
//	typedef-name

type_name:
	  class_name
	| enum_name
	| typedef_name
	;

// elaborated-type-specifier:
//	class-key ::(opt) nested-name-specifier(opt) identifier
//	"enum" ::(opt) nested-name-specifier(opt) identifier
//	"typename" ::(opt) nested-name-specifier identifier
//	"typename" ::(opt) nested-name-specifier identifier 
//				< template-argument-list >

elaborated_type_specifier:
	  class_key identifier
		{ do3($$, AggrRef, $1, $2, curscope);
		  newdecl(&_stmtlist, $$); }
	| KW_ENUM identifier
		{ do3($$, EnumRef, $1, $2, curscope);
		  newdecl(&_stmtlist, $$); }
//x	| KW_TYPENAME identifier
//x	| KW_TYPENAME identifier OP_LSS template_argument_list OP_GTR
	;

// enum-specifier:
//	"enum" identifier(opt) { enumerator-list(opt) }

enum_specifier:
	  KW_ENUM identifier_opt lbrace %uniq
		{ do3(_enum_type, NewEnum, $1, $2, curscope);
		  _enum_val = none;	
		  newdecl(&_stmtlist, _enum_type); }
	    enumerator_list_opt comma_opt RBRACE
		{ $$ = _enum_type; }
	;

// enumerator-list:
//	enumerator-definition
//	enumerator-list , enumerator-definition

enumerator_list:
	  enumerator_definition
	| enumerator_list COMMA enumerator_definition
	;

enumerator_list_opt: none | enumerator_list ;

// enumerator-definition:
//	enumerator
//	enumerator = constant-expression

// enumerator:
//	identifier

enumerator_definition:
	  identifier
		{ Item s;
		  do3(s, DeclSym, _enum_type, $1, curscope);
		  do1(_enum_val, EnumNext, _enum_val);
		  do2($$, Define, s, _enum_val); }
	| identifier OP_ASN constant_expression
		{ do3($2, DeclSym, _enum_type, $1, curscope);
		  _enum_val = $3;
		  do2($$, Define, $2, $3); }
	;

// namespace-definition:
//	named-namespace-definition
//	unnamed-namespace-definition

namespace_definition:
	  named_namespace_definition
	| unnamed_namespace_definition
	;

// named-namespace-definition:
//	original-namespace-definition
//	extension-namespace-definition

named_namespace_definition:
	  original_namespace_definition
	| extension_namespace_definition
	;

// original-namespace-definition:
//	"namespace" identifier { namespace-body }

original_namespace_definition:
	  KW_NAMESPACE identifier lbrace %uniq
		{ savescope_new($$, NewNspace, $1, $2); } // $3
	      namespace_body RBRACE %uniq
		{ getscope_restore($$, $3); }
	;

// extension-namespace-definition:
//	"namespace" original-namespace-name { namespace-body }

extension_namespace_definition:
	  KW_NAMESPACE original_namespace_name lbrace %uniq
		{ savescope_set($$, $2); } // $$ = $4 
	      namespace_body RBRACE %uniq
		{ getscope_restore($$, $4); }
	;

// unnamed-namespace-definition:
//	"namespace" { namespace-body }

unnamed_namespace_definition:
	  KW_NAMESPACE lbrace %uniq
		{ savescope_new($$, NewNspace, $1, none); } // $3
	      namespace_body RBRACE %uniq
		{ getscope_restore($$, $3); }
	;

// namespace-body:
//	declaration-seq(opt)

namespace_body:
	  declaration_seq_opt
	;

// namespace-alias-definition:
//	"namespace" identifier = qualifier-namespace-specifier ;

namespace_alias_definition:
	  KW_NAMESPACE identifier OP_ASN namespace_name EOS %uniq
		{ UNSUPPORTED /* 1 */ }
	;

// qualified-namespace-specifier:
//	::(opt) nested-name-specifier (opt) namespace-name

// using-declaration:
//	"using" "typename"(opt) ::(opt) nested-name-specifier unqualified-id ;
//	"using" :: unqualified-id ;

using_declaration:
	  KW_USING KW_TYPENAME unqualified_id EOS %uniq
		{ UNSUPPORTED /* 2 */ }
	| KW_USING unqualified_id EOS %uniq
		{ UNSUPPORTED /* 3 */ }
	| KW_USING SCOPE unqualified_id EOS %uniq
		{ UNSUPPORTED /* 4 */ }
	;

// using-directive:
//	"using" "namespace" ::(opt) nested-name-specifier(opt) namespace-name ;

using_directive:
	  KW_USING KW_NAMESPACE namespace_name EOS %uniq
		{ UNSUPPORTED /* 5 */ }
	;

// asm-directive:
//	"asm" ( string-literal ) ;

asm_directive:
	  KW_ASM LPAREN STR RPAREN EOS %uniq
		{ UNSUPPORTED /* 6 */ }
	;

// linkage-specification:
//	"extern" string-literal { declaration-seq(opt) }
//	"extern" string-literal declaration

linkage_specification:
	  KW_EXTERN STR lbrace declaration_seq_opt RBRACE
	| KW_EXTERN STR declaration
	;

//////////////////////////////////////////////////////////////////////////////
// Declarators (A.7.1)

// init-declarator-list:
//	init-declarator
//	init-declarator-list , init-declarator

// init-declarator:
//	declarator initializer(opt)

save_scope:
	  { savescope($$); }
	;

init_declarator_list1:
	  save_scope init_declarator COMMA save_scope
		{ $$ = $4; }
	| init_declarator_list1 init_declarator COMMA save_scope
		{ $$ = $4; }
	;

init_declarator_list:
	  save_scope init_declarator
		{ $$ = $1; }
	| init_declarator_list1 init_declarator
		{ $$ = $1; }
	;

init_declarator_list_opt: none | init_declarator_list;


init_declarator:
	  declarator attribute_opt 
		{ eval_decl($$, $1, $-1);
		  newdecl(&_stmtlist, $$);
		  restorescope($0); }
	| declarator attribute_opt 
		{ eval_decl($$, $1, $-1); 
		  newdecl(&_stmtlist, $$); }
	     initializer 
		{ do2($$, Define, $3, $4); 
		  restorescope($0); 
		  $$ = $3; }
	| declarator attribute_opt 
		{ eval_decl($$, $1, $-1); 
		  newdecl(&_stmtlist, $$); }
	     OP_CONDASN assignment_expression
		{ do2($$, Define, $3, $5); 
		  restorescope($0); 
		  $$ = $3; }
	;
 
// declarator:
//	direct-declarator
//	ptr-operator declarator

declarator:
	  direct_declarator
	| ptr_operator declarator
		{ declPtr($$, $1, $2); }
	;

// direct-declarator:
//	declarator-id
//	direct-declarator ( parameter-declaration-clause ) 
//			cv-qualifier-seq(opt) exception-specification(opt)
//	direct-declarator [ constant-expression(opt) ]
//	( declarator )

direct_declarator:
	  declarator_id
	| direct_declarator LPAREN parameter_declaration_clause RPAREN
			cv_qualifier_seq_opt //x exception_specification_opt
		{ declFunc($$, $1, $3, $5); }
	| direct_declarator LBRACK constant_expression RBRACK
		{ declArray($$, $1, $3); }
	| direct_declarator LBRACK RBRACK
		{ declArray($$, $1, none); }
	| LPAREN declarator RPAREN
		{ $$ = $2; }
	;

ctor_declarator_id:
	  CTOR_ID
		{ do3($1, SpecialId, T(CTOR_ID), $1, curscope);
		  declSym($$, curscope, $1); 
		  if($1.scope) {
			if(curscope.token == T(NONE)) {
				curscope.token = T(SCOPE);
				curscope.scope = $1.scope;
			} else 
				yyerror("scope cannot be specified for"
					" non-global constructor");
		  }
		}
	;

ctor_declarator:
	  ctor_declarator_id LPAREN parameter_declaration_clause RPAREN
			cv_qualifier_seq_opt //x exception_specification_opt
		{ declFunc($$, $1, $3, $5); }
	;

declarator_or_ctor:
	  ctor_declarator	%rprec 1
	| declarator		%rprec 2
	;

// ptr-operator:
//	* cv-qualifier-seq(opt)
//	&
//	::(opt) nested-name-specifier * cv-qualifier-seq(opt)

ptr_operator:
	  OP_MUL cv_qualifier_seq_opt
		{ $$ = T(OP_PTR); $$.val = $2.val; }
	| OP_AND
		{ $$ = T(OP_REF); }
	| SCOPE OP_MUL cv_qualifier_seq_opt
		{ $$ = T(OP_MPTR); $$.scope = $1.scope; $$.val = $3.val; }
	;

// cv-qualifier-seq:
//	cv-qualifier cv-qualifier-seq(opt)

cv_qualifier_seq:
	  cv_qualifier cv_qualifier_seq_opt
		{ do2($$, BuildTypeList, $2, $1); }
	;

cv_qualifier_seq_opt: none | cv_qualifier_seq ;

// cv-qualifier:
//	"const"
//	"volatile"

cv_qualifier:
	  KW_CONST
	| KW_VOLATILE
	;

// declarator-id:
//	::(opt) id-expression
//	::(opt) nested-name-specifier(opt) type-name

declarator_id:
	  id_expression
		{ 
		  declSym($$, curscope, $1); 
		  if($1.scope) {
			if(curscope.token == T(NONE)) {
				curscope.token = T(SCOPE);
				curscope.scope = $1.scope;
			} else 
				yyerror("scope cannot be specified for"
					" non-global declarator");
		  }
		}
	;

// type-id:
//	type-specifier-seq abstract-declarator(opt)

type_id:
	  type_specifier_seq abstract_declarator_opt
		{ eval_decl($$, $2, $1); }
	;

// type-specifier-seq:
//	type-specifier type-specifier-seq(opt)

type_specifier_notypename_seq:
	  type_specifier_notypename
		{ do2($$, BuildTypeList, none, $1); }
	| type_specifier_notypename_seq type_specifier_notypename
		{ do2($$, BuildTypeList, $1, $2); }
	;

type_specifier_seq:
	  cv_qualifier_seq_opt type_name cv_qualifier_seq_opt
		{ do2($$, BuildTypeList, $1, $2);
		  if($3.token != T(NONE))
		  	do2($$, BuildTypeList, $$, $3); }
	| type_specifier_notypename_seq
	;

// abstract-declarator:
//	ptr-operator abstract-declarator(opt)
//	direct-abstract-declarator

abstract_declarator:
	  ptr_operator abstract_declarator_opt
		{ declPtr($$, $1, $2); }
	| direct_abstract_declarator
	;

abstract_declarator_opt: none | abstract_declarator ;

// direct-abstract-declarator:
//	direct-abstract-declarator(opt) ( parameter-declaration-clause )
//			cv-qualifier-seq(opt) exception-specification(opt)
//	( abstract-declarator )

direct_abstract_declarator:
	  LPAREN parameter_declaration_clause
		RPAREN cv_qualifier_seq_opt //x exception_specification_opt
		{ declFunc($$, none, $2, $4); }
	| direct_abstract_declarator LPAREN parameter_declaration_clause
		RPAREN cv_qualifier_seq_opt //x exception_specification_opt
		{ declFunc($$, $1, $3, $5); }
	| LPAREN abstract_declarator RPAREN
		{ $$ = $2; }
	;

// parameter-declaration-clause:
//	parameter-declaration-list(opt) ...(opt)
//	parameter-declaration-list , ...

parameter_declaration_clause:	// optional
		{ do3($$, BuildParamList, none, none, curscope); }
	| parameter_declaration_list
	| parameter_declaration_list comma_opt ELIPSIS
		{ do3($$, BuildParamList, $1, T(ELIPSIS), curscope); }
	| ELIPSIS
		{ do3($$, BuildParamList, none, T(ELIPSIS), curscope); }
	;

// parameter-declaration-list:
//	parameter-declaration
//	parameter-declaration-list , parameter-declaration

parameter_declaration_list:
	  parameter_declaration
		{ do3($$, BuildParamList, none, $1, curscope); }
	| parameter_declaration_list COMMA parameter_declaration
		{ do3($$, BuildParamList, $1, $3, curscope); }
	;

// parameter-declaration:
//	decl-specifier-seq declarator
//	decl-specifier-seq declarator = assignment-expression
//	decl-specifier-seq abstract-declarator(opt)
//	decl-specifier-seq abstract-declarator(opt) = assignment-expression

parameter_declaration:
	  decl_specifier_seq declarator OP_ASN assignment_expression %rprec 1
		{ eval_pdecl($3, $2, $1);
		  do2($$, DefParam, $3, $4); }
	| decl_specifier_seq declarator %rprec 1
		{ eval_pdecl($$, $2, $1); }
	| decl_specifier_seq abstract_declarator_opt
				 OP_ASN assignment_expression %rprec 2
		{ eval_pdecl($3, $2, $1);
		  do2($1, DeclParam, $3, none);
		  do2($$, DefParam, $1, $4); }
	| decl_specifier_seq abstract_declarator_opt %rprec 2
		{ eval_pdecl($$, $2, $1); $2 = $$;
		  do2($$, DeclParam, $2, none); }
	; 

// function-definition:
//	decl-specifier-seq(opt) declarator ctor-initializor(opt) function-body
//	decl-specifier-seq(opt) declarator function-try-block

eval_declarator_or_ctor:
	  declarator_or_ctor
		{ eval_decl($$, $1, $-1); 
			yywarning("%p,%p,%p\n", &$-1, &$0, &$1); }
	;

function_definition:
	  decl_specifier_seq_opt save_scope declarator_or_ctor 
		{ eval_decl($$, $3, $1); }
	    ctor_initializer_opt function_body
		{ if($5.token == T(TOKENLIST)) {
		  	$5.TokenList->append_delete($6.TokenList);
			$6 = $5;
		  } else if($5.token != T(NONE))
			do2($6, CtorInit, $5, $6);
		  if(_savecode && !_gsavecode) {
			if(!_deftok)
				_deftok = new TokenList;
			_deftok->append(T(DEFINE));
			_deftok->append(curscope);
			_deftok->append($4);
			_deftok->append_delete($6.TokenList);
			$$ = $4;
		  } else 
		  	do2($$, Define, $4, $6);
		  restorescope($2); }
//x	| decl_specifier_seq_opt declarator function_try_block
//x		{ UNSUPPORTED /* 7 */ }
	;


// function-body:
//	compound-statement

function_body:
		{ savescope_newfunc($$, $-1); }	// $1
	    compound_statement
		{ restorescope($1); $$ = $2; }
	| save_compound_statement
	;

// initializer:
//	= initializer-clause
//	( expression-list )
//    DC++:
//	?= assignment_expression (handled where 'initializer' is used)

initializer:
	  OP_ASN initializer_clause
		{ $$ = $2; }
	| LPAREN expression_list RPAREN
		{ $$ = $2; }
	;

// initializer-clause:
//	assignment-expression
//	{ initializer-list ,(opt) }
//	{ }

initializer_clause:
	  assignment_expression
	| lbrace initializer_list_opt comma_opt RBRACE %uniq
		{ $$ = $2; }
	;

comma_opt: | COMMA ;

// initializer-list:
//	initializer-clause
//	initializer-list , initializer-clause
//    C/C++ extension:
//	[ constant-expression ] initializer-clause
//	initializer-list , [ constant-expression ] initializer-clause
//	[ constant-expression ] initializer-clause
//	initializer-list , [ constant-expression ] initializer-clause

prev_initializer_list:
	  initializer_list COMMA
	| none
	;

initializer_list:
	  prev_initializer_list initializer_clause
		{ do2($$, BuildArray, $1, $2); }
	| prev_initializer_list LBRACK constant_expression RBRACK
					initializer_clause
		{ do2($$, BuildArray, $1, $5); }
		// XXX - need to use constant-expression
	| prev_initializer_list OP_DOT id_expression initializer_clause
		{ do2($$, BuildArray, $1, $4); }
		// XXX - need to use id-expression
	;

initializer_list_opt:	// optional
		{ do2($$, BuildArray, none, none); }
	| initializer_list
	;


//////////////////////////////////////////////////////////////////////////////
// CLASSES (A.8)


// class-specifier:
//	class-head { member-specification(opt) }

// class-head:
//	class-key identifier(opt) base-clause(opt)
//	class-key nested-name-specifier identifier base-clause(opt)

class_specifier:
	  class_key identifier_opt base_clause_opt lbrace %uniq
		{ do3($4, NewAggr, $1, $2, curscope); 
		  newdecl(&_stmtlist, $4);
		  if($3.token != T(NONE)) { Item t;
		  	eval_decl(t, $3, $4); $4 = t; }
		  savescope_set($$, $4);
		  $1.Bool = _savecode; _savecode = true; } // $5
	     member_specification_opt RBRACE %uniq
		{ do2($$, EndAggr, $4, $5);
		  restorescope($5);
		  _savecode = $1.Bool; }
	;


// class-key:
//	"class"
//	"struct"
//	"union"
//    DC++ extension:
//	"dynamic" "class"
//	"dynamic" "virtual" "class"

class_key:
	  KW_CLASS
	| KW_STRUCT
	| KW_UNION

	// DC++ extension:
	| KW_DYNAMIC KW_CLASS
		{ $$ = T(KW_DCLASS); }
	| KW_DYNAMIC KW_VIRTUAL KW_CLASS
		{ $$ = T(KW_DTYPE); }
	// for re-parsing:
	| KW_DCLASS
	| KW_DTYPE
	;

// member-specification:
//	member-declaration member-specification(opt)
//	access-specifier : member-specification(opt)
//    DC++ extension:
//	class-name : member-specification(opt)

member_specification:
	  member_declaration member_specification_opt
	| access_specifier colon
		{ do2($$, SetAccess, curscope, $1); curscope = $$; }
		member_specification_opt
	// DC++ extension:
	| class_name colon
		{ do3($1, FindSym, T(AGGR_ID), $1, curscope); 
		  if($1.token != T(NONE)) {
			do2($$, SetAccess, curscope, $1); curscope = $$;
		  }}
		member_specification_opt
	| KW_DYNAMIC class_name colon
		{ $1.token = T(AGGR_ID);
		  $1.tval = T(KW_DYNAMIC);
		  do3($2, FindSym, $1, $2, curscope);
		  if($1.token != T(NONE)) {
			do2($$, SetAccess, curscope, $2); curscope = $$;
		  }}
		member_specification_opt	
	;

member_specification_opt: none | member_specification ;

// member-declaration:
//	decl-specifier-seq(opt) member-declarator-list(opt) ;
//	function-definition ;(opt)
//	qualified-id ;
//	using-declaration
//	template-declaration

member_declaration:
	  EOS %uniq
	| class_key identifier EOS %rprec 1 %uniq
		{ do3($$, AggrDecl, $1, $2, curscope);
		  newdecl(&_stmtlist, $$); }
	| decl_specifier_seq member_declarator_list EOS %rprec 2 %uniq 
		{ eval_decl($$, $2, $1); }
	| decl_specifier_seq EOS %rprec 3 %uniq
		{ do3($$, DeclSym, $1, none, curscope); }
	| member_declarator_list EOS %rprec 4 %uniq
		{ eval_decl($$, $1, none); }
	| function_definition
		{ do3($$, SetAttr, $1, T(KW_INLINE), none); }
//x	| qualified_id EOS %uniq
	| using_declaration
//x	| template_declaration
	;
 
// member-declarator-list:
//	member-declarator
//	member-declarator-list , member-declarator

member_declarator_list:
	  member_declarator attribute_opt
	| member_declarator_list COMMA member_declarator attribute_opt
		{ buildDeclList($$, $1, $3); }
	;

// member-declarator:
//	declarator pure-specifier(opt)
//	declarator constant-initializer(opt)
//	identifier(opt) : constant-expression

// pure-specifier:
//	= 0

// constant-initializer:
//	= constant-expression

member_declarator:
	  declarator_or_ctor
	| declarator OP_ASN constant_expression
		{ defineSym($$, $1, $3); }
	| colon constant_expression
		{ $$ = none; }
	| identifier colon constant_expression
		{ setAttr($2, T(TBC), T(COLON), $3);
		  declSym2($$, $2, curscope, $1); }
	;


//////////////////////////////////////////////////////////////////////////////
// Derived Classes (A.8.1)

// base-clause:
//	: base-specifier-list

base_clause_opt: 
	  none
	| colon base_specifier_list
		{ $$ = $2; }
	;

// base-specifier-list:
//	base-specifier
//	base-specifier-list , base-specifier

base_specifier_list:
	  base_specifier
	| base_specifier_list COMMA base_specifier
		{ buildDeclList($$, $1, $3); }
	;

// base-specifier:
//	::(opt) nested-name-specifier(opt) class-name
//	"virtual" access-specifier(opt) ::(opt) nested-name-specifier(opt) 
//						class-name
//	access-specifier "virtual"(opt) ::(opt) nested-name-specifier(opt)
//						class-name

base_specifier:
	  class_name
		{ declInherit($$, none, $1); }
	| KW_VIRTUAL class_name
		{ $1.token = T(NONE); $1.tval = T(KW_VIRTUAL); declInherit($$, $1, $2); }
	| KW_VIRTUAL access_specifier class_name
		{ $2.tval = T(KW_VIRTUAL); declInherit($$, $2, $3); }
	| access_specifier class_name
		{ declInherit($$, $1, $2); }
	| access_specifier KW_VIRTUAL class_name
		{ $1.tval = T(KW_VIRTUAL); declInherit($$, $1, $3); }
	;

// access-specifier:
//	"private"
//	"protected"
//	"public"

access_specifier:
	  KW_PRIVATE
	| KW_PROTECTED
	| KW_PUBLIC
	;

//////////////////////////////////////////////////////////////////////////////
// Special Member Functions (A.8.2)

// conversion-function-id:
//	"operator" conversion-type-id

conversion_function_id:
	  KW_OPERATOR conversion_type_id
		{ $2.scope = 0;
		  do3($$, SpecialId, T(CONV_ID), $2, curscope); }
	| SCOPE KW_OPERATOR conversion_type_id
		{ $3.scope = $1.scope;
		  do3($$, SpecialId, T(CONV_ID), $3, curscope); }
	;

// conversion-type-id:
//	type-specifier-seq conversion-declarator(opt)

conversion_type_id:
	  type_specifier_seq conversion_declarator_opt
		{ eval_decl($$, $2, $1); }
	;

// conversion-declarator:
//	ptr-operator conversion-declarator(opt)

conversion_declarator:
	  ptr_operator conversion_declarator_opt
		{ declPtr($$, $1, $2); }
	;

conversion_declarator_opt: none | conversion_declarator ;

// ctor-initializer:
//	: mem-initializer-list

ctor_initializer:
	  COLON_SAVE 
		{ $$.Bool = _rawid; _rawid = true; }
	     save_mem_initializer_list
		{ _rawid = $2.Bool;
		  $1.token = T(COLON);
		  ($3.TokenList)->prepend($1); 
		  $$ = $3; }
	| COLON 
		{ savescope_newfunc($$, $0); } // $2
	     mem_initializer_list
		{ $$ = $3; restorescope($2); }
	;

ctor_initializer_opt: none | ctor_initializer;

// mem-initializer-list:
//	mem-initializer
//	mem-initializer , mem-initializer-list

save_mem_initializer_list:
	  save_mem_initializer
	| save_mem_initializer_list COMMA save_mem_initializer
		{ ($1.TokenList)->append($2); 
		  ($1.TokenList)->append_delete($3.TokenList); 
		  $$ = $1; }
	;

mem_initializer_list:
	  mem_initializer
		{ do3($$, BuildStmtList, none, $1, curscope); }
	| mem_initializer_list COMMA mem_initializer
		{ do3($$, BuildStmtList, $1, $3, curscope); }
	;

// mem-initializer:
//	mem-initializer-id ( expression-list(opt) )

// mem-initializer-id:
//	::(opt) nested-name-specifier(opt) class-name
//	identifier

save_mem_initializer:
	  identifier LPAREN save_statement_seq_opt RPAREN
		{ $1.token = T(ID); 
		  ($3.TokenList)->prepend($2);
		  ($3.TokenList)->prepend($1);
		  ($3.TokenList)->append($4);
		  $$ = $3; }
	;

mem_initializer:
	  identifier LPAREN expression_list_opt RPAREN
		{ do3($$, CtorCall, $1, $3, curscope); }
	;


//////////////////////////////////////////////////////////////////////////////
// Overloading (A.8.3)

// operator-function-id:
//	"operator" operator

operator_function_id:
	  KW_OPERATOR operator
		{ $2.scope = 0;
		  do3($$, SpecialId, T(OPER_ID), $2, curscope); }
	| SCOPE KW_OPERATOR operator
		{ $3.scope = $1.scope;
		  do3($$, SpecialId, T(OPER_ID), $3, curscope); }
	;

// operator: one of
//   "new" 	"delete" 	"new[]" 	"delete[]"
//   +     -     *     /     %     ^     &     |     ~     !     =     <     >
//   +=    -=    *=    /=    %=    ^=    &=    |=    <<    >>    >>=   <<=   ==
//   !=    <=    >=    &&    ||    ++    --    ,     ->*   ->    ()    []

simple_operator:
	  OP_NEW | OP_DELETE | OP_NEWA | OP_DELETEA
	| OP_ADD | OP_SUB | OP_MUL | OP_DIV | OP_MOD | OP_XOR | OP_AND
	| OP_OR | OP_NOT | OP_LNOT | OP_ASN | OP_LSS | OP_GTR
	| OP_ASNADD | OP_ASNSUB | OP_ASNMUL | OP_ASNDIV | OP_ASNMOD
	| OP_ASNXOR | OP_ASNAND | OP_ASNOR | OP_LSHIFT | OP_RSHIFT
	| OP_ASNRSHIFT | OP_ASNLSHIFT | OP_EQU
	| OP_NEQ | OP_LEQ | OP_GEQ | OP_LAND | OP_LOR | OP_INC
	| OP_DEC | COMMA | OP_MPTRTO | OP_PTRTO 
	;

operator:
	  simple_operator
	| LPAREN RPAREN
		{ $$ = T(OP_FUNC); }
	| LBRACK RBRACK
		{ $$ = T(OP_INDEX); }
	;


//////////////////////////////////////////////////////////////////////////////
// TEMPLATES (A.9)

// template-declaration:
//	"export"(opt) "template" < template-parameter-list > declaration

// template-parameter-list:
//	template-parameter
//	template-parameter-list , template-parameter

// template-parameter:
//	type-parameter
//	parameter-declaration

// type-parameter:
//	"class" identifier(opt)
//	"class" identifier(opt) = type-id
//	"typename" identifier(opt)
//	"typename" identifier(opt) = type-id
//	"template" < template-parameter-list > "class" identifier(opt)
//	"template" < template-parameter-list > "class" identifier(opt)
//					 = template-name

// template-id:
//	template-name < template-argument-list >

// template-argument-list:
//	template-argument
//	template-argument-list , template-argument

// template-argument:
//	assignment-expression
//	type-id
//	template-name

// explicit-instantiation:
//	template declaration

// explicit-specialization:
//	"template" < > declaration


//////////////////////////////////////////////////////////////////////////////
// EXCEPTION HANDLING (A.10)

// try-block:
//	"try" compound-statement handler-seq

// function-try-block:
//	"try" ctor-initializer(opt) function-body handler-seq

// handler-seq:
//	handler handler-seq(opt)

// handler:
//	"catch" ( exception-declaration ) compound-statement

// exception-declaration:
//	type-specifier-seq declarator
//	type-specifier-seq abstract-declarator
//	type-specifier-seq
//	...

// throw-expression:
//	"throw" assignment-expression(opt)

// exception-specification:
//	"throw" ( type-id-list(opt) )

// type-id-list:
//	type-id
//	type-id-list , type-id



//////////////////////////////////////////////////////////////////////////////
// EXPRESSIONS (A.5)

// primary-expression:
//	literal
//	"this"
//	:: identifier
//	:: operator-function-id
//	:: qualified-id
//	( expression )
//	id-expression

literal:
	  NUM
	| CHAR
	| STR
	| FLOAT
	| BOOL
	;

primary_expression:
	  literal
		{ do1($$, Literal, $1); }
	| KW_THIS
		{ do3($$, SpecialId, $1, $1, curscope); }
	| LPAREN expression RPAREN
		{ $$ = $2; }
	| id_expression_val
		{ do2($$, Primary, $1, curscope); }
	;

// id-expression:
//	unqualified-id
//	qualified-id

id_expression:
	  unqualified_id
//x	| qualified_id
	;

id_expression_val:
	  unqualified_id_val
	;

// unqualified-id:
//	identifier
//	operator-function-id
//	conversion-function-id
//	~ class-name
//	template-id

unqualified_id:
	  identifier
	| operator_function_id
	| conversion_function_id
	| DTOR_ID
		{ do3($$, SpecialId, T(DTOR_ID), $1, curscope); }
//x	| template_id
	;

unqualified_id_val:
	  VAL_ID
	| ID
		{ do2($$, Undeclared, $1, curscope); }
	| operator_function_id
	| conversion_function_id
	| DTOR_ID
		{ do3($$, SpecialId, T(DTOR_ID), $1, curscope); }
	;

// qualified-id:
//	nested-name-specifier "template"(opt) unqualified-id

// nested-name-specifier:
//	class-or-namespace-name :: nested-name-specifier(opt)

// class-or-namespace-name:
//	class-name
//	namespace-name

// NOTE: this probably won't be necessary since CParse.cc handles scoping



// postfix-expression:
//	primary-expression
//	postfix-expression [ expression ]
//	postfix-expression ( expression-list(opt) )
//	simple-type-specifier ( expression-list(opt) )
//	postfix-expression . "template"(opt) ::(opt) id-expression
//	postfix-expression -> "template"(opt) ::(opt) id-expression
//	postfix-expression . pseudo-destructor-name
//	postfix-expression -> pseudo-destructor-name
//	postfix-expression ++
//	postfix-expression --
//	"dynamic_cast" < type-id > ( expression )
//	"static_cast" < type-id > ( expression )
//	"reinterpret_cast" < type-id > ( expression )
//	"const_cast" < type-d > ( expression )
//	"typeid" ( expression )
//	"typeid" ( type-id )

postfix_expression:
	  primary_expression
	| postfix_expression LBRACK expression RBRACK
		{ do2($$, Index, $1, $3); }
	| postfix_expression LPAREN expression_list_opt RPAREN
		{ do3($$, Call, $1, $3, curscope); }
	| simple_type_specifier_notypename LPAREN expression_list_opt RPAREN
		{ do3($$, CtorCall, $1, $3, curscope); }
	| type_name LPAREN expression_list_opt RPAREN
		{ do3($$, CtorCall, $1, $3, curscope); }
	| postfix_expression OP_DOT id_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	| postfix_expression OP_PTRTO id_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	| postfix_expression OP_DOT pseudo_destructor_name
		{ do3($$, BinaryOp, $2, $1, $3); }
	| postfix_expression OP_PTRTO pseudo_destructor_name
		{ do3($$, BinaryOp, $2, $1, $3); }
	| postfix_expression OP_INC
		{ do2($$, UnaryOp, T(OP_INCP), $1); }
	| postfix_expression OP_DEC
		{ do2($$, UnaryOp, T(OP_DECP), $1); }
	| newcast OP_LSS type_id OP_GTR LPAREN expression RPAREN
		{ do3($$, Cast, $6, $3, $1); }
	| KW_TYPEID LPAREN expression RPAREN
		{ UNSUPPORTED /* 8 */ }
	| KW_TYPEID LPAREN type_id RPAREN
		{ UNSUPPORTED /* 9 */ }
	;

newcast:
	  KW_DCAST
	| KW_SCAST
	| KW_RCAST
	| KW_CCAST
	;

// expression-list:
//	assignment-expression
//	expression-list , assignment-expression

expression_list:
	  assignment_expression
		{ do2($$, BuildExprList, none, $1); }
	| expression_list COMMA assignment_expression
		{ do2($$, BuildExprList, $1, $3); }
	;

expression_list_opt: { do2($$, BuildExprList, none, none); } | expression_list;

// pseudo-destructor-name:
//	::(opt) nested-name-specifier(opt) type-name :: ~ type-name
//	::(opt) nested-name-specifier(opt) ~ type-name

pseudo_destructor_name:
	  DTOR_ID
		{ do3($$, SpecialId, T(DTOR_ID), $1, curscope); }
/*
	  type_name SCOPE OP_NOT type_name
	| OP_NOT type_name
*/
	;

// unary-expression:
//	postfix-expression
//	++ cast-expression
//	-- cast-expression
//	unary-operator cast-expression
//	"sizeof" unary-expression
//	"sizeof" ( type-id )
//	new-expression
//	delete-expression

unary_expression:
	  postfix_expression
	| OP_INC cast_expression
		{ do2($$, UnaryOp, $1, $2); }
	| OP_DEC cast_expression
		{ do2($$, UnaryOp, $1, $2); }
	| unary_operator cast_expression
		{ do2($$, UnaryOp, $1, $2); }
	| OP_SIZEOF unary_expression
		{ do2($$, UnaryOp, $1, $2); }
	| OP_SIZEOF LPAREN type_id RPAREN
		{ do2($$, UnaryOp, T(OP_TSIZEOF), $3); }
	| new_expression
	| delete_expression
	;

// unary-operator: one of
//	* & + - ! ~

unary_operator:
	  OP_MUL	{ $$ = T(OP_PTR); }
	| OP_AND	{ $$ = T(OP_REF); }
	| OP_ADD	{ $$ = T(OP_UPLUS); }
	| OP_SUB	{ $$ = T(OP_UMINUS); }
	| OP_LNOT
	| OP_NOT
	;

// new-expression:
//	::(opt) "new" new-placement(opt) new-type-id new-initializer(opt)
//	::(opt) "new" new-placement(opt) ( type-id ) new-initializer(opt)

new_expression:
	  OP_NEW new_placement_opt new_type_id new_initializer_opt
		{ do3($$, OpNew, $3, $4, $2); }
	| OP_NEW new_placement_opt LPAREN type_id RPAREN new_initializer_opt
		{ do3($$, OpNew, $4, $6, $2); }
	;

// new-placement:
//	( expression-list )

new_placement:
	  LPAREN expression_list RPAREN
		{ $$ = $2; }
	;

new_placement_opt: none | new_placement;

// new-type-id:
//	type-specifier-seq new-declarator(opt)

new_type_id:
	  type_specifier_seq new_declarator_opt
		{ eval_decl($$, $2, $1); }
	;

// new-declarator:
//	ptr-operator new-declarator(opt)
//	direct-new-declarator

new_declarator:
	  ptr_operator new_declarator_opt
		{ declPtr($$, $1, $2); }
	| direct_new_declarator
	;

new_declarator_opt: none | new_declarator;

// direct-new-declarator:
//	[ expression ]
//	direct-new-declarator [ constant-expression ]

direct_new_declarator:
	  LBRACK expression RBRACK
		{ declArray($$, none, $2); }
	| direct_new_declarator LBRACK constant_expression RBRACK
		{ declArray($$, $1, $3); }
	;

// new-initializer:
//	( expression-list(opt) )

new_initializer:
	  LPAREN expression_list_opt RPAREN
		{ $$ = $2; }
	;

new_initializer_opt: none | new_initializer;


// delete-expression:
//	::(opt) "delete" cast-expression
//	::(opt) "delete" [ ] cast-expression

delete_expression:
	  OP_DELETE cast_expression
		{ do2($$, UnaryOp, $1, $2); }
	| OP_DELETEA cast_expression
		{ do2($$, UnaryOp, $1, $2); }
	;

// cast-expression:
//	unary-expression
//	( type-id ) cast-expression
//   C/C++ extension:
//	( type-id ) { initializer_list(opt) ,(opt) }

cast_expression:
	  unary_expression
	| LPAREN type_id RPAREN cast_expression
		{ do3($$, Cast, $4, $2, none); }
	| LPAREN type_id RPAREN lbrace
	   initializer_list_opt comma_opt RBRACE %uniq
		{ do3($$, Cast, $5, $2, none); }
	;

// pm-expression:
//	cast-expression
//	pm-expression .* cast-expression
//	pm-expression ->* cast-expression

pm_expression:
	  cast_expression
	| pm_expression OP_MDOT cast_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	| pm_expression OP_MPTRTO cast_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	;

// multiplicative-expression:
//	pm-expression
//	multiplicative-expression * pm-expression
//	multiplicative-expression / pm-expression
//	multiplicative-expression % pm-expression

multiplicative_expression:
	  pm_expression
	| multiplicative_expression OP_MUL pm_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	| multiplicative_expression OP_DIV pm_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	| multiplicative_expression OP_MOD pm_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	;

// additive-expression:
//	multiplicative-expression
//	additive-expression + multiplicative-expression
//	additive-expression - multiplicative-expression

additive_expression:
	  multiplicative_expression
	| additive_expression OP_ADD multiplicative_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	| additive_expression OP_SUB multiplicative_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	;
 
// shift-expression:
//	additive-expression
//	shift-expression << additive-expression
//	shift-expression >> additive-expression

shift_expression:
	  additive_expression
	| shift_expression OP_LSHIFT additive_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	| shift_expression OP_RSHIFT additive_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	;
	
// relational-expression:
//	shift-expression
//	relational-expression < shift-expression
//	relational-expression > shift-expression
//	relational-expression <= shift-expression
//	relational-expression >= shift-expression

relational_expression:
	  shift_expression
	| relational_expression OP_LSS shift_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	| relational_expression OP_GTR shift_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	| relational_expression OP_LEQ shift_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	| relational_expression OP_GEQ shift_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	;

// equality-expression:
//	relational-expression
//	equality-expression == relational-expression
//	equality-expression != relational-expression

equality_expression:
	  relational_expression
	| equality_expression OP_EQU relational_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	| equality_expression OP_NEQ relational_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	;

// and-expression:
//	equality-expression
//	and-expression & equality-expression

and_expression:
	  equality_expression
	| and_expression OP_AND equality_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	;

// exclusive-or-expression:
//	and-expression
//	exclusive-or-expression ^ and-expression


exclusive_or_expression:
	  and_expression
	| exclusive_or_expression OP_XOR and_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	;

// inclusive-or-expression:
//	exclusive-or-expression
//	inclusive-or-expression | exclusive-or-expression

inclusive_or_expression:
	  exclusive_or_expression
	| inclusive_or_expression OP_OR exclusive_or_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	;

// logical-and-expression:
//	inclusive-or-expression
//	logical-and-expression && inclusive-or-expression

logical_and_expression:
	  inclusive_or_expression
	| logical_and_expression OP_LAND inclusive_or_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	;

// logical-or-expression:
//	logical-and-expression
//	logical-or-expression || logical-and-expression

logical_or_expression:
	  logical_and_expression
	| logical_or_expression OP_LOR logical_and_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
	;

// conditional-expression:
//	logical-or-expression
//	logical-or-expression ? expression : assignment-expression

conditional_expression:
	  logical_or_expression
	| logical_or_expression QMARK expression colon assignment_expression
		{ do3($$, CondExpr, $1, $3, $5); }
	;

// assignment-expression:
//	conditional-expression
//	logical-or-expression assignment-operator assignment-expression
//	throw-expression

assignment_expression:
	  conditional_expression
	| logical_or_expression assignment_operator assignment_expression
		{ do3($$, BinaryOp, $2, $1, $3); }
//x	| throw_expression
	;

// assignment-operator: one of
//	= *= /= %= += -= >>= <<= &= ^= |=
//    DC++:
//	?=

assignment_operator:
	  OP_ASN
	| OP_ASNMUL
	| OP_ASNDIV
	| OP_ASNMOD
	| OP_ASNADD
	| OP_ASNSUB
	| OP_ASNRSHIFT
	| OP_ASNLSHIFT
	| OP_ASNAND
	| OP_ASNXOR
	| OP_ASNOR
	| OP_CONDASN
	;

// expression:
//	assignment-expression
//	expression , assignment-expression

expression:
	  assignment_expression
	| expression COMMA assignment_expression
		{ do3($$, BinaryOp, T(OP_COMMA), $1, $3); }
	;

expression_opt: none | expression;

// constant-expression:
//	conditional-expression

constant_expression:
	  conditional_expression
	;


//////////////////////////////////////////////////////////////////////////////
// STATEMENTS (A.6)

// statement:
//	labeled-statement
//	expression-statement
//	compound-statement
//	selection-statement
//	iteration-statement
//	jump-statement
//	declaration-statement
//	try-block

// note: this got a little messy to avoid the dangling-else problem
// once shift/reduce precedence is added to shacc, this can be cleaned up

statement:
	  matched_statement
	| unmatched_statement
	;

other_statement:
	  labeled_statement
	| expression_statement
	| compound_statement
	| selection_statement
	| iteration_statement
	| jump_statement
	| declaration_statement
//x	| try_block
	;

// selection-statement:
//	"if" ( condition ) statement
//	"if" ( condition ) statement "else" statement
//	"switch" ( condition ) statement

ifsavescope:
	  KW_IF LPAREN
		{ savescope_newlocal($$); }
	;

restsc:	/* restore scope */
		{ restorescope($-3); }
	;

matched_statement:
	  other_statement
	| ifsavescope condition RPAREN matched_statement restsc 
	     KW_ELSE matched_statement
		{ do3($$, IfElse, $2, $4, $7); }
	;

unmatched_statement:
	  ifsavescope condition RPAREN statement restsc
		{ do3($$, IfElse, $2, $4, none); }
	| ifsavescope condition RPAREN matched_statement restsc
	     KW_ELSE unmatched_statement
		{ do3($$, IfElse, $2, $4, $7); }
	;

selection_statement:
	  KW_SWITCH LPAREN 
		{ savescope_newlocal($2); }
	     condition RPAREN statement
		{ do2($$, Switch, $4, $6);
		  restorescope($2); }
	;

// labeled-statement:
//	identifier : statement
//	"case" constant-expression : statement
//	"default" : statement

labeled_statement:
	  identifier colon statement
		{ NOSCOPE($1); do3($$, Label, none, $1, $3); }
	| KW_CASE constant_expression colon statement
		{ do3($$, Label, $1, $2, $4); }
	| KW_DEFAULT colon statement
		{ do3($$, Label, $1, none, $3); }
	;

// expression-statement:
//	expression(opt) ;

expression_statement:
	  expression_opt EOS %uniq
	;

// compound-statement:
//	{ statement-seq(opt) }

compound_statement:
	  LBRACE %uniq
		{ savescope_newlocal($$);
		  $1 = _stmtlist;
		  do3(_stmtlist, BuildStmtList, none, none, curscope);
		  _stmtlist.scope = curscope.scope; }
	    statement_seq_opt RBRACE %uniq
		{ $$ = _stmtlist; _stmtlist = $1; 
		  restorescope($2); }
	;

// statement-seq:
//	statement
//	statement-seq statement

statement_seq_opt:	// optional
	| statement_seq_opt statement
		{ do3(_stmtlist, BuildStmtList, _stmtlist, $2, curscope); }
	;

// condition:
//	expression
//	type-specifier-seq declarator = assignment-expression
//    DC++:
//	type-specifier-seq declarator ?= assignment-expression

condition:
	  expression
	| type_specifier_seq declarator OP_ASN assignment_expression
		{ 
		  defineSym($3, $2, $4); 
		  eval_decl($2, $3, $1);
		  do3($$, BuildStmtList, none, none, curscope);
		  $$.scope = curscope.scope; 
		  newdecl(&$$, $2);
		}
	| type_specifier_seq declarator OP_CONDASN assignment_expression
		{ 
		  defineSym($3, $2, $4); 
		  eval_decl($2, $3, $1);
		  do3($$, BuildStmtList, none, none, curscope);
		  $$.scope = curscope.scope; 
		  newdecl(&$$, $2);
		}
	;

condition_opt: none | condition;

// iteration-statement:
//	"while" ( condition ) statement
//	"do" statement "while" ( expression ) ;
//	"for" ( for-init-statement condition(opt) ; expression(opt) ) statement

iteration_statement:
	  KW_WHILE LPAREN 
		{ savescope_newlocal($2); }
	     condition RPAREN statement
		{ do3($$, While, $4, $6, none);
		  restorescope($2); }
	| KW_DO statement KW_WHILE LPAREN expression RPAREN EOS %uniq
		{ do3($$, DoWhile, none, $2, $5); }
	| KW_FOR LPAREN
		{ savescope_newlocal($2); }
	     for_init_statement condition_opt EOS expression_opt RPAREN statement 
		{ do3($1, ForExpr, $4, $5, $7);
		  do2($$, For, $1, $9);
		  restorescope($2); }
	;

// for-init-statement:
//	expression-statement
//	simple-declaration

for_init_statement:
	  expression_statement
	| 
		{ $$ = _stmtlist;
	    	  do3(_stmtlist, BuildStmtList, none, none, curscope);
		  _stmtlist.scope = curscope.scope; } // $1
	   simple_declaration %uniq
		{ $$ = _stmtlist; _stmtlist = $1; }
	;

// jump-statement:
//	"break" ;
//	"continue" ;
//	"return" expression(opt) ;
//	"goto" identifier ;

jump_statement:
	  KW_BREAK EOS %uniq
		{ do1($$, Break, $1); }
	| KW_CONTINUE EOS %uniq
		{ do1($$, Continue, $1); }
	| KW_RETURN expression_opt EOS %uniq
		{ do1($$, Return, $2); }
	| KW_GOTO identifier EOS %uniq
		{ NOSCOPE($2); do2($$, Goto, none, $2); }
	;

// declaration-statement:
//	block-declaration

declaration_statement:
	  block_declaration
		{ $$ = none; }
	;

//////////////////////////////////////////////////////////////////////////////
// KEYWORDS (A.2)

// typedef-name:
//	identifier

typedef_name:
	  TYPE_ID
	;

// namespace-name:
//	original-namespace-name
//	namespace-alias

namespace_name:
	  original_namespace_name
	| namespace_alias
	;

// original-namespace-name:
//	identifier

original_namespace_name:
	  NSPACE_ID
	;

// namespace-alias:
//	identifier

namespace_alias:
	  ASPACE_ID
	;

// class-name:
//	identifier
//	template-id

class_name:
	  AGGR_ID
//x	| CTOR_ID
//x		{ $$.token = T(AGGR_ID); }
//x	| template_id
	;

// enum-name:
//	identifier

enum_name:
	  ENUM_ID
	;

// template-name:
//	identifier

template_name:
	  TPLATE_ID
	;


identifier:
	  ID
	| VAL_ID
	| TYPE_ID
	| ENUM_ID
	| AGGR_ID
//x	| CTOR_ID
//x		{ $$.token = T(AGGR_ID); }
	| NSPACE_ID
	| ASPACE_ID
	| TPLATE_ID
	;

identifier_opt: none | identifier ;


//////////////////////////////////////////////////////////////////////////////
// misc helpful rules:

// for handling optional parts:
none: { $$ = none; };

//////////////////////////////////////////////////////////////////////////////
// save up expressions/compound statements for processing later

lbrace: LBRACE | LBRACE_SAVE { $$.token = T(LBRACE); } ;

colon: COLON | COLON_SAVE { $$.token = T(COLON); } ;

ignore_savecode: /* optional */ | TOKENLIST;

save_compound_statement:
		{ $$.Bool = _rawid; _rawid = true; }
	  LBRACE_SAVE save_statement_seq_opt RBRACE %uniq
		{ _rawid = $1.Bool;
		  $2.token = T(LBRACE);
		  $3.TokenList->prepend($2); 
		  $3.TokenList->append($4);
		  $$ = $3; }
	;

save_statement_seq:
	  save_statement_seq_opt any_statement_token
		{ $$.TokenList->append($2); }
	| save_statement_seq_opt save_statement_list
		{ $$.TokenList->append_delete($2.TokenList); }
	;

save_statement_seq_opt:
		{ $$.token = T(TOKENLIST); $$.TokenList = new TokenList; }
	| save_statement_seq
	;

save_statement_list:
	  lbrace save_statement_seq_opt RBRACE %uniq
		{ $2.TokenList->prepend($1); 
		  $2.TokenList->append($3);
		  $$ = $2; }
	| LPAREN save_statement_seq_opt RPAREN
		{ $2.TokenList->prepend($1); 
		  $2.TokenList->append($3);
		  $$ = $2; }
	| LBRACK save_statement_seq_opt RBRACK
		{ $2.TokenList->prepend($1); 
		  $2.TokenList->append($3);
		  $$ = $2; }
	;

any_statement_token:
	  NUM | CHAR | STR | FLOAT | BOOL
	| EOS | colon | QMARK 
	| simple_operator
	| OP_DOT | OP_MDOT
	| OP_SCOPE | ELIPSIS
	| cv_qualifier
	| simple_type_specifier_notypename
	| storage_class_specifier
	| KW_IF | KW_ELSE | KW_DO | KW_WHILE | KW_FOR
	| KW_SWITCH | KW_CASE | KW_DEFAULT
	| KW_BREAK | KW_CONTINUE | KW_GOTO
	| KW_RETURN
	| KW_ENUM | KW_TYPEDEF
	| class_key
	| function_specifier
	| OP_SIZEOF | OP_TYPEOF
	| KW_ASM
	| KW_OPERATOR | KW_THIS
	| KW_TYPEID | KW_TYPENAME
	| KW_NAMESPACE | KW_USING
	| KW_TEMPLATE
	| access_specifier
	| KW_FRIEND
	| KW_THROW | KW_CATCH | KW_TRY
	| newcast
	| identifier
		{ $$.token = T(ID); }
	| DTOR_ID
	;


//////////////////////////////////////////////////////////////////////////////
// PREPROCESSING DIRECTIVES (A.11) [ the book says A.10.1 but
//	    this seems odd, since it's not a subsection of Exceptions


//////////////////////////////////////////////////////////////////////////////
// Attributes (GNU-ish extension)

attribute:
	  KW_ATTRIBUTE LPAREN LPAREN save_statement_seq RPAREN RPAREN
		{ /* just ignore it for now */ }
	;

attribute_opt: none | attribute ;

%%//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

