/*  $Id: CRep.hh,v 1.15 1998/08/22 03:35:16 shaggy Exp $
 *      C++ internals (knowledge representation and manipulation)
 *      
 *      relative path: .//lib/cparse
 *  
 *      Shag/OS: an Object-Oriented MicroKernel Operating System
 *       Libraries
 *        C/C++ Parsing classes
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the Shag/OS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare Non-profit Public License (SNPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare Non-profit Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

//  $Log: CRep.hh,v $
//  Revision 1.15  1998/08/22 03:35:16  shaggy
//  header update
//
//  Revision 1.14  1998/08/12 15:18:35  shaggy
//  better typing; fixed some type coercions;
//  cleanup for #include's, and fixed some constructor call bugs
//
//  Revision 1.13  1998/08/04 00:45:35  shaggy
//  lots of fixes so it builds under Solaris again
//
//  Revision 1.12  1998/08/03 15:00:13  shaggy
//  conversion operators, constructor conversions, default params
//
//  Revision 1.11  1998/08/03 04:11:33  shaggy
//  deco can *finally* generate code which binds a type to a class
//
//  Revision 1.10  1998/08/03 00:12:58  shaggy
//  tons of work-- friends, class scope fixes, tons of bugfixes,
//  lots more cleanup, more DC++ work, etc
//
//  Revision 1.9  1998/08/02 04:29:39  shaggy
//  more fixes
//
//  Revision 1.8  1998/08/01 21:02:07  shaggy
//  much more cleanup; many bugfixes; DC++ progress
//
//  Revision 1.7  1998/08/01 14:56:47  shaggy
//  about to restructure deco
//
//  Revision 1.6  1998/07/31 23:44:27  shaggy
//  running out of time
//
//  Revision 1.5  1998/07/30 16:16:30  shaggy
//  switched to new vsn of shacc and removed useless underscore prefixes
//  as well as lot of extra configuration code that is no longer needed
//
//  Revision 1.4  1998/07/30 04:40:32  shaggy
//  much more dtype and dclass work
//
//  Revision 1.3  1998/07/29 03:35:12  shaggy
//  dtype support nearly complete, initial dclass support
//
//  Revision 1.2  1998/07/28 19:55:26  shaggy
//  moved most of generic parsing from Deco into CAction
//
//  Revision 1.1  1998/07/28 16:43:20  shaggy
//  moved lots of Deco into generic CParse lib
//  CAction split into vCAction and CAction
//  printaction -> part of CAction
//  ccint -> CRep (C Representation)
//  ccout -> COut (C output)
//
//  Revision 1.57  1998/07/28 01:24:22  shaggy
//  initial support for implementation specifiers
//
//  Revision 1.56  1998/07/26 17:34:44  shaggy
//  deco 0.8 release
//
//  Revision 1.55  1998/07/21 14:57:27  shaggy
//  trying to get dtype conversion working right
//
//  Revision 1.54  1998/07/20 17:51:46  shaggy
//  added bitfield support
//
//  Revision 1.53  1998/07/20 14:59:35  shaggy
//  more dtype conv work
//
//  Revision 1.52  1998/07/19 22:33:35  shaggy
//  part 2 of major overhaul (needed for type-safty and to start
//  being more consistent)
//
//  Revision 1.51  1998/07/19 06:46:46  shaggy
//  1st stage of Item cleanup
//
//  Revision 1.50  1998/07/11 23:06:49  shaggy
//  fixed labels and goto
//
//  Revision 1.49  1998/07/11 19:03:00  shaggy
//  declarations in for() work now (even multiple decls)
//
//  Revision 1.48  1998/07/08 13:42:42  shaggy
//  started adding proper function calls
//
//  Revision 1.47  1998/07/07 13:41:41  shaggy
//  enum cleanup, and can define types inside functions now
//
//  Revision 1.46  1998/07/06 22:49:42  shaggy
//  reworked symtab, output ordering, etc
//
//  Revision 1.45  1998/07/06 13:39:06  shaggy
//  many more changes/fixes for first pre-release
//
//  Revision 1.44  1998/07/05 02:47:38  shaggy
//  tons of bug fixes
//
//  Revision 1.43  1998/07/04 20:38:59  shaggy
//  lots of bug fixes; fn, aggr, etc improvements, new/delete
//
//  Revision 1.42  1998/07/04 01:23:30  shaggy
//  anonymous unions work now
//
//  Revision 1.41  1998/07/03 19:06:25  shaggy
//  changed scope semantics; fn name is available in scope now
//
//  Revision 1.40  1998/07/03 16:53:52  shaggy
//  new Scope wrapper class, and dcast_XXX functions eliminate lots of casts
//
//  Revision 1.39  1998/07/02 13:25:51  shaggy
//  public/private/protected works for inheritance
//
//  Revision 1.38  1998/07/01 13:14:32  shaggy
//  handles default code for dtype member fns
//
//  Revision 1.37  1998/06/30 15:38:37  shaggy
//  better support for member fns and member fn ptrs
//
//  Revision 1.36  1998/06/30 13:23:32  shaggy
//  fcall fixes for d++ conv
//
//  Revision 1.35  1998/06/30 12:22:32  shaggy
//  many fixes for "this"
//
//  Revision 1.34  1998/06/29 01:36:54  shaggy
//  fixed inline attr, output of pure virtuals, etc...
//
//  Revision 1.33  1998/06/07 18:09:04  shaggy
//  added types to Expr()'s, and fixed output of members (. and ->)
//
//  Revision 1.32  1998/06/02 14:50:42  shaggy
//  in the midst of more full parsing updates
//
//  Revision 1.31  1998/05/26 13:40:14  shaggy
//  many more fixes
//
//  Revision 1.30  1998/05/26 05:02:54  shaggy
//  parameter scoping works correctly
//  most basic C/C++ statements now work
//
//  Revision 1.29  1998/05/25 23:28:44  shaggy
//  major rework of func/param info
//
//  Revision 1.28  1998/05/25 20:08:41  shaggy
//  more bugfixes
//
//  Revision 1.27  1998/05/25 04:22:13  shaggy
//  enough for tonight
//
//  Revision 1.26  1998/05/24 23:39:33  shaggy
//  various bug fixes and code cleanup
//
//  Revision 1.25  1998/05/24 01:41:41  shaggy
//  internal support for most statements
//
//  Revision 1.24  1998/05/20 14:41:24  shaggy
//  better code output
//
//  Revision 1.23  1998/05/19 14:55:36  shaggy
//  a bit closer to generating code (stmtlists work)
//
//  Revision 1.22  1998/05/18 13:22:36  shaggy
//  more decl changes
//
//  Revision 1.21  1998/05/18 05:24:27  shaggy
//  more cleanup of declaration support functions
//
//  Revision 1.20  1998/05/17 23:14:57  shaggy
//  expressions, expression lists, etc
//
//  Revision 1.19  1998/05/16 23:30:43  shaggy
//  added CTOR_ID and some rules to help fix ctor ambiguity problem
//  output turned off for #include'ed files
//  better error checking, and ability to list locations
//  of symbol declaration and definition
//
//  Revision 1.18  1998/05/15 07:03:37  shaggy
//  dependency ordering fixed; constructor initializers; lots of bug fixes, etc
//
//  Revision 1.17  1998/05/14 05:08:41  shaggy
//  inheritence works and lots of bugs fixed
//
//  Revision 1.16  1998/05/13 21:39:36  shaggy
//  switched to dtype/ptype/rtype
//
//  Revision 1.15  1998/05/13 21:02:48  shaggy
//  deinitions/initializations work better now
//
//  Revision 1.14  1998/05/12 05:00:14  shaggy
//  more scope work
//
//  Revision 1.13  1998/05/10 18:53:41  shaggy
//  just about ready to start fixing scoping
//
//  Revision 1.12  1998/05/08 13:03:23  shaggy
//  fixed const/volatile
//
//  Revision 1.11  1998/05/07 14:46:40  shaggy
//  const/volatile partially working
//
//  Revision 1.10  1998/05/04 04:39:49  shaggy
//  first stage of function overloading
//
//  Revision 1.9  1998/05/04 03:09:20  shaggy
//  support for ctors, dtors, conversion operators, operator overloading, etc
//
//  Revision 1.8  1998/04/23 01:32:47  shaggy
//  more Solaris changes
//
//  Revision 1.7  1998/04/22 12:17:51  shaggy
//  changed ID to Id for consistency
//
//  Revision 1.6  1998/04/16 13:35:33  shaggy
//  access control
//
//  Revision 1.5  1998/04/15 14:47:58  shaggy
//  array output works
//
//  Revision 1.4  1998/04/15 13:39:56  shaggy
//  added types to constants
//
//  Revision 1.3  1998/04/14 03:57:09  shaggy
//  forward reference support
//
//  Revision 1.2  1998/04/06 14:27:43  shaggy
//  lots more changes, and added some tests
//
//  Revision 1.1  1998/02/08 23:41:58  shaggy
//  Initial revision
//

#include <CObj>
#include <CLex>
#include <SymTab>
#include <StrTab>
#include <vCAction>

#define T(x)	vCAction::x

struct Decl;
class TokenList;

class Scope {	// more like a ScopePtr or ScopeRef actually...
	vSymTab *_sym;
public:
	typedef vSymTab::Id Id;
	Scope() { _sym = 0; }
	Scope(vSymTab *s) { _sym = s; }
	Scope(const Scope &s) { _sym = s._sym; }
	const Scope& operator= (vSymTab *s) const
		{ const_cast<vSymTab*&>(_sym) = s; return *this; }
	const Scope& operator= (Scope s) const
		{ const_cast<vSymTab*&>(_sym) = s._sym; return *this; }
	operator vSymTab* () const { return _sym; }
	bool operator== (Scope s) const { return _sym == s._sym; }
	bool operator!= (Scope s) const { return _sym != s._sym; }
	bool operator== (vSymTab *s) const { return _sym == s; }
	bool operator!= (vSymTab *s) const { return _sym != s; }
	Decl* get(Id id) const
		{ return reinterpret_cast<Decl*>(_sym->get(id)); }
	bool set(Id id, Decl *d) { return _sym->set(id, d); }
	bool add(Id id, Decl *d) { return _sym->add(id, d); }
	bool del(Id id) { return _sym->del(id); }
	Id index(uint n) const { return _sym->index(n); }
	Scope parent() const { return Scope(_sym->parent()); }
	bool parent(Scope s) { return _sym->parent(s); }
	Decl* decl() const
		{ return reinterpret_cast<Decl*>(_sym->data()); }
	bool decl(Decl *d)
		{ return _sym->data(d); }
	Id pchid() const;
	vStrTab* strtab() const { return _sym->strtab(); }
	operator bool () const { return _sym != 0; }

	// untagged lookup:
	inline Decl* uget(vStrTab::Id id) const;
	inline Decl* ufind(vStrTab::Id id) const;

	// tagged lookup:
	inline Decl* tget(vStrTab::Id id) const;
	inline Decl* tget(vStrTab::Id id, bool t) const;
	inline Decl* tfind(vStrTab::Id id) const;

	// selectable tagged lookup:
	inline Decl* tfind(vStrTab::Id id, bool t) const;
	inline Decl* tfind(vStrTab::Id id, bool t, bool s) const;
};

// templates could be used to do this generically, but then
// deco wouldn't be able to parse itself, so until deco has
// template support, none of the code shall use templates
class ScopeList {
	struct Node {
		vSymTab *sym;
		Node *next;
	};
	Node *_first;
	Node *_last;
public:
	ScopeList() { _first = _last = 0; }
	ScopeList(Node *first, Node *last) { _first = first; _last = last; }
	Scope head() const { return Scope(_first ? _first->sym : 0); }
	ScopeList tail() const { return _first ? ScopeList(_first->next, _last)
				 : ScopeList(0, 0); }
	void prepend(Scope s) { Node *n = new Node; n->sym = s;
				n->next = _first;
				if(!_last) _last = n;
				_first = n; }
	void append(Scope s) { Node *n = new Node; n->sym = s;
				n->next = 0;
				if(_last) _last->next = n; 
				else _first = n;
				_last = n; }
	ScopeList& operator += (Scope s) { append(s); return *this; }
};

// currently unused:
class DeclList {
	struct Node {
		Decl *decl;
		Node *next;
	};
	Node *_first;
	Node *_last;
public:
	DeclList() { _first = _last = 0; }
	DeclList(Node *first, Node *last) { _first = first; _last = last; }
	Decl* head() const { return _first ? _first->decl : 0; }
	DeclList tail() const { return _first ? DeclList(_first->next, _last)
				 : DeclList(0, 0); }
	void prepend(Decl *d) { Node *n = new Node; n->decl = d;
				n->next = _first;
				if(!_last) _last = n;
				_first = n; }
	void append(Decl *d) { Node *n = new Node; n->decl = d;
				n->next = 0;
				if(_last) _last->next = n; 
				else _first = n;
				_last = n; }
	DeclList& operator += (Decl *d) { append(d); return *this; }
};


struct Litstruct {
	type_t	type;
	union {
		uint64 val;
		void *ptr;
	};
};

class Lit : public Litstruct {
public:
	Lit(type_t t, uint64 v) { type = t; val = v; }
	Lit(type_t t, int v) { type = t; val = v; }
	Lit(type_t t, void *p) { type = t; ptr = p; }
	bool operator==(const Lit &n) const
		{ return type == n.type && val == n.val; }
	bool operator!=(const Lit &n) const { return !(*this == n); }
};

struct ExprNode {
	vCAction::Token op;
	union {
		struct {
			ExprNode *left;
			union {
				ExprNode *mid;
				vSymTab *sc_mid;
				vCAction::Token tok_mid;
			};
			ExprNode *right;
		} e;
		Litstruct lit;
		Decl *decl;
		TokenList *tlist;
		// vSymTab *scope;
		vStrTab::Id id;
	};
	type_t rtype;
	ExprNode *next;
};

class Expr {
	ExprNode *p;
	friend class ExprList;
	friend class Stmt;
public:
	Expr() { p = 0; }
	Expr(ExprNode *np) { p = np; }
	Expr(const Lit &);
	Expr(Decl *);
	Expr(type_t);
	Expr(TokenList *);
	// Expr(const Scope scope);
	Expr(vStrTab::Id);
	Expr(vCAction::Token tok, Expr e);
	Expr(vCAction::Token tok, Expr e1, Expr e2);
	Expr(vCAction::Token tok, Expr e1, Expr e2, Expr e3);
	Expr& operator+= (Expr e1)
		{ p->next = e1.p; return *this; }
	Expr operator+ (Expr e1) const
		{ Expr e(*this); e.p->next = e1.p; return e; }
	bool operator== (Expr) const;
	bool operator!= (Expr e) const { return !(*this == e); }
	bool is(Expr e) const { return p == e.p; }
	Expr& set_rtype(type_t t) { p->rtype = t; return *this; }
	vCAction::Token op() const { return p->op; }
	const Lit& lit() const { return (Lit&)p->lit; }
	Decl *decl() const { return p->decl; }
	TokenList *tlist() const { return p->tlist; }
	vStrTab::Id id() const { return p->id; }
	const Expr left() const { return p->e.left; }
	const Expr mid() const { return p->e.mid; }
	const Expr right() const { return p->e.right; }
	type_t rtype() const { return p->rtype; }
	~Expr() {  }
	operator ExprNode*() const { return p; }
	operator bool () const { return p != 0; }
	Expr next() const { return Expr(p->next); }
	Expr& operator++() { p = p->next; return *this; }
	Expr operator++(int) { Expr e = *this; p = p->next; return e; }
	Expr dup() const;
};

static const Expr null_expr = Expr((ExprNode*)0);

class ExprList {
	ExprNode *p;
public:
	ExprList(vCAction::Token tok = T(EXPRLIST));
	ExprList(vCAction::Token tok, vCAction::Token fmt);
	ExprList(ExprNode *np) { p = np; }
	ExprList& append(Expr e);
	ExprList& operator += (Expr e) { return append(e); }
	operator bool () const { return p != 0; }
	bool empty() const { return !p->e.left; }
	Expr first() const { return Expr(p->e.left); }
	Expr last() const { return Expr(p->e.right); }
	vCAction::Token fmt() const { return p->e.tok_mid; }
	int len() const;
	operator ExprNode*() const { return p; }
};

class Stmt {
	ExprNode *p;
	friend class StmtList;
public:
	Stmt(ExprNode *np) { p = np; }
	Stmt(const Lit &);
	Stmt(Decl *);
	Stmt(TokenList *);
	Stmt(vCAction::Token tok, Expr e);
	Stmt(vCAction::Token tok, Expr e1, Expr e2);
	Stmt(vCAction::Token tok, Expr e1, Expr e2, Expr e3);
	vCAction::Token op() const { return p->op; }
	const Lit& lit() const { return (Lit&)p->lit; }
	Decl *decl() const { return p->decl; }
	TokenList *tlist() const { return p->tlist; }
	const Expr left() const { return p->e.left; }
	const Expr mid() const { return p->e.mid; }
	const Expr right() const { return p->e.right; }
	~Stmt() {  }
	operator ExprNode*() const { return p; }
	operator bool () const { return p != 0; }
	Stmt next() const { return Stmt(p->next); }
	Stmt& operator++() { p = p->next; return *this; }
	Stmt operator++(int) { Stmt e = *this; p = p->next; return e; }
};

static const Stmt null_stmt = Stmt((ExprNode*)0);

class StmtList {
	ExprNode *p;
public:
	StmtList(vCAction::Token tok = T(STMTLIST), Scope scope = 0);
	StmtList(Scope scope, vCAction::Token tok = T(STMTLIST));
	StmtList(ExprNode *np) { p = np; }
	StmtList& append(Stmt e);
	StmtList& operator += (Stmt e) { return append(e); }
	operator bool () const { return p != 0; }
	Scope scope() const { return p->e.sc_mid; }
	bool empty() const { return !p->e.left; }
	Stmt first() const { return Stmt(p->e.left); }
	Stmt last() const { return Stmt(p->e.right); }
	operator ExprNode*() const { return p; }
};

class StmtIf : public Stmt { public: StmtIf(Expr c, Stmt s)
		: Stmt(T(KW_IF), c, Expr(s), null_expr) {} };
class StmtIfElse : public Stmt { public: StmtIfElse(Expr c, Stmt s, Stmt e)
		: Stmt(T(KW_IF), c, Expr(s), Expr(e)) {} };
class StmtWhile : public Stmt { public: StmtWhile(Expr c, Stmt s)
		: Stmt(T(KW_WHILE), c, Expr(s), null_expr) {} };
class StmtDoWhile : public Stmt { public: StmtDoWhile(Stmt s, Expr c)
		: Stmt(T(KW_DO), null_expr, Expr(s), c) {} };
class ExprFor : public Expr { public: ExprFor(Stmt i, Expr c, Expr n)
		: Expr(T(KW_FOR), Expr(i), c, n) {} };
class StmtFor : public Stmt { public: StmtFor(Stmt i, Expr c, Expr n, Stmt s)
		: Stmt(T(KW_FOR), ExprFor(i, c, n), Expr(s), null_expr) {} 
		StmtFor(Expr fe, Stmt s)
		: Stmt(T(KW_FOR), fe, Expr(s), null_expr) {} };
class StmtSwitch : public Stmt { public: StmtSwitch(Expr v, Stmt s)
		: Stmt(T(KW_SWITCH), v, Expr(s), null_expr) {} };
class StmtReturn : public Stmt { public: StmtReturn(Expr r)
		: Stmt(T(KW_RETURN), r, null_expr, null_expr) {} };
class StmtGoto : public Stmt { public: StmtGoto(Expr w)
		: Stmt(T(KW_GOTO), w, null_expr, null_expr) {} };
class StmtBreak : public Stmt { public: StmtBreak()
		: Stmt(T(KW_BREAK), null_expr, null_expr, null_expr) {} };
class StmtContinue : public Stmt { public: StmtContinue()
		: Stmt(T(KW_CONTINUE), null_expr, null_expr, null_expr) {} };
class StmtAsm : public Stmt { public: StmtAsm(Expr s)
		: Stmt(T(KW_ASM), s, null_expr, null_expr) {} };
class StmtLabel : public Stmt { public: StmtLabel(Expr n)
		: Stmt(T(COLON), n, null_expr, null_expr) {} 
		StmtLabel(Expr n, Stmt s)
		: Stmt(T(COLON), n, Expr(s), null_expr) {} };
class StmtCase : public Stmt { public: StmtCase(Expr n)
		: Stmt(T(KW_CASE), n, null_expr, null_expr) {} 
		StmtCase(Expr n, Stmt s)
		: Stmt(T(KW_CASE), n, Expr(s), null_expr) {} };
class StmtDefault : public Stmt { public: StmtDefault()
		: Stmt(T(KW_DEFAULT), null_expr, null_expr, null_expr) {} 
		StmtDefault(Stmt s)
		: Stmt(T(KW_DEFAULT), null_expr, Expr(s), null_expr) {} };
class ExprDecl : public Expr { public: ExprDecl(Decl *d)
		: Expr(T(DECL), Expr(d), null_expr, null_expr) {} };
class StmtDecl : public Stmt { public: StmtDecl(Decl *d)
		: Stmt(T(DECL), Expr(d), null_expr, null_expr) {} };
class ExprCall : public Expr { public: ExprCall(Expr c, ExprList e)
		: Expr(T(OP_FUNC), c, null_expr, Expr(e)) {} };
class StmtCall : public Stmt { public: StmtCall(Expr c, ExprList e)
		: Stmt(T(OP_FUNC), c, null_expr, Expr(e)) {} };
class ExprCtor : public Expr { public: ExprCtor(StmtList c, StmtList s)
		: Expr(T(CTORINIT), Expr(c), Expr(s), null_expr) {} };
class ExprCode : public Expr {
public:
	ExprCode(Scope /*sc*/, StmtList c, StmtList s)
		: Expr(T(CTORINIT), Expr(c), Expr(s), null_expr) {}
};


struct DeclCommon {
	enum StorageClass {
		sc_default = 0,
		sc_auto = 1,
		sc_static = 2,
		sc_extern = 3,
		sc_register = 4,
		sc_typedef = 5,
		sc_mutable = 6,
		sc_constant = 7,
		sc_param = 8,
		sc_friend = 9,
	};
	enum DataType {
		dt_default = 0,
		dt_int = 1,
		dt_char = 2,
		dt_wchar = 3,
		dt_bool = 4,
		dt_void = 5,
		dt_float = 6,
		dt_double = 7,
		dt_enum = 8,
		dt_struct = 9,
		dt_class = 10,
		dt_union = 11,
		dt_namespace = 12,
		dt_enumval = 13,	// constant from an enum
		dt_dtype = 14,		// dynamic type
		dt_dclass = 15,		// dynamic class
		dt_dimpl = 16,		// dynamic implementation
	};
	enum DataTypeFlags {
		DTF_SIGNOK = 0x01,
		DTF_SHORTOK = 0x02,
		DTF_LONGOK = 0x04,
		DTF_SCOPE = 0x08,	// defines a new scope
		DTF_AGGR = 0x10,
		DTF_TAG = 0x20,		// tagged types can co-exist in symtab
	};
	enum Access {
		ac_public = 0,
		ac_protected = 1,
		ac_private = 2,
		ac_unspecified = 3,
	};

	// note: the uchars below used to be bools, but there's this
	// problem with the struct getting aligned to the largest
	// contained alignment.  On Sparcs, bool is 32-bits (and
	// aligned as such), so even though we're only using one bit
	// for each bool, and 16-bits total here, this structure
	// gets rounded up to 32 bits, causing TDecl to be more than 64.
	DataType	dtype:5;
	Access		access:2;
	char 		rsvd:1;
	StorageClass	storage:4;
	uchar		is_inline:1;
	uchar		is_virtual:1;
	uchar		is_explicit:1;
	char		rsvd2:1;

	static csz* scname[];
	static csz* dtname[];
	static csz* acname[];
	static uint8 dtflags[];

	bool tagged() const
		{ return dtflags[dtype] & DTF_TAG; }
	bool dtaggr() const
		{ return dtflags[dtype] & DTF_AGGR; }
	bool dtscope() const
		{ return dtflags[dtype] & DTF_SCOPE; }
	bool dtlongok() const
		{ return dtflags[dtype] & DTF_LONGOK; }
	bool dtshortok() const
		{ return dtflags[dtype] & DTF_SHORTOK; }
	bool dtsignok() const
		{ return dtflags[dtype] & DTF_SIGNOK; }
	Access defacc() const
		{ return dtype == dt_class ? ac_private : ac_public; }
};

struct TDecl : DeclCommon {
	uint		longcount:3;
	bool		is_unsigned:1;
	bool		is_signed:1;
	bool		is_short:1;
	bool		is_const:1;
	bool		is_volatile:1;
	bool		useful:1;
	bool		rsvd:7;
	type_t		ptype;		// printable type (based on definition)

	void clear() { *(uint64*)this = 0; }
};
 
struct Decl : DeclCommon {
	enum OutType {
		OT_NONE=0,	// it's as if it never existed...
		OT_FROMHDR=1,	// came from an include file
		OT_HDR=2,	// destined for an include file
		OT_CODE=3,	// destined for main code body
		OT_REPOS=4,	// belongs in a code repository
	};

	bool		defined:1;
	bool		initialized:1;
	bool		anon:1;
	bool		newtype:1;
	bool		field:1;
	bool		bitfield:1;
	bool		converted:1;	// if true, `next' link points to conversion
	OutType		otype:3;

	// the following flags could be derived from the type:
	// (for that reason, they probably should be changed to predicates)
	bool		is_func:1;
	bool		is_method:1;

	// flags for output dependency ordering:
	bool		odefined:1;
	bool		odeclared:1;	// also used to report duplicates
	bool		oneeddef:1;
	bool		oneeddecl:1;

	type_t		ptype;		// printable type (based on definition)
	type_t		rtype;		// real type
	Scope		scope;
	vStrTab::Id	id;
	int32		where;
	union {
		struct {
			ExprNode	*expr;
		} v;			// for variable values
		struct {
			int64		val;
		} c;			// for simple constant values
		struct {
			Decl		*decl;
		} i;			// for inheritance
		struct {
			int		tsize;	// total size
			Decl		*first;
		} a;			// for aggregates
		struct {
			int		bsize;	// bitsize
		} b;			// for bitfields
		struct {
			vSymTab		*pscope;	// parameter scope
		} f;			// for uninitialized functions
	};
	Decl 		*next;	// next in-order declaration
	Decl 		*link;	// link to other Decl's that share the same id 

	static csz* otname[];

	void clear() { memset(this, 0, sizeof *this); }

	// TDecl& operator=(const TDecl &t)
	//	{ return static_cast<TDecl>(*this) = t; }

	Decl& set_storage(StorageClass sc) { storage = sc; return *this; }
	Decl& set_access(Access ac) { access = ac; return *this; }
	Decl& set_inline() { is_inline = true; return *this; }
	Decl& set_virtual() { is_virtual = true; return *this; }
	Decl& set_val(Expr e)
		{ v.expr = e; 
		  defined = true; initialized = true; return *this; }
	Decl& set_val(StmtList s)
		{ v.expr = Expr(s);
		  defined = true; initialized = true; return *this; }
	Decl& set_val(ExprList e)
		{ v.expr = Expr(e);
		  defined = true; initialized = true; return *this; }
	Decl& set_val(int64 v)
		{ c.val = v; defined = true; initialized = true; return *this; }
	bool is_purevirtual() const
		{ return Expr(v.expr).op() == T(LITERAL); }
	bool is_aggr() const
		{ return rtype && rtype->is_type(Type_Aggr); }
	bool nooutput(int mask)	const
		{ return !((1<<otype) & mask); }
	Decl*	nextfn() const
		{ return !link || link->tagged() ? 0 : link; }
	bool is_ctor() const
		// XXX - utilizing a major hack:
		{ return is_method && scope.strtab()->string(id)[0] == ' '; }
};

// Note: constant values can be represented with Decl's with
// the appropriate type, and a default value (which will never change)


extern obj_t Type_DeclRef;

class DeclRef {
	Decl* _decl;
public:
	DeclRef(Decl *decl) : _decl(decl) {}
	type_t rtype() const { return _decl->rtype; }
	Decl*  decl() const { return _decl; }
	static type_t type() { return &Type_DeclRef; }
};

class obj_DeclRef : public Object {
public:
	obj_DeclRef(Decl *decl) : Object(DeclRef::type(), Inline)
		{ new(&_val)DeclRef(decl); }
	GEN_COPYREFCTOR(obj_DeclRef, DeclRef)
};

GEN_DCAST(DeclRef)


extern obj_t Type_Enum;

class Enum {
	Decl *_decl;
public:
	Enum(Decl *decl) : _decl(decl) {}
	vStrTab::Id name() const { return _decl->id; }
	Decl *decl() const { return _decl; }
	static type_t type() { return &Type_Enum; }
};

class obj_Enum : public Object {
public:
	obj_Enum(Decl *decl) : Object(Enum::type(), Inline)
		{ new(&_val)Enum(decl); }
	GEN_COPYREFCTOR(obj_Enum, Enum)
};

GEN_DCAST(Enum)


#define case_ANY_SCOPE_ID  \
			case AGGR_ID: case ENUM_ID: \
			case NSPACE_ID: case ASPACE_ID

#define case_ANY_TYPE_ID  case TYPE_ID: case AGGR_ID: \
			case ENUM_ID: case TPLATE_ID

#define case_ANY_VAL_ID	case VAL_ID: case CTOR_ID: \
			case DTOR_ID: case CONV_ID: case OPER_ID

#define case_ANY_DECL_ID case_ANY_TYPE_ID: case_ANY_VAL_ID: \
			 case NSPACE_ID: case ASPACE_ID

#define case_ANY_ID	case ID: case_ANY_DECL_ID



/* if this gives an error, then the size isn't 8 */
static const void * const test_sizeof_TDecl = sizeof(TDecl)-8;

Decl* symget(vStrTab::Id, const Scope, bool tagged);
Decl* symfind(vStrTab::Id, const Scope, bool tagged, bool search = true);

inline Decl*
Scope::uget(vStrTab::Id id) const
{ return symget(id, *this, false); }

inline Decl*
Scope::tget(vStrTab::Id id) const 
{ return symget(id, *this, true); }

inline Decl*
Scope::tget(vStrTab::Id id, bool t) const
{ return symget(id, *this, t); }

inline Decl*
Scope::ufind(vStrTab::Id id) const
{ return symfind(id, *this, false); }

inline Decl*
Scope::tfind(vStrTab::Id id) const
{ return symfind(id, *this, true); }

inline Decl*
Scope::tfind(vStrTab::Id id, bool t) const
{ return symfind(id, *this, t); }

inline Decl*
Scope::tfind(vStrTab::Id id, bool t, bool s) const
{ return symfind(id, *this,t,s); }


/* Note: it is assumed that these particular ID's won't occur naturally... */
const vStrTab::Id ID_ANON 	= (vStrTab::Id)-1;
const vStrTab::Id ID_INHERIT 	= (vStrTab::Id)-2;
const vStrTab::Id ID_USING	= (vStrTab::Id)-3;
const vStrTab::Id ID_FRIEND	= (vStrTab::Id)-4;
const vStrTab::Id ID_CONV	= (vStrTab::Id)-5;	// for conversion operators
const vStrTab::Id ID_CTOR	= (vStrTab::Id)-6;	// for conversion constructors

inline bool idreal(vStrTab::Id id) { return (ulong)id < (ulong)ID_CTOR; }

int tokenprec(vCAction::Token t, int assoc);


bool scope_empty(const Scope);
bool scope_derived(const Scope);
void dump_scope(const Scope);
void dump_scopes(ScopeList);

Decl * param_decl(Func *func, int i);

bool subscope(Scope scope, Scope sub);

void decladd(Decl *d);
void declrename(Decl *d, vStrTab::Id id);

// unfortunate arbitrary limits due to fixed size buffers 
// scattered throughout the code...  one of many things that
// needs to be fixed at some point in the future:
#define MAXTYPELEN 256	// (sometimes including id) 
#define MAXIDLEN 128	// (including all scoping)
#define MAXLINELEN 4096	// (for any single logical C++ line)

