/*  $Id: CAction.cc,v 1.19 1998/05/13 00:17:40 shaggy Exp $
 *      standard C/C++ parsing actions
 *      
 *      relative path: .//lib/cparse/save
 *  
 *      Shag/OS: an Object-Oriented MicroKernel Operating System
 *       Libraries
 *        C/C++ Parsing classes
 *         (save)
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the Shag/OS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare Non-profit Public License (SNPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare Non-profit Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

//  $Log: CAction.cc,v $
//  Revision 1.19  1998/05/13 00:17:40  shaggy
//  scoping works and enums fixed (mostly)
//
//  Revision 1.18  1998/05/12 05:00:11  shaggy
//  more scope work
//
//  Revision 1.17  1998/05/10 18:53:39  shaggy
//  just about ready to start fixing scoping
//
//  Revision 1.16  1998/05/10 00:03:29  shaggy
//  string concatenation works, and comments are being
//  added to get ready to fix scoping
//
//  Revision 1.15  1998/05/04 03:09:17  shaggy
//  support for ctors, dtors, conversion operators, operator overloading, etc
//
//  Revision 1.14  1998/05/03 02:47:06  shaggy
//  added inheritence to parser
//
//  Revision 1.13  1998/04/16 13:35:43  shaggy
//  access control
//
//  Revision 1.12  1998/04/15 13:39:42  shaggy
//  added types to constants
//
//  Revision 1.11  1998/04/11 21:49:26  shaggy
//  added token saving (for regurgitation)
//
//  Revision 1.10  1998/04/06 14:28:06  shaggy
//  got enums almost working
//
//  Revision 1.9  1998/03/22 03:14:55  shaggy
//  almost there...
//
//  Revision 1.8  1998/03/21 00:23:48  shaggy
//  All kinds of new changes...
//  and it works!!!
//
//  Revision 1.7  1998/02/11 04:33:44  shaggy
//  print scopes
//
//  Revision 1.6  1998/02/08 19:07:05  shaggy
//  about to start splitting up everything
//
//  Revision 1.5  1998/01/12 04:12:48  shaggy
//  got dependencies working when re-generating code
//
//  Revision 1.4  1998/01/11 05:06:33  shaggy
//  about to switch to standard ref/ptr semantics
//
//  Revision 1.3  1996/11/23  17:21:35  shaggy
//  last checkin before switching to virtual functions
//
//  Revision 1.2  1996/10/19  21:47:37  shaggy
//  modified to handle new tokens
//
//  Revision 1.1  1996/04/10  12:13:48  shaggy
//  Initial revision
//


#include <defines>
#include <types.h>
#include <inttypes.h>
#include <CAction>
#include <posix/stdio.h>
#include <posix/assert.h>


const CAction::Item CAction::null_item(CAction::NONE);


#if 0
void
CAction::ItemStack::push(Item_ item)
{
	Node *new_node = new Node;
	new_node->item = item;
	new_node->next = _ptr;
	_ptr = new_node;
}


bool
CAction::ItemStack::pop(Item &item)
{
	Node *old_node = _ptr;
	if(!old_node)
		return false;
	item = old_node->item;
	_ptr = old_node->next;
	delete old_node;
}
#endif


CAction::CAction(CLex &clp, vStrTab &strtab, vSymTab &globlsym)
                  : _clex(clp), _strtab(strtab), _global_scope(globlsym)
{
	_complete = true;
	_num_errors = 0;
	_num_warnings = 0; 
}


void
CAction::accept()
{
}


void
CAction::eol()
{
}


csz *
CAction::string(vStrTab::Id id) const
{
	return _strtab.string(id);
}
 
csz *
CAction::string(Token token) const
{
	switch(token) {
	case TEMP:	return "<TEMP>";
	case RESULT:	return "<RESULT>";
	case INSN:	return "<INSN>";
	case LIST:	return "<LIST>";
	case TREE:	return "<TREE>";
	case ARRAY:	return "<ARRAY>";
	case OBJ:	return "<OBJ>";
	case TYPE:	return "<TYPE>";
	case TYPE_LIST:	return "<TYPE_LIST>";
	case PARM_LIST:	return "<PARM_LIST>";
	case EXPR_LIST:	return "<EXPR_LIST>";
	case STMT_LIST:	return "<STMT_LIST>";
	case ERROR:	return "<ERROR>";
	case ERROR_LIST:return "<ERROR_LIST>";
	case STRUCT:	return "<STRUCT>";
	case CLASS:	return "<CLASS>";
	case VAR:	return "<VAR>";
	case FUNC:	return "<FUNC>";
	default:
		return _clex.token_string((CLex::Token)token);
	}                                                   
}


sz *
CAction::string(Item_ item, sz *buf, size_t buflen,
				bool show_type) const
{
	char internal_buf[4096];
	csz *s = internal_buf;
	sz *bp = buf;

	if(show_type) {
		size_t len = sprintf(buf, "%s:", string(item.token));
		if(len >= buflen)
			return 0;
		bp += len;	
		buflen -= len;
	}

	switch(item.token) {
        case NUM:
		s = _clex.token_string((CLex::Token)item.token, item.val,
				item.c.is_signed, item.c.is_long);
                break;
        case CHAR:
		s = _clex.token_string((CLex::Token)item.token, item.val);
                break;
        case LINE:
        case CSTR:
		s = _clex.token_string((CLex::Token)item.token, item.str);
                break;
        case ID:
        case STR:
        case BACKSTR:
        case ANGSTR:
		s = _clex.token_string((CLex::Token)item.token,
					_strtab.string(item.id));
		if(item.token == ID) {
			sprintf(internal_buf, "%p:%s", item.scope, s);
			s = internal_buf;
		}
                break;
        default:
                sprintf(internal_buf, "%p:0x%llx ", item.scope, item.val);       
	}

	size_t len = strlen(s);
	if(len < buflen)
		strcpy(bp, s);
	else {
		memcpy(bp, s, buflen/2);
		memcpy(bp+buflen/2, "...", 3);
		memcpy(bp+buflen/2+3, s+len-(buflen+1)/2+4, (buflen+1)/2-3);
	}
	
	return buf; 
}


void
CAction::eprintf(csz *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	veprintf(fmt, ap);
	va_end(ap);
}


void
CAction::vmprintf(csz *prefix, csz *fmt, va_list args)
{
	eprintf("%s:%d: %s", _clex.source(), _clex.line_num(), prefix);
	veprintf(fmt, args);
	switch(_clex.last_token()) {
	case NONE:
		eprintf("\n");
		break;
	case EOF_COMPLETE:
	case EOF_INCOMPLETE:
		eprintf("; at end of file\n"); 
		break;
	case INVALID:
		eprintf("; at `%s'\n", _clex.token_string());
		break;
	default:
		eprintf("; before `%s'\n", _clex.token_string());
	}
	_clex.set_token((CLex::Token)NONE);
}


void
CAction::mprintf(csz *prefix, csz *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	vmprintf(prefix, fmt, ap);
	va_end(ap);
}


void
CAction::error(csz *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	_num_errors++;
	vmprintf("", fmt, ap);
	va_end(ap);
}


void
CAction::warning(csz *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	_num_warnings++;
	vmprintf("warning: ", fmt, ap);
	va_end(ap);
}


void
CAction::abort(csz *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	vmprintf("abort: ", fmt, ap);
	va_end(ap);
	abort_exit();
}


void
CAction::internal_error(csz *file, int line, csz *module)
{
	if(module) 
		mprintf("Internal Error ", "in %s (%s:%d)",
			module, file, line);
	else
		mprintf("Internal Error ", "(%s:%d)", file, line);
	internal_error_exit();
}


void
CAction::_assertfail(csz *cond, csz *file, int line)
{
	mprintf("", "Assertion \"%s\" failed", cond);
	internal_error(file, line, 0);
}


void
CAction::mprint_action(csz *prefix, csz *name,
			Item_ p1, Item_ p2, Item_ p3)
{
	char buf1[50], buf2[50], buf3[50];
	mprintf(prefix, "%s(%s, %s, %s)", name,
		string(p1, buf1, sizeof(buf1), true),
		string(p2, buf2, sizeof(buf2), true),
		string(p3, buf3, sizeof(buf3), true));
}


void
CAction::mprint_result(csz *prefix, Item_ r)
{
	char buf[50];
	mprintf(prefix, "\tresult: %s", string(r, buf, sizeof(buf), true));
}



vSymTab*
CAction::get_symtab(Item_ id, Item_ scope)
{
	vSymTab *symtab = (scope.token == NONE) ? &_global_scope : scope.scope;
	if(id.scope)
		symtab = id.scope;
	assert(symtab);
	return symtab;
}



void
CAction::unsupported(csz *name,
			Item_ p1, Item_ p2, Item_ p3)
{
	mprint_action("unsupported: ", name, p1, p2, p3);
	internal_error(__FILE__, __LINE__);
}

void
CAction::Test(Item *r, Item_ p1, Item_ p2, Item_ p3)
{
	mprint_action("", "Test", p1, p2, p3);
	*r = p2;
}

void
CAction::Undeclared(Item *r, Item_ sym, Item_ /*scope*/, Item_)
{
	error("`%s' undeclared", _strtab.string(sym.id));
	*r = null_item;
}

void
CAction::Index(Item *r, Item_ array, Item_ index, Item_ )
{
	BinaryOp(r, OP_INDEX, array, index);
}


void
CAction::NewAggr(Item *r, Item_ type, Item_ sym, Item_ scope)
{
	NewScope(r, type, sym, scope);
}

void
CAction::NewNspace(Item *r, Item_ type, Item_ sym, Item_ scope)
{
	NewScope(r, type, sym, scope);
}


#define UNSUPPORTED(n)  void \
	CAction::n(Item *, Item_ p1, Item_ p2, Item_ p3) \
	{ unsupported(#n, p1, p2, p3); }

UNSUPPORTED(Literal)
UNSUPPORTED(UnaryOp)
UNSUPPORTED(BinaryOp)

UNSUPPORTED(BuildTypeList)
UNSUPPORTED(BuildParamList)
UNSUPPORTED(BuildExprList)
UNSUPPORTED(BuildStmtList)

UNSUPPORTED(DeclSym)
UNSUPPORTED(DeclParam)
UNSUPPORTED(DeclPtr)
UNSUPPORTED(DeclArray)
UNSUPPORTED(DeclFunc)
UNSUPPORTED(Define)
UNSUPPORTED(DefParam)
UNSUPPORTED(AggrRef)
UNSUPPORTED(EnumRef)
UNSUPPORTED(NewScope)
UNSUPPORTED(FreeScope)
UNSUPPORTED(NewEnum)
UNSUPPORTED(SetAccess)
UNSUPPORTED(Inherit)

UNSUPPORTED(FindSym)
UNSUPPORTED(GetMember)
UNSUPPORTED(GetThis)
UNSUPPORTED(GetDestructor)
UNSUPPORTED(SpecialId)

UNSUPPORTED(Call)
UNSUPPORTED(Cast)
UNSUPPORTED(Label)
UNSUPPORTED(CaseLabel)
UNSUPPORTED(DefaultLabel)
UNSUPPORTED(Switch)

UNSUPPORTED(Return)
UNSUPPORTED(Goto)
UNSUPPORTED(Jump)
UNSUPPORTED(IfElse)
UNSUPPORTED(CondExpr)

UNSUPPORTED(Patch)

