/*  $Id: vCAction.hh,v 1.7 1998/08/22 03:35:16 shaggy Exp $
 *      Default C/C++ Parsing Actions
 *      
 *      relative path: .//lib/cparse
 *  
 *      Shag/OS: an Object-Oriented MicroKernel Operating System
 *       Libraries
 *        C/C++ Parsing classes
 *  
 *  Copyright (c) 1996,1997 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the Shag/OS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare Non-profit Public License (SNPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare Non-profit Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

//  $Log: vCAction.hh,v $
//  Revision 1.7  1998/08/22 03:35:16  shaggy
//  header update
//
//  Revision 1.6  1998/08/12 15:18:36  shaggy
//  better typing; fixed some type coercions;
//  cleanup for #include's, and fixed some constructor call bugs
//
//  Revision 1.5  1998/08/03 00:12:59  shaggy
//  tons of work-- friends, class scope fixes, tons of bugfixes,
//  lots more cleanup, more DC++ work, etc
//
//  Revision 1.4  1998/08/01 14:56:47  shaggy
//  about to restructure deco
//
//  Revision 1.3  1998/07/30 16:16:31  shaggy
//  switched to new vsn of shacc and removed useless underscore prefixes
//  as well as lot of extra configuration code that is no longer needed
//
//  Revision 1.2  1998/07/30 04:40:32  shaggy
//  much more dtype and dclass work
//
//  Revision 1.1  1998/07/28 16:43:21  shaggy
//  moved lots of Deco into generic CParse lib
//  CAction split into vCAction and CAction
//  printaction -> part of CAction
//  ccint -> CRep (C Representation)
//  ccout -> COut (C output)
//


#include <CLex>
#include <vStrTab>
#include <vSymTab>




#ifdef NDEBUG
#	define assertfail(c) internal_error(__FILE__, __LINE__)
#else
#	define assertfail(c) _assertfail(#c, __FILE__, __LINE__)
#endif


class vErrorReport {
public:
	virtual void error(csz *fmt, ...) = 0;
	virtual void warning(csz *fmt, ...) = 0;
	virtual void abort(csz *fmt, ...) = 0;
	virtual void internal_error(csz *file, int line,
				csz *module = 0) = 0;
};


#define CACTION0(fn)	virtual void fn(Item *r, Item_, Item_, Item_) = 0;
#define CACTION1(fn,p1)	virtual void fn(Item *r, Item_ p1, Item_, Item_) = 0;
#define CACTION2(fn,p1,p2)	\
	virtual void fn(Item *r, Item_ p1, Item_ p2, Item_) = 0;
#define CACTION3(fn,p1,p2,p3)	\
	virtual void fn(Item *r, Item_ p1, Item_ p2, Item_ p3) = 0;



class vCAction : public vErrorReport {
protected:
	bool _complete;
	int _num_errors;
	int _num_warnings;
public:
	enum Token {
		NONE = 0,
		EOF_COMPLETE,
		EOF_INCOMPLETE,
		INCOMPLETE, INVALID,

#include "CParse.tab.h"

		TEMP,		// used to designate temporaries
		LITERAL,	// a converted literal (constant) value
		RESULT,		// result of specified code
		INSN,		// pointer to another instruction
		LIST,		// linked list
		TREE,		// binary (or more) tree
		ARRAY,		// linear array
		OBJ,		// generic object
		TYPE,		// generic type
		TYPELIST,	// declaration list
		PARM,		// parameter definition
		PARMLIST,	// parameter list
		EXPR,		// expression
		EXPRLIST,	// expression list
		STMT,		// statement
		STMTLIST,	// statement list
		ERR,		// error information
		ERRLIST,	// list of error information
		DECL,		// a declaration
		DECLLIST,	// a list of declarations
		STRUCT,		// structure information
		CLASS,		// class information
		VAR,		// variable
		FUNC,		// function
		CTORINIT,	// a constructor initializer

		NUM_TOKENS
	};

	struct Item {
		Token token;
		ushort line;	// more than 65535 lines in a single C++
				// source file seems really absurd...
				// (I'm not denying that it happens,
				// but such files *really* need to be
				// broken up!)
		union {
			vSymTab *scope;
			struct {
				bool is_signed;
				char is_long;
			} c;	// for constants/literals only
			void *cptr;	// for constants/literals only
		};
		union {
			uint64 val;		// LITERAL (is_signed=0)
			int64 sval;		// LITERAL (is_signed=1)
			Token tval;		//
			uint64 Num;		// NUM
			bool Bool;		// BOOL
			uint Char;		// CHAR
			vStrTab::Id id;		// STR, ID, BACKSTR, ANGSTR
			double Float;		// FLOAT
			class TokenList *TokenList; // TOKENLIST
			struct {
				int32 top;
				int32 bottom;
			} Action;		// ACTION
		};

		Item() {}
		Item(Token t) { token = t; val = 0; }
		Item(CLex::Token t) { token = (Token)t; val = 0; }
		Item(Token t, uint64 v) { token = t; val = v; }
		Item(CLex::Token t, uint64 v) { token = (Token)t; val = v; }
		// Item(Token t, double f) { token = t; fval = f; }
		// Item(CLex::Token t, double f) { token = (Token)t; fval = f; }
		// Item(Token t, csz *s) { token = t; str = s; }
		// Item(CLex::Token t, csz *s) { token = (Token)t; str = s; }
		// Item(Token t, void *p) { token = t; ptr = p; }
		// Item(CLex::Token t, void *p) { token = (Token)t; ptr = p; }
		// Item(csz *s) { token = CSTR; str = s; }
		Item(int n) { token = NUM; val = n; }
		Item& operator=(Token t) { token = t; val = 0; return *this; }
		Item& operator=(CLex::Token t)
			{ token = (Token)t; val = 0; return *this; }
	};
	typedef const Item &Item_;

	static const Item none;

	
	bool complete() const { return _complete; }
	void complete(bool tf) { _complete = tf; }

	int errors() const { return _num_errors; }
	int warnings() const { return _num_warnings; }

	virtual void accept() = 0;
	virtual void get_input(bool split_token) = 0;
	virtual vStrTab::Id strid(Token, csz *str) = 0;

	virtual void eol() = 0;
	virtual csz* cstr(Token token) const = 0;
	virtual csz* cstr(vStrTab::Id id) const = 0;
	virtual sz * cstr(Item_ item, sz *buf, size_t buflen,
				bool show_type = false) const = 0;

	virtual void eprintf(csz *fmt, ...) = 0;
	virtual void veprintf(csz *fmt, va_list args) = 0;
	virtual void mprintf(csz *prefix, csz *fmt, ...) = 0;
	virtual void vmprintf(csz *prefix,
				csz *fmt, va_list args) = 0;
	virtual void _assertfail(csz *cond, csz *filename,
				int line) = 0;

	virtual void mprint_action(csz *prefix, csz *name,
			Item_ p1, Item_ p2, Item_ p3) = 0;
	virtual void mprint_result(csz *prefix, Item_ r) = 0;
	virtual void unsupported(csz *name, Item_ p1, Item_ p2, Item_ p3) = 0;

	virtual vSymTab* get_symtab(Item_ id, Item_ scope) = 0;
	
	typedef void action(Item *r, Item_ p1, Item_ p2, Item_ p3);
	typedef action vCAction::* Action;

	CACTION3(Test, p1, p2, p3);		/* r = p1 */
	// used for testing parts of the parser during development
	
	CACTION2(Undeclared, sym, scope);	/* r = p1 */
	//	sym: NewId
	//	scope: AnyScope
	// report an error about an undeclared symbol

	CACTION1(Literal, item);		/* r = literal(item) */
	//	item: RawLit|Lit|DeclId|Expr
	//	r: Lit
	// create a descriptor for a literal
	// should return NONE if no valid literal can be generated

	CACTION2(Primary, item, scope);
	// use item as a primary expression, providing any necessary
	// extra functionality required (such as `this->') to access
	// it in the current scope

	CACTION2(UnaryOp, op, arg); 		/* r = op(arg) */
	//	op: OP_INC|OP_DEC|OP_INCP|OP_DECP|OP_SIZEOF
	//		|OP_MUL|OP_AND|OP_ADD|OP_SUB|OP_LNOT|OP_NOT
	//	arg: Lit|Id|Expr
	//	r: Lit|Expr
	// perform a unary operation
	// should handle NONE without generating additional errors
	
	CACTION3(BinaryOp, op, arg1, arg2);	/* r = op(arg1, arg2) */
	//	op: OP_INDEX|OP_DOT|OP_PTRTO|OP_MDOT|OP_MPTRTO
	//		|OP_MUL|OP_DIV|OP_MOD|OP_ADD|OP_SUB
	//		|OP_LSHIFT|OP_RSHIFT
	//		|OP_LSS|OP_GTR|OP_LEQ|OP_GEQ|OP_EQU|OP_NEQ
	//		|OP_AND|OP_XOR|OP_OR
	//		|OP_LAND|OP_LOR
	//		|OP_COMMA
	//		|OP_ASN|OP_ASNMUL|OP_ASNDIV|OP_ASNMOD
	//		|OP_ASNADD|OP_ASNSUB
	//		|OP_ASNRSHIFT|OP_ASNLSHIFT
	//		|OP_ASNAND|OP_ASNXOR|OP_ASNOR
	//	arg1: Lit|Id|Expr
	//	arg2: Lit|Id|Expr
	// 	r: Lit|Expr
	// perform a binary operation
	// should handle NONE without generating additional errors

	CACTION2(Index, array, index);		/* r = array[index] */
	// 	array: Lit|Id|Expr
	// 	index: Lit|Id|Expr
	// note that if index is a pointer, r = index[array] instead,
	// for backwards compatibility with C
	// should handle NONE without generating additional errors

	CACTION2(DeclParam, type, sym);
	//	type:
	//	sym:
	//	r:

	CACTION3(DeclSym, type, sym, scope);
	//	type: TypeList | EnumType | NONE
	//	sym: Id
	//	scope: AnyScope
	//	r: DeclId
	// declare a symbol in the given scope to be of type `type'

	CACTION2(DeclPtr, ptrtype, type);
	//	ptrtype: OP_PTR | OP_REF 
	//	type: Type
	//	r: Type
	// return a type that is a pointer or reference to type `type'

	CACTION3(DeclArray, type, size, lower);
	//	type: Type
	//	size: Lit|Id|Expr|NONE
	//	lower: Lit|Id|Expr|NONE
	//	r: Type
	// return a type that is an array from [lower..lower+size-1] of `type'
	// a lower bound of NONE implies 0, while a size of NONE
	// implies an array with an unknown/variable size

	CACTION3(DeclFunc, type, plist, cv);
	//	type: Type
	//	plist: ParamList
	//	cv: TypeList
	//	r: Type
	// return a function type for the specified param list and return value

	CACTION3(SetAttr, decl, attr, val);
	//	decl: Id
	//	attr: KW_INLINE
	//	val: Expr
	// sets an attribute on a type, variable, fn, etc...

	CACTION2(Define, var, val);
	//	var: Id
	//	val: Lit|Id|Expr
	//	r: Id
	// define the default value for a variable, or the
	// code body for a function
	// should handle NONE without generating additional errors

	CACTION2(DefParam, par, val);
	//	par: 
	//	val:
	//	r:
	
	CACTION3(AggrRef, type, sym, scope);
	//	type:
	//	sym:
	//	scope:
	//	r:

	CACTION3(AggrDecl, type, sym, scope);
		
	CACTION3(NewAggr, type, sym, scope);
	//	type: KW_STRUCT | KW_CLASS | KW_UNION | KW_DCLASS | KW_DTYPE
	//	sym: Id | NONE
	//	scope: AnyScope (enclosing scope)
	//	r: AggrId
	// note: if undefined, defaults to calling NewScope,
	// and NewScope is expected to be able to handle NewAggr as well

	CACTION1(EndAggr, aggr);
	//	aggr: 
	//	r: AggrId

	CACTION3(NewNspace, type, sym, scope);
	//	type: KW_NAMESPACE
	//	sym: Id | NONE
	//	scope: AnyScope (enclosing scope)
	//	r: NspaceId
	// note: if undefined, defaults to calling NewScope
	// and NewScope is expected to be able to handle NewNspace as well

	CACTION3(NewLocal, type, sym, scope)
	//	type: NONE
	//	sym: NONE
	//	scope: AnyScope (enclosing scope)
	//	r: Local
	// note: if undefined, defaults to calling NewScope,
	// and NewScope is expected to handle NewLocal as well

	CACTION1(FuncScope, sym)
	//	sym: VAL_ID
	//	r: Scope
 
	CACTION3(NewScope, type, sym, scope);
	//	type: KW_STRUCT | KW_CLASS | KW_UNION | KW_NAMESPACE | NONE
	//	sym: Id | NONE
	//	scope: AnyScope (enclosing scope)
	//	r: Scope
	// create a new scope, and enter a handle to it in
	// the enclosing scope, with the specified name, or anonymous

	CACTION1(FreeScope, scope);
	//	scope: Scope
	//	r: Scope (parent of freed scope)
	// free the specified scope to reclaim its memory
	// all symbols created within it can be freed as well

	CACTION3(EnumRef, type, sym, scope);
	//	type:
	//	sym:
	//	scope:
	//	r:
	
	CACTION3(NewEnum, type, sym, scope);
	//	type: KW_ENUM
	//	sym: Id
	//	scope: Scope
	//	r: EnumType

	CACTION1(EnumNext, prev);
	//	prev: Lit | NONE
	//	r: Lit

	CACTION2(SetAccess, scope, acc);
	//	scope:
	//	acc:
	//	r:

	CACTION3(Inherit, scope, itype, base);
	//	scope:
	//	itype:
	//	base:
	//	r:

	CACTION2(BuildTypeList, typelist, item);
	//	typelist: TypeList|NONE
	//	item: RawTypeKw|TypeId|TypeList
	//	r: TypeList
	// builds a compound type list from simple type keywords
	// and predefined types
	// note that joining two typelists must be supported

	CACTION2(BuildParamList, plist, item);
	//	plist: ParamList|NONE
	//	item: Id|ELIPSIS|NONE
	//	r: ParamList
	// builds a list of parameters used in formal function definitions
	// if item==ELIPSIS the paramlist should be marked as having
	// a variable number of parameters

	CACTION2(BuildArray, exprlist, item);
	//	exprlist: ExprList | NONE
	//	item: Lit|Id|Expr|NONE
	//	r: ExprList
	// builds a list of expressions to be used for a function call
	// should handle NONE without generating additional errors

	CACTION2(BuildExprList, exprlist, item);
	//	exprlist: ExprList | NONE
	//	item: Lit|Id|Expr|NONE
	//	r: ExprList
	// builds a list of expressions to be used for a function call
	// should handle NONE without generating additional errors

	CACTION2(BuildStmtList, stmtlist, item);
	//	stmtlist: StmtList | NONE
	//	item: Stmt
	//	r: StmtList
	// builds a list of statements

	CACTION3(FindSym, type, sym, scope);
	//	type: NONE|SCOPE|VAL_ID|AGGR_ID
	//		|KW_CLASS|KW_STRUCT|KW_UNION|KW_ENUM
	//	sym: Id|NONE
	//	scope: ANyScope|NONE
	//	r: Id
	// find a symbol, starting in the specified scope
	// if sym.scope is set, only that scope will be used
	// if sym==NONE && type==SCOPE, the global scope should be returned

	CACTION2(Cast, type, val);
	//	type:
	//	val:
	//	r:

	CACTION3(Call, fn, param, scope);
	//	fn:
	//	param:
	//	r:

	CACTION3(SpecialId, type, val, scope);
	//	type:
	//	val:
	//	scope:
	//	r:

	CACTION2(SpecialStr, type, scope);
	
	CACTION2(CtorInit, id, params);
	//	id:	AggrId
	//	params:	TokenList
	//	r:	CtorInit

	CACTION2(CtorCall, fn, param);

	CACTION3(OpNew, type, init, param);

	// XXX - need to have parameters properly defined:
	CACTION3(Label, a, b, c);
	CACTION3(Switch, a, b, c);
	CACTION3(Return, a, b, c);
	CACTION3(Goto, a, b, c);
	CACTION3(Break, a, b, c);
	CACTION3(Continue, a, b, c);
	CACTION3(IfElse, a, b, c);
	CACTION3(CondExpr, a, b, c);
	CACTION3(While, a, b, c);
	CACTION3(DoWhile, a, b, c);
	CACTION3(For, a, b, c);
	CACTION3(ForExpr, a, b, c);
};

// General guidelines on all actions:
//	- since the arguments passed to the action are passed as `const Item&',
//	  they might sometimes refer to the same object as `Item *r'.  
//	  Therefore, no portions of any arguments can be used after
//	  *r is changed in any way.  This means updates to *r should
//	  probably be done last, after all information is gathered.


//    implementation-defined types:
//
//	(Tokens listed in square brackets can be optionally
//	replaced with any token-- the one listed is just
//	a standard preference, but it doesn't affect the parser---
//	if other tokens are listed as well, then those ones have special
//	meanings, and should not be reused)
//
//	AnyId:		any identifier (might not be declared)
//		= DeclId | NewId
//	AnyDeclId:	any declared indentifier
//		= AnyTypeId | AnyValId | NspaceId | TplateId
//	AnyTypeId:	identifier for any type descriptor
//		= AggrId | EnumId | TypeId
//	AnyScope:	identifier for any scope
//		= AggrId | NspaceId | CodeScope
//	AnyValId:	identifier for any symbol with a value
//		token= ID | DTOR_ID | CONV_ID | OPER_ID
//	AggrId:		identifier for an aggregate (class/struct/union)
//		token= AGGR_ID
//	EnumId:		identifier for an enumeration
//		token= ENUM_ID
//	TypeId:		identifier for a type (not enum/aggr)
//		token= TYPE_ID
//	ValId:		identifier for a symbol with a value
//		token= VAL_ID
//	NewId:		undefined identifier
//		token= ID
//	NspaceId:
//		token= NSPACE_ID | ASPACE_ID
//	CodeScope:
//		token= SCOPE
//	TypeList:	compound list of identifiers to describe a type
//		token= [TYPELIST]
//	Lit:		literal (and synthesized literals, which
//				are essentially constants)
//		token= [LITERAL] | NONE
//	Expr:		compound expression
//
//	
//   CParse-defined types:
//
//	RawId:		
//		token= ID
//		id= string id
//
//	RawLit:
//	    integer number:
//		token= NUM 
//		val= 64-bit number
//		c.is_signed= literal has U suffix
//		c.is_long= # of L suffixes
//
//	    character:
//		token= CHAR
//		val= character number
//
//	    string:
//		token= STR
//		id= string id
//
//	    floating-point number:
//		token= FLOAT
//		fval= number (as a double)
//
//	    boolean:
//		token= BOOL
//		val= 0:false, 1:true
//
//	RawTypeKw:
//	    storage class:
//		token= KW_AUTO | KW_REGISTER | KW_STATIC | KW_EXTERN
//			| KW_MUTABLE
//
//	    simple type:
//		token= KW_CHAR | KW_WCHAR | KW_BOOL | KW_INT
//			| KW_FLOAT | KW_DOUBLE | KW_VOID
//
//	    type modifier:
//		token= KW_SHORT | KW_LONG | KW_SIGNED | KW_UNSIGNED
//
//	    function specifier:
//		token= KW_INLINE | KW_VIRTUAL | KW_EXPLICIT
//
//	    special declarations:
//		token= KW_FRIEND | KW_TYPEDEF
//
//	    cv qualifier:
//		token= KW_CONST | KW_VOLATILE	
//
