/*  Lexer.cc                                                    (CRC:41cafee4)
 *      relative path: ./lib/parse
 *      creation date: Mon Mar  4 00:48:00 1996
 *      last modified: Fri Aug 21 22:43:47 1998
 *      header update: Fri Aug 21 22:54:51 1998
 *  
 *      Shag/OS: an Object-Oriented MicroKernel Operating System
 *       Libraries
 *        Generic parsing functions
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the Shag/OS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

#define IMPLEMENTATION_Lexer
#include <Lexer>
#include <new>


// Currently, this initializes a ctype table for 
// standard 7-bit ASCII, with all values above 126 marked 
// as unknown.  In the future, Unicode UTF8 support ought to be
// added, and this will have to change.

Lexer::Lexer()
{
	if(init_ctypes) {
		int i;
		for(i=0; i<256; i++)
			ctype[i] = 0;
		ctype[' '] |= WHITESPACE;
		ctype['\t'] |= WHITESPACE;
		ctype['\r'] |= WHITESPACE;
		ctype['\n'] |= WHITESPACE;
		ctype['\f'] |= WHITESPACE;
		ctype['\v'] |= WHITESPACE;
		ctype['$'] |= ID | ID_START;
		ctype['_'] |= ID | ID_START;
		for(i='A'; i<='Z'; i++) {
			ctype[i] |= UPPER | ID | ID_START;
			ctype[i-'A'+'a'] |= LOWER | ID | ID_START;
		}
		for(i='A'; i<='F'; i++) {
			ctype[i] |= HEXNUM;
			ctype[i-'A'+'a'] |= HEXNUM;
		}
		for(i='0'; i<='9'; i++) 
			ctype[i] |= NUM | HEXNUM | ID;
		for(i='0'; i<='7'; i++)
			ctype[i] |= OCTNUM;
		for(i='0'; i<='1'; i++)
			ctype[i] |= BINNUM;
		for(i=32; i<=126; i++)
			if(!ctype[i])
				ctype[i] |= PUNC;
	}
	set_empty();
	_sol = true;
	_line_num = 1;
}

Lexer::Lexer(const char *str, size_t len)
{
	new(this)Lexer();
	set_src(str, len);
}

