/*  Lexer.hh                                                    (CRC:2cc614c0)
 *      relative path: ./lib/parse
 *      creation date: Mon Mar  4 00:21:39 1996
 *      last modified: Fri Aug 21 22:43:47 1998
 *      header update: Fri Aug 21 22:54:52 1998
 *  
 *      Shag/OS: an Object-Oriented MicroKernel Operating System
 *       Libraries
 *        Generic parsing functions
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the Shag/OS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

#include <types.h>

class Lexer {
	const char *_str;	// source buffer
	size_t _len;		// remaining size of source buffer
	bool _sol;		// start of line
	int _line_num;		// current source line number
	enum ctypes {
		UPPER =		0x0001,
		LOWER =		0x0002,
		ALPHA =		0x0003, // for scanning only
		NUM =		0x0004,
		ALPHANUM = 	0x0007, // for scanning only
		PUNC = 		0x0008,
		WHITESPACE = 	0x0020,
		HEXNUM = 	0x0040,
		OCTNUM = 	0x0080,
		BINNUM = 	0x0100,
		ID_START = 	0x0400,
		ID = 		0x0800,
	}; 
	static short ctype[256];
	static bool init_ctypes;
public:
	void set_src(const char *str, size_t len) 
		{ _str = str; _len = len; }
	Lexer();
	Lexer(const char *str, size_t len);
	bool sol() const { return _sol; }
	bool eol() const { return *_str == '\n'; }
	bool eof() const { return !_str; }
	void set_eof() { set_src(0, 0); }
	void set_empty() { set_src("", 0); }
	char get_char() const { return *_str; }
	int get_digit() const { return *_str-'0'; }
	int get_hex() const { if(is_num()) return get_digit();
			else return (*_str-'A'+10)&0xf; }
	bool match(char ch) const { return *_str == ch; }
	bool next_char() {
		if((_sol = (*_str == '\n'))) _line_num++;
		if(!--_len) return false;
		else { _str++; return true; }
	}
	int line_num() const { return _line_num; }
	void set_line_num(int num) { _line_num = num; }
	static bool is_upper(char ch) { return ctype[ch] & UPPER; }
	static bool is_lower(char ch) { return ctype[ch] & LOWER; }
	static bool is_alpha(char ch) { return ctype[ch] & ALPHA; }
	static bool is_num(char ch) { return ctype[ch] & NUM; }
	static bool is_bin(char ch) { return ctype[ch] & BINNUM; }
	static bool is_oct(char ch) { return ctype[ch] & OCTNUM; }
	static bool is_hex(char ch) { return ctype[ch] & HEXNUM; }
	static bool is_alphanum(char ch) 
			{ return ctype[ch] & ALPHANUM; }
	static bool is_punc(char ch) 
			{ return ctype[ch] & PUNC; }
	static bool is_whitespace(char ch) 
			{ return ctype[ch] & WHITESPACE; }
	static bool is_id_start(char ch) 
			{ return ctype[ch] & ID_START; }
	static bool is_id(char ch) 
			{ return ctype[ch] & ID; }
	bool is_empty() const { return !_len; }
	bool is_upper() const { return is_upper(*_str); }
	bool is_lower() const { return is_lower(*_str); }
	bool is_alpha() const { return is_alpha(*_str); }
	bool is_num() const { return is_num(*_str); }
	bool is_bin() const { return is_bin(*_str); }
	bool is_oct() const { return is_oct(*_str); }
	bool is_hex() const { return is_hex(*_str); }
	bool is_alphanum() const { return is_alphanum(*_str); }
	bool is_punc() const { return is_punc(*_str); }
	bool is_whitespace() const { return is_whitespace(*_str); }
	bool is_id_start() const { return is_id_start(*_str); }
	bool is_id() const { return is_id(*_str); }
};


#ifdef IMPLEMENTATION_Lexer
bool Lexer::init_ctypes = true;
short Lexer::ctype[256];
#endif


