/*  ParsePath.hh                                                (CRC:d4d4ed5b)
 *      relative path: ./lib/parse
 *      creation date: Oct 1995
 *      last modified: Sat Jun 21 08:24:41 1997
 *      header update: Fri Aug 21 22:54:52 1998
 *  
 *      Shag/OS: an Object-Oriented MicroKernel Operating System
 *       Libraries
 *        Generic parsing functions
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the Shag/OS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

#ifndef NAME_MAX
#define NAME_MAX	255	// max chars in a filename
#endif
#ifndef PATH_MAX
#define PATH_MAX	1024	// max chars in a pathname
#endif



// OPT: could compute each part when it is first accessed only
//	or could just make into separate classes, and inherit only
//	the parts needed


class ParsePath {
	bool absolute;
	char *end;
	char *next_dir;
	char base[NAME_MAX];
	char dir[PATH_MAX];
public:
	void parse(const char *path);
	ParsePath();
	ParsePath(const char *path)	{ parse(path); }
	const bool is_absolute() const	{ return absolute; }
	const bool has_dir() const	{ return dir[0] != 0; }
	const bool has_base() const	{ return base[0] != 0; }
	const bool is_root() const	{ return absolute && !has_dir() && !has_base(); }
	const char* get_dir() const	{ return dir; }
	const char* get_base() const	{ return base; }
	const char* get_ext(const char *div) const;
	const char* get_suffix(const char *div) const;
	bool has_suffix(const char *suffix) const;
	bool change_suffix(char *nb, const char *olds, const char *news) const;
	bool get_next_dir(char *nd = 0);
	void first_dir()
		{ next_dir = dir; }
	bool first_dir(char *nd)
		{ first_dir(); return get_next_dir(nd); }
	void get_dir_path(char *np) const;
};

