/*  $Id: string.cc,v 1.5 1998/08/22 03:35:37 shaggy Exp $
 *      string operations optimized for i386
 *      
 *      relative path: ./lib/posix/m-i386
 *  
 *      Shag/OS: an Object-Oriented MicroKernel Operating System
 *       Libraries
 *        Posix-compliant code
 *         (m-i386)
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the Shag/OS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

//  $Log: string.cc,v $
//  Revision 1.5  1998/08/22 03:35:37  shaggy
//  header update
//
//  Revision 1.4  1998/02/11 04:53:54  shaggy
//  proper linkage directives
//
//  Revision 1.3  1996/11/08  22:24:51  shaggy
//  header update
//
//  Revision 1.2  1996/10/26  19:09:32  shaggy
//  header update
//
//  Revision 1.1  1996/10/06  13:45:29  shaggy
//  Initial revision
//
//      last modified: Wed May 22 18:27:33 1996
//      creation date: Dec 1994


#pragma Make Flags_Banned += -pedantic

#define Q(n)	#n
#ifdef __ELF__
#	define C(n)	Q(n)
#else
#	define C(n)	Q(_##n)
#endif


#ifndef FUNC_memcpy
#define FUNC_memcpy
#ifndef ARCH_i386sx	/* attempt to move longwords */
/* void memcpy(char *dest, char *src, int size) */
asm("
.globl "C(memcpy)"	
"C(memcpy)":
	pushl	%ebp
	movl	%esp,%ebp
	pushl	%esi	
	pushl	%edi
	movl	8(%ebp),%edi
	movl	12(%ebp),%esi
	movl	16(%ebp),%ecx
	cmpl	$0,%ecx		/* leave if done */
	je	3f
1:	test	$3,%edi		/* align dest to longword boundary */
	jz	2f						
	movsb	(%esi),(%edi)	/* copy a byte */	
	loop	1b				
	jmp	3f			
2:	cmpl	$0,%ecx		/* leave if done */
	je	3f		 			
	movl	%ecx,%edx			
	shrl	$2,%ecx		/* number of longwords */
	andl	$3,%edx		/* number of bytes remainder */
	rep	
	movsl	(%esi),(%edi)	/* copy longwords */	
	cmpl	$0,%edx		/* leave if done */
	je	3f		
	movl	%edx,%ecx	/* get remainder */	
	rep	
	movsb	(%esi),(%edi)	/* copy bytes */
3:	popl	%edi	
	popl	%esi	
	leave		
	ret	
");
#else		/* 16-bit bus -- move shorts only */
/* void memcpy(char *dest, char *src, int size) */
asm("
.globl "C(memcpy)"
"C(memcpy)":
	pushl	%ebp				
	movl	%esp,%ebp			
	pushl	%esi			
	pushl	%edi		
	movl	8(%ebp),%edi
	movl	12(%ebp),%esi
	movl	16(%ebp),%ecx		
	cmpl	$0,%ecx		/* leave if done */
	je	3f
1:	test	$1,%edi		/* align dest to word boundary */
	jz	2f
	movsb	(%esi),(%edi)	/* copy a byte */
	dec	%ecx	
2:	cmpl	$0,%ecx		/* leave if done */
	je	3f		
	movl	%ecx,%edx	
	shrl	$1,%ecx		/* number of longwords */
	rep		
	movsw	(%esi),(%edi)	/* copy words */
	test	$1,%edx		/* check remainder */	
	jz	3f	
	movsb	(%esi),(%edi)	/* copy leftover byte */
3:	popl	%edi	
	popl	%esi
	leave	
	ret
");
#endif
#endif



#ifndef FUNC_memset
#define FUNC_memset
/* void memset(char *dest, uchar data, int nbytes) */
asm("
.globl "C(memset)"
"C(memset)":
	pushl	%ebp
	movl	%esp,%ebp
	pushl	%edi		
	movl	8(%ebp),%edi	
	movl	16(%ebp),%ecx
	movl	12(%ebp),%eax
	cmpl	$0,%ecx	
	je	1f
	rep	
	stosb	
1:	popl	%edi	
	leave
	ret
");
#endif

#ifndef FUNC_memset_w
#define FUNC_memset_w
/* void memset_w(char *dest, ushort data, int nwords) */
asm("
.globl "C(memset_w)"
"C(memset_w)":
	pushl	%ebp	
	movl	%esp,%ebp	
	pushl	%edi	
	movl	8(%ebp),%edi
	movl	16(%ebp),%ecx	
	movl	12(%ebp),%eax
	cmpl	$0,%ecx	
	je	1f
	rep	
	stosw
1:	popl	%edi
	leave
	ret
");
#endif

#ifndef FUNC_memset_l
#define FUNC_memset_l
/* void memset_l(char *dest, int nwords, ulong data) */
asm("
.globl "C(memset_l)"
"C(memset_l)":
	pushl	%ebp
	movl	%esp,%ebp
	pushl	%edi	
	movl	8(%ebp),%edi
	movl	16(%ebp),%ecx	
	movl	12(%ebp),%eax
	cmpl	$0,%ecx	
	je	1f
	rep	
	stosl
1:	popl	%edi
	leave
	ret
");
#endif


#ifndef FUNC_strcpy
#define FUNC_strcpy
asm("
.globl "C(strcpy)"
"C(strcpy)":
	pushl	%ebp
	movl	%esp,%ebp
	pushl	%edi
	pushl	%esi	
	movl	8(%ebp),%edi
	movl	12(%ebp),%esi
1:	lodsb
	stosb	
	cmpb	$0,%al
	jnz	1b
	mov	12(%ebp),%eax
	popl	%esi
	popl	%edi
	leave
	ret
");
#endif

#ifndef FUNC_strcat
#define FUNC_strcat
asm("
.globl "C(strcat)"
"C(strcat)":
	pushl	%ebp	
	movl	%esp,%ebp
	pushl	%edi	
	pushl	%esi	
	movl	8(%ebp),%esi
1:	lodsb	
	cmpb	$0,%al	
	jnz	1b
	leal	-1(%esi),%edi
	movl	12(%ebp),%esi
1:	lodsb	
	stosb
	cmpb	$0,%al	
	jnz	1b
	mov	8(%ebp),%eax
	popl	%esi
	popl	%edi
	leave	
	ret
");
#endif

#ifndef FUNC_strlen
#define FUNC_strlen
asm("
.globl "C(strlen)"
"C(strlen)":	
	pushl	%ebp	
	movl	%esp,%ebp
	pushl	%esi		
	movl	8(%ebp),%esi
1:	lodsb		
	cmpb	$0,%al
	jnz	1b
	movl	%esi,%eax	
	subl	8(%ebp),%eax
	decl	%eax	
	popl	%esi
	leave	
	ret	
");
#endif

