/*  $Id: stdio.h,v 1.3 1998/08/22 03:35:35 shaggy Exp $
 *      relative path: ./lib/posix
 *  
 *      Shag/OS: an Object-Oriented MicroKernel Operating System
 *       Libraries
 *        Posix-compliant code
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the Shag/OS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

/*
 *  $Log: stdio.h,v $
 *  Revision 1.3  1998/08/22 03:35:35  shaggy
 *  header update
 *
 *  Revision 1.2  1997/06/21 12:24:02  shaggy
 *  header update
 *
 *  Revision 1.1  1997/01/21 05:04:01  shaggy
 *  Initial revision
 *
 *      creation date: 1992
 */


#include <posix/posix.h>
#include <posix/sys/types.h>
#include <posix/stddef.h>
#include <posix/fcntl.h>
#include <posix/stdarg.h>

typedef ulong fpos_t;

typedef struct {
	/* obj_t	fd; */
	fpos_t	pos;
} FILE;

extern_C FILE _stdfile[];

#define EOF	ERR_EOF
#define stdin	&_stdfile[0]
#define stdout	&_stdfile[1]
#define stderr	&_stdfile[2]

#define BUFSIZ		XXX
#define FILENAME_MAX	MAXLABEL
#define FOPEN_MAX	MAXUSHORT
#define L_tmpnam	XXX
#define TMP_MAX		XXX
#define _IOFBF		XXX
#define _IOLBF		XXX
#define _IONBF		XXX

/* extern_C  int voprintf(obj_t oid, const char *fmt, va_list args); */
/* extern_C  int oprintf(obj_t oid, const char *fmt, ...); */
/* extern_C  int cprintf(const char *fmt, ...); */
extern_C  int printf(const char *fmt, ...);
/* extern_C  int kprintf(const char *fmt, ...); */
/* extern_C  int vkprintf(const char *fmt, va_list args); */
extern_C  int eprintf(const char *fmt, ...);
extern_C  int vprintf(const char *fmt, va_list args);
extern_C  int vsprintf(char *s, const char *fmt, va_list args);
extern_C  int sprintf(char *s, const char *fmt, ...);
extern_C  int puts(const char *s);
extern_C  int putchar(char ch);
extern_C  int getchar(void);
/* extern_C  char *ogets(char *s, int len, obj_t oid); */
/* char *gets(char *str);  
 * XXX - gets() is a major security hole, because it has no
 * way to specify the buffer size , therefore it has been
 * removed.  Use getstrn(buf,len) instead  */
extern_C char *getstrn(char *s, int len);

extern_C int fclose(FILE *stream);
extern_C FILE *fdopen(int fildes, const char *type);
extern_C int feof(FILE *stream);
extern_C int ferror(FILE *stream);
extern_C int fflush(FILE *stream);
extern_C int fgetc(FILE *stream);
extern_C int fgetpos(FILE *stream, fpos_t *pos);
extern_C char *fgets(char *s, int n, FILE *stream);
extern_C int fileno(FILE *stream);
extern_C FILE *fopen(const char *filename, const char *mode);
extern_C int fprintf(FILE *stream, const char *fmt, ...);
extern_C int fputc(int c, FILE *stream);
extern_C int putc(int c, FILE *stream);
extern_C int fputs(const char *s, FILE *stream);
extern_C size_t fread(void *ptr, size_t size, size_t nelem, FILE *stream);
extern_C FILE *freopen(const char *filename, const char *mode, FILE *stream);
extern_C int fscanf(FILE *stream, const char *format, ...);
extern_C int fseek(FILE *stream, long offset, int whence);
extern_C int fsetpos(FILE *stream, const fpos_t *pos);
extern_C long ftell(FILE *stream);
extern_C size_t fwrite(const void *ptr, size_t size, size_t nelem, FILE *stream);

extern void perror(const char *msg);
extern void print_error(const char *msg, ulong errcode);

