/*  $Id: StrTab.hh,v 1.5 1998/08/22 03:35:40 shaggy Exp $
 *      Symbol Table support
 *      
 *      relative path: ./lib/symtab
 *  
 *      Shag/OS: an Object-Oriented MicroKernel Operating System
 *       Libraries
 *        (symtab)
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the Shag/OS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

//  $Log: StrTab.hh,v $
//  Revision 1.5  1998/08/22 03:35:40  shaggy
//  header update
//
//  Revision 1.4  1998/05/24 01:41:41  shaggy
//  internal support for most statements
//
//  Revision 1.3  1998/04/22 12:02:04  shaggy
//  Changed ID to Id to be consistent with naming conventions
//
//  Revision 1.2  1998/01/12 04:12:48  shaggy
//  got dependencies working when re-generating code
//
//  Revision 1.1  1998/01/11 05:11:03  shaggy
//  Initial revision
//
//  Revision 1.1  1996/04/08  00:07:21  shaggy
//  Initial revision
//


#include <defines>
#include <types.h>
#include <vStrTab>

// XXX - this string table can't know where string pointers are coming
//	from, so it always allocates space for new strings and copies them.
//	This could be inefficient for strings that already have
//	been allocated, or that are static.

// for now, this is simply a linked list (very slow)
// for simplicity.  Later this should be some sort of fast
// hash table or something.

class StrTab : public vStrTab {
private:
	struct _StrEntry {
		csz 	*string;
		uint	refcount;
		_StrEntry *next;		
	};
	_StrEntry *_list;
public:
	StrTab();
	~StrTab();
	Id id(csz *string);
	csz* string(Id id) const { return id ? ((_StrEntry*)id)->string : 0; }
	uint ref(Id id) { return id ? ++((_StrEntry*)id)->refcount : 0; }
	uint unref(Id id);
};

