/*  $Id: SymTab.cc,v 1.12 1998/08/22 03:35:40 shaggy Exp $
 *      Symbol Table support
 *      
 *      relative path: ./lib/symtab
 *  
 *      Shag/OS: an Object-Oriented MicroKernel Operating System
 *       Libraries
 *        (symtab)
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the Shag/OS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

//  $Log: SymTab.cc,v $
//  Revision 1.12  1998/08/22 03:35:40  shaggy
//  header update
//
//  Revision 1.11  1998/07/30 04:40:32  shaggy
//  much more dtype and dclass work
//
//  Revision 1.10  1998/07/06 22:49:41  shaggy
//  reworked symtab, output ordering, etc
//
//  Revision 1.9  1998/07/05 02:47:37  shaggy
//  tons of bug fixes
//
//  Revision 1.8  1998/07/03 19:06:24  shaggy
//  changed scope semantics; fn name is available in scope now
//
//  Revision 1.7  1998/06/30 18:42:00  shaggy
//  reversed ordering in symbol table to be consistent with input
//
//  Revision 1.6  1998/05/04 03:09:19  shaggy
//  support for ctors, dtors, conversion operators, operator overloading, etc
//
//  Revision 1.5  1998/04/22 12:02:05  shaggy
//  Changed ID to Id to be consistent with naming conventions
//
//  Revision 1.4  1998/01/12 04:12:48  shaggy
//  got dependencies working when re-generating code
//
//  Revision 1.3  1998/01/11 05:06:33  shaggy
//  about to switch to standard ref/ptr semantics
//
//  Revision 1.2  1996/10/18  01:40:54  shaggy
//  mostly finished
//
//  Revision 1.1  1996/03/27  22:57:34  shaggy
//  Initial revision
//
  

#define IMPLEMENTATION_SymTab
#include <SymTab>
#include <new>

// XXX - not thread safe


// obj_t Type_vSymTab;

// note: the cache mechanism here is currently broken


SymTab::SymTab(vStrTab *strtab) 
{
	_strtab = strtab;
	_first = 0;
	_cache = 0;
	_inum = 0;
	_inext = 0;
	_parent = 0;
	_data = 0;
}

SymTab::SymTab(vSymTab *parent)
{
	new(this)SymTab(parent->strtab());
	_parent = parent;
}

SymTab::SymTab(vSymTab *parent, void *data)
{
	new(this)SymTab(parent);
	_data = data;
}


SymTab::~SymTab()
{
	_SymEntry *sp = _first;

	while(sp) {
		_SymEntry *spnext = sp->next;
		// XXX - should probably try to free the value...
		// but we don't really know its type or any way to 
		// access it...

		delete sp;
		sp = spnext;
	}
}


vSymTab*
SymTab::parent() const
{
	return _parent;
}

bool
SymTab::parent(vSymTab *p)
{
	_parent = p;
	return true;
}

void*
SymTab::data() const
{
	return _data;
}

bool
SymTab::data(void *d)
{
	// if(_data)
	//	return false;
	// else {
		_data = d;
		return true;
	// }
}

vStrTab*
SymTab::strtab() const
{
	return _strtab;
}


bool SymTab::get(Id id, Val &val) const
{
	_SymEntry *psp = 0;
	_SymEntry *sp = _inext;
	if(!sp || sp->id != id)
		for(sp = _cache; sp && sp->id != id; sp = sp->cnext)
			psp = sp;
	if(!sp)
		return false;
	if(psp) {
		psp->cnext = sp->cnext;
		sp->cnext = _cache;
		_cache = sp;
	}
	val = sp->val;
	return true;
}	

bool SymTab::set(Id id, Val val)
{
	_SymEntry *psp = 0;
	_SymEntry *sp = _inext;
	if(!sp || sp->id != id)
		for(sp = _cache; sp && sp->id != id; sp = sp->cnext)
			psp = sp;
	if(!sp)
		return false;
	if(psp) {
		psp->cnext = sp->cnext;
		sp->cnext = _cache;
		_cache = sp;
	}
	sp->val = val;
	return true;
}

bool SymTab::add(Id id, Val val)
{
	Val tval;
	if(get(id, tval))
		return false;
	_SymEntry *sp = new _SymEntry;
	sp->id = id;
	sp->val = val;
	sp->next = _first;
	sp->cnext = _cache;
	_first = sp;
	_cache = sp;
	return true;
}


bool SymTab::del(Id id)
{
	_SymEntry *psp = 0;
	_SymEntry *sp = _inext;
	if(sp && sp->id == id) {
		_inext = _inext->cnext;
		++_inum;
	} else {
		for(sp = _cache; sp && sp->id != id; sp = sp->cnext)
			psp = sp;
		if(!sp)
			return false;
		if(psp) 
			psp->cnext = sp->cnext;
		else
			_cache = sp->cnext;
	}
	psp = _first;
	if(psp == sp)
		_first = sp->next;
	else {
		while(psp->next != sp)
			psp = psp->next;
		psp->next = sp->next;
	}
	delete sp;
	return true;
}


// Although index() will allow random access, only forward sequential
// access will be efficient
// Also, as long as the cached entry is used, other threads inserting
// or deleting elements will not cause dropouts 
// but if the cache isn't used, the ordering could change, and
// dropouts or repeats could occur

SymTab::Id SymTab::index(uint index) const
{
	if(!index || index < _inum) {
		_inum = 0;
		_inext = _first;
	}
	while(_inext && _inum < index) {
		_inext = _inext->next;
		++_inum;
	}
	return _inext ? _inext->id : 0;
}	

