/*  $Id: types.h,v 1.6 1998/04/17 18:02:02 shaggy Exp $
 *      relative path: ./lib
 *  
 *      Shag/OS: an Object-Oriented MicroKernel Operating System
 *       Libraries
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the Shag/OS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@mail.csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

/*  $Log: types.h,v $
 *  Revision 1.6  1998/04/17 18:02:02  shaggy
 *  more stuff to make it build under Solaris
 *
 *  Revision 1.5  1998/02/11 04:55:47  shaggy
 *  added cz and csz
 *
 *  Revision 1.4  1997/06/21 12:25:29  shaggy
 *  header update
 *
 *  Revision 1.3  1997/01/21 05:18:28  shaggy
 *  header update
 *
 *  Revision 1.2  1996/12/20 18:14:17  shaggy
 *  modified to use new standard types
 *
 *  Revision 1.1  1996/12/08  18:07:31  shaggy
 *  Initial revision
 *
 *      creation date: Dec 1994
 */

#include <inttypes.h>

#define _K_	1024			// Kilo (2^10)
#define	_M_	1048576			// Mega (2^20)
#define _G_	1073741824		// Giga (2^30)
#define _T_	1099511627776		// Tera (2^40)

#ifndef ASM

#if 0	/* Not needed unless inttypes.h is switched to use '_t' extensions */
typedef int8_t		int8;
typedef int16_t		int16;
typedef int32_t		int32;
typedef int64_t		int64;
typedef uint8_t		uint8;
typedef uint16_t	uint16;
typedef uint32_t	uint32;
typedef uint64_t	uint64;
#endif

/* The general use of these should be phased out, and replaced with
 * more explicit types from inttypes.h
 * However, inttypes.h relies on these types being defined, so these
 * must remain here. (actually, they should be defined earlier,
 * or in inttypes.h...)
 */
typedef unsigned char		uchar;
typedef signed char		schar;
typedef unsigned short int	ushort;
typedef unsigned long int	ulong;
typedef unsigned int		uint;
typedef unsigned long long int	uvlong;
typedef long long int		vlong;
 

typedef uint8	flag8;
typedef uint16	flag16;
typedef uint32	flag32;
typedef uint64	flag64;

/* Strictly speaking, these aren't necessary, but they aid in 
 * the clarity and conciseness of code.  `char*' by itself is fairly
 * short, but `const char *' starts getting rather large, and is used
 * quite commonly.  In addition, `char*' doesn't make it clear whether
 * it is a pointer to a single character or an array of characters, and
 * in no way shows whether that array is supposed to be zero terminated.
 * The following types make it clear for zero terminated strings:
 * (Note that the pointer was NOT included in these typedefs, so that
 * where they are used a pointer is still required, making it very
 * clear that they aren't self-contained entities, but instead that
 * they reference memory elsewhere, which may need to be explicitly copied)
 */
typedef char sz;
typedef const char csz;

// Note that 'signed char' and 'unsigned char'
// are *not* included as builtin char types, because
// when the signed-ness of a char type is being depended upon,
// it is generally being used as an integer, not a character.

#ifndef FOR_ALL_BUILTIN_CHAR_TYPES
#define FOR_ALL_BUILTIN_CHAR_TYPES(x)	\
		x(char) x(wchar_t)
#endif

#ifndef FOR_ALL_BUILTIN_INT_TYPES
#define FOR_ALL_BUILTIN_INT_TYPES(x)	\
		x(int8) x(uint8) \
		x(int16) x(uint16) \
		x(int32) x(uint32) \
		x(int64) x(uint64)
#endif

#ifndef FOR_ALL_BUILTIN_FLOAT_TYPES
#define FOR_ALL_BUILTIN_FLOAT_TYPES(x)	\
		x(float) x(double) x(long double)
#endif

#ifndef FOR_ALL_BUILTIN_TYPES
#define FOR_ALL_BUILTIN_TYPES(x)	FOR_ALL_BUILTIN_CHAR_TYPES(x)	\
					FOR_ALL_BUILTIN_INT_TYPES(x)	\
					FOR_ALL_BUILTIN_FLOAT_TYPES(x)	
#endif


#ifndef USE_STDLIB

#if !defined(__cplusplus) || defined(need_wchar_t)
	typedef ushort		wchar_t;
#endif

#ifndef _TIME_T
#	define _TIME_T
	typedef int64 time_t;		// in nanosecondss,
#endif					// starting at the epoch,
					// midnight Jan 1, 1970

#define	SEC	(time_t)1000000000	// one second in nanoseconds
#define MSEC	(time_t)1000000		// one millisecond
#define USEC	(time_t)1000		// one microsecond
#define NSEC	(time_t)1		// one nanosecond

#ifndef _PTRDIFF_T
#	define _PTRDIFF_T
#	ifdef __PTRDIFF_TYPE__
		typedef __PTRDIFF_TYPE__ ptrdiff_t;
#	else
		typedef vaddr_t	ptrdiff_t;
#	endif
#endif

#ifndef _SIZE_T
#	define _SIZE_T
#	ifdef __SIZE_TYPE__
		typedef __SIZE_TYPE__ size_t;
#	else
		typedef	vaddr_t	size_t;
#	endif
#endif

#ifndef _SSIZE_T
#	define _SSIZE_T
	typedef svaddr_t ssize_t;
#endif

#ifndef _COUNT_T
#	define _COUNT_T
	typedef int32	count_t;
#endif

#ifndef _WINT_T
#	define _WINT_T
	typedef wchar_t	wint_t;
#endif

#ifndef NULL
#	ifdef __cplusplus
#		define NULL		0
#	else
#		define NULL		(void*)0
#	endif
#endif

#ifndef NULLP
#	define NULLP(t)	(t*)0
#endif

#endif 	/* USE_STDLIB */
#endif	/* ASM */

