/*  $Id: dcobj.hh,v 1.13 1998/08/22 03:36:08 shaggy Exp $
 *      Dynamic Object standard header
 *      
 *      relative path: ./util/deco
 *  
 *      Shag/OS: an Object-Oriented MicroKernel Operating System
 *       Utilities for compilation, configuration, and development
 *        Dynamic Encapsulator of C++ Objects
 *  
 *  Copyright (c) 1996,1997 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the Shag/OS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

//  $Log: dcobj.hh,v $
//  Revision 1.13  1998/08/22 03:36:08  shaggy
//  header update
//
//  Revision 1.12  1998/08/11 12:50:44  shaggy
//  getting conditional assignment to work
//
//  Revision 1.11  1998/08/10 14:33:35  shaggy
//  misc
//
//  Revision 1.10  1998/08/09 21:49:55  shaggy
//  DECO *finally* converts all of my original testcases automatically
//

#if (__GNUC_MINOR__ > 7) || (__GNU__ > 2)
/* so we can  convert method pointers to ordinary pointers: */
#pragma Make Flags_Needed += -Wno-pmf-conversions
#endif

#include <inttypes.h>
#include <defines>

class __dcobj__;	// to let deco know this file has been included

// some lexical shortcuts:
#ifndef dclass
#	ifdef __dcplusplus
#		define dclass	dynamic class
#	else
#		define dclass	class
#	endif
#endif
#ifndef dtype
#	ifdef __dcplusplus
#		define dtype	dynamic virtual class
#	else
#		define dtype	class
#	endif
#endif


// class _DCBind;
dclass D_Type;


extern "C" void *malloc(size_t);

#if 0

typedef struct {
	char byte[1];
} *paddr_t;

typedef struct {
	char byte[1];
} *vaddr_t;


#define ALLOCA(nbytes)	reinterpret_cast<vaddr_t>(alloca(nbytes))

// #define MFPTR_TO_FPTR(f)  (reinterpret_cast<void**>(&f)[1])

#endif


// subject to change... therefore, the macros ought to be used

// DBIND = dynamically checked bind
// DASSIGN = dynamically checked assignment

#if __dcplusplus
#	define DBIND(t,r,o)		t r = o
#	define DCBIND(t,r,o)		t r ?= o
#	define DREBIND(r,o)		&r = o
#	define DCREBIND(r,o)		&r ?= o
#	define DASSIGN(r,o)		r = o
#	define DCASSIGN(r,o)		r ?= o
#else
#	define DBIND(t,r,o)		t r = t(o)
//#	define DCBIND(t,r,o)		t r = t(o, t::_DCBIND)
//#	define DCBIND(t,r,o)		t r = o.__dcheck()
#	define DCBIND(t,r,o)		t r = o.dcbind(r)
#	define DREBIND(r,o)		o.dbind(&r)
#	define DCREBIND(r,o)		o.dcbind(&r)
#	define DASSIGN(r,o)		r = o
#	define DCASSIGN(r,o)		r.dcassign(o)
//#define ????				o.dbind(static_cast<t&>(0))
#endif

// Examples:

// if(DBIND(r_type, r, obj)) foo(r);
// if(r_type r ?= obj) foo(r);

// DASSIGN(w,obj);
// w.dassign(obj);
// w ?= obj;
 
// r.dbind(obj);
// &r = obj;		// allowable 

// DASSIGN(&r,obj);
// r.dcheck_dbind(obj);
// &r ?= obj;
// &r.dassign(obj);



inline void* operator new(size_t, void *where)
{
	return where;
}



enum __NoInit { __NO_INIT };


//class it_class;
//class it_type;
//typedef const it_class &it_class_;
//typedef const it_type &it_type_;
class ot_class;
class ot_type;
typedef const ot_class &ot_class_;
typedef const ot_type &ot_type_;

class Obj {
public:
	uintptr _class;		// bits 0-1:
				//	0:  inline data (3*sizeof(void*) bytes)
				//      1:  remote data
				//      2:  reserved
				//      3:  local data
	union {
	    struct {
		void *p1;
		void *p2;
		void *p3;
	    } _inline;
	    struct {
		mutable uintptr datamgr;// bit 0: data has been read
				// bit 1: data has been written
		void *data;	// pointer to data
		void *spare;	// reserved for use by data manager
	    } _local;
	    struct {
		uintptr datamgr;// low 2 bits are reserved for datamgr flags	
		struct {
			void *p1;
			void *p2;
		} handle;
	    } _remote;
	} u;

	// typedef int (Obj::**_itable_entry)(int param);

	Obj() {}
	inline Obj(ot_class_ clss, void *data); 
	inline Obj(ot_class_ clss, size_t size);
	operator bool() const { return _class != 0; }
// protected:
	// data as well?
	inline const void* _this_r() const;
	inline void* _this_w();
	inline void* _this_rw();
	void* _this_inline()
		{ _class &= ~3; return &u._inline; }
	void* _this_local(void *p)
		{ _class |= 3; return (u._local.data = p); }
	void _this_remote() { _class = (_class & ~3) | 1; }
	inline void* _init_alloc(ot_class_, size_t);
	inline void* _init_ref(ot_class_, void *);
};


class ot_type : public Obj {};
class ot_class : public Obj {};

inline const void* Obj::_this_r() const
{
	switch(_class & 3) {
	case 0: return &u._inline;
	case 3: u._local.datamgr |= 2; return u._local.data;
	default: return 0;	// invoke datamgr
	}
}

inline void* Obj::_this_w()
{
	switch(_class & 3) {
	case 0: return &u._inline;
	case 3: u._local.datamgr |= 1; return u._local.data;
	default: return 0;	// invoke datamgr
	}
}

inline void* Obj::_this_rw()
{
	switch(_class & 3) {
	case 0: return &u._inline;
	case 3: u._local.datamgr |= 3; return u._local.data;
	default: return 0;	// invoke datamgr
	}
}

	
inline void* Obj::_init_alloc(ot_class_ clss, size_t size)
{
	if(size <= sizeof(u._inline)) {
		_class = reinterpret_cast<uintptr>(&clss);
		return &u._inline;
	} else {
		_class = reinterpret_cast<uintptr>(&clss) | 3;	
		return (u._local.data = malloc(size));
	}
}

inline void* Obj::_init_ref(ot_class_ clss, void *data)
{
	_class = reinterpret_cast<uintptr>(&clss) | 3;
	return (u._local.data = data);
}

inline Obj::Obj(ot_class_ clss, void *data) 
{
	_init_ref(clss, data);
}

inline Obj::Obj(ot_class_ clss, size_t size)
{
	_init_alloc(clss, size);
}



// class _OCF;


class OCFrame		// Object Call Frame
{
};

dclass D_Type;
typedef const D_Type &D_Type_;
dclass D_Class;
typedef const D_Class &D_Class_;

class BTRef		// Bound Typed Reference 
{
public:
	void **_itabp;
	Obj *_objp;
	Obj *self;
	const D_Type * const _type;
	// Obj dcall(ObjType *type, int method, Obj **param, int nparam);
	// Obj _obj;
protected:
	BTRef(__NoInit) {}
	BTRef(D_Type_ type) : _type(&type) {}
	BTRef(D_Type_ type, Obj *obj);
public:
	enum _DCBind { _DCBIND };

	// should the bind operation itself return a bool, such that
	// operator bool() is again useable?  If so, how can we later
	// check the validity of a type reference?
	// If operator bool not allowed, then how does one implement
	// a boolean object?  Explicitly compare to TRUE/FALSE?
	operator bool() const { return _objp != 0; }

	Obj* _this_obj() const { return _objp; }
	BTRef* _this_ref() const
		{ return reinterpret_cast<BTRef*>(_objp); }
	const void* _this_r() const { return _objp->_this_r(); }
	void* _this_w() const { return _objp->_this_w(); }
	void* _this_rw() const { return _objp->_this_rw(); }
	
	bool _this_is_self() const { return _objp == self; }
#if 0
	Obj mcall(int method, Obj **param, int nparam)
		{ return (_objp->*(_itb()->mcall__0))(method, param, nparam); }
#endif
	// bool d(Obj &);
	operator Obj& () const { return *_objp; }
//	_BTRef& operator& () { return reinterpret_cast<_BTRef&>(*this); }
};

#if 0
class _BTRef : public BTRef {
protected:
	_BTRef() : BTRef({}
public:
	// BTRef& operator= (BTRef
	// bool dcheck_assign()
};
#endif


/*
int Obj::dbind(class BTDesc *type) const
{
	if(_class->_class == &D_Class) {
		((mcall_func*)(_class->_data))(this, int(type));
	} else
		printf("Unknown class type\n"); 
}
*/


// The complexity below is needed to handle all the
// "chicken and egg" problems that result from trying to
// build the standard types and classes:

#ifdef __dcplusplus
#	include "t_type.dh"
#	include "t_class.dh"
#	include "D_Class.dh"
#	include "D_Type.dh"
#else
#	include "t_type.dh.h"
#	include "t_class.dh.h"
#	include "D_Class.dh.h"
#	include "D_Type.dh.h"
#endif

#if 0

#include "t_type.dh.h"
#include "t_class.dh.h"

#include "D_Class.dh.h"
#include "D_Type.dh.h"

#include "t_int.dh.h"
//#include "t_char.dh.h"
//#include "t_ptr.dh.h"
//#include "t_ref.dh.h"
//#include "t_arr.dh.h"
//#include "t_str.dh.h"

//#include "DM_Int.dh.h"
//#include "DM_Char.dh.h"
//#include "DM_Ptr.dh.h"
//#include "DM_Ref.dh.h"
//#include "DM_Arr.dh.h"
//#include "D_Str.dh.h"
#include "D_Int32.dh.h"


#endif

